/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.assemblies;

import ist.palcom.communication.CommunicationManager;
import ist.palcom.connections.ConnectionsManager;
import ist.palcom.device.AbstractDeviceContext;
import ist.palcom.device.DeviceContext;
import ist.palcom.device.IDeviceIO;
import ist.palcom.device.RemoteConnectManager;
import ist.palcom.device.Service;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.manager.HgraphManager;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.MediaManager;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.pubsub.PubSubManager;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.manager.ResourceManager;
import ist.palcom.resourceAddressing.ResourceAddressingManager;
import ist.palcom.routing.RoutingManager;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * An AssemblyDeviceContext is a DeviceContext for devices that support
 * assemblies. It is created as a decorator for another DeviceContext (decorator
 * pattern).
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class AssemblyDeviceContext implements DeviceContext {

    private DeviceContext context;

    private IAssemblyManager assemblyManager;

    private IResourceManager resourceManager;

    public AssemblyDeviceContext(DeviceContext context) {
        this.context = context;
    }

    // FIXME: this should probably do addService. Correspondingly, setting to
    // null should do removeservice of any previous AssemblyManager
    // ditto for ResourceManager
    public void setAssemblyManager(IAssemblyManager m) {
        assemblyManager = m;
    }

    public IAssemblyManager getAssemblyManager() {
        return assemblyManager;
    }

    public void setResourceManager(IResourceManager m) {
        resourceManager = m;
    }

    public IResourceManager getResourceManager() {
        return resourceManager;
    }

    public void addServiceAtDeviceLevel(Service service) {
        context.addServiceAtDeviceLevel(service);
    }
    
    public void addServiceToList(Service service, LocalSID listLocalSID) {
    	context.addServiceToList(service, listLocalSID);
    }
    
    public void addListOfServicesToList(LocalSID listLocalSID, ArrayList services, PRDSubList subList) {
    	context.addListOfServicesToList(listLocalSID, services, subList);
    }

    public ConnectionsManager getConnectionsManager() {
        return context.getConnectionsManager();
    }

    public IDeviceIO getDeviceIO() {
        return context.getDeviceIO();
    }

    public PRDDevice getDeviceInfo() {
        return context.getDeviceInfo();
    }

    public DiscoveryManager getDiscoveryManager() {
        return context.getDiscoveryManager();
    }

    public ResourceManager getResourceManager2() {
    	return context.getResourceManager2();
    }
    
    public MediaManager getMediaManager() {
        return context.getMediaManager();
    }

    public String getName() {
        return context.getName();
    }

    public PointToPointManager getPointToPointManager() {
        return context.getPointToPointManager();
    }

    public PubSubManager getPubSubManager() {
        return context.getPubSubManager();
    }

    public RemoteConnectManager getRemoteConnectManager() {
        return context.getRemoteConnectManager();
    }

    public PalcomScheduler getScheduler() {
        return context.getScheduler();
    }

    public void removeService(Service service) {
        context.removeService(service);
    }
    
	public void removeSubList(PRDSubList subList) {
		context.removeSubList(subList);
	}

    public Iterator serviceIterator() {
        return context.serviceIterator();
    }

    public void setDeviceIO(IDeviceIO io) {
        context.setDeviceIO(io);
    }
    
    public void setHeartBeatPeriod(long period) {
    	context.setHeartBeatPeriod(period);
    }
    
    public String getNewInstanceNumber(ServiceID serviceID) {
    	return context.getNewInstanceNumber(serviceID);
    }
    
    public String getNewLocalSIDString() {
    	return context.getNewLocalSIDString();
    }
    
    public LocalSID getNewLocalSID() {
    	return context.getNewLocalSID(); 
    }

    public void setName(String name) {
        context.setName(name);
    }
    
    public HgraphManager getHgraphManager(){
    	return context.getHgraphManager();
    }
    
    public CommunicationManager getCommunicationManager() {
    	return context.getCommunicationManager();
    }
    
    public RoutingManager getRoutingManager() {
    	return context.getRoutingManager();
    }
    
    public ResourceAddressingManager getResourceAddressingManager() {
    	if(context instanceof DeviceContext) 
    		return ((DeviceContext)context).getResourceAddressingManager();
    	else
    		throw new Error("Context does not provide a ResourceAddressingManager");
    }
    
	public void addListOfServicesAtDeviceLevel(ArrayList newServices, PRDSubList subList) {
		context.addListOfServicesAtDeviceLevel(newServices, subList);
	}

    
/****************hgraph impl below********************/
	
	public String getNodeName(){return context.getNodeName();}
	public int setNodeName(String name, Path callerPath){return context.setNodeName(name, callerPath);}
	public Path getPath(){return context.getPath();}
	public void updatePath(){context.updatePath();}
	public boolean hasChildNamed(String name){return context.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return context.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return context.removeNodeChild(name, callerPath);}
	public String toXML(){return context.toXML();}
	public String toXMLFull(){return context.toXMLFull();}
	public String toXMLRecursively(){return context.toXMLRecursively();}
	public String toXMLFullRecursively(){return context.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return context.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return context.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return context.getSiblingFunc();};
	public ValueFunc getValueFunc(){return context.getValueFunc();};
	public FuncFunc getFuncFunc(){return context.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return context.getNodeParent();}
	public Iterator getChildIterator(){return context.getChildIterator();}
}
