/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.assemblies;

import ist.palcom.debugging.DebugDispatcher;
import ist.palcom.device.DeviceContext;
import ist.palcom.discovery.CommandEvent;
import ist.palcom.resource.descriptor.AbstractResourceAddress;
import ist.palcom.resource.descriptor.Action;
import ist.palcom.resource.descriptor.ActionWithParams;
import ist.palcom.resource.descriptor.AssemblyID;
import ist.palcom.resource.descriptor.AssignAction;
import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.CommandInfo;
import ist.palcom.resource.descriptor.ControlItem;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.EventHandlerClause;
import ist.palcom.resource.descriptor.EventHandlerList;
import ist.palcom.resource.descriptor.EventHandlerScript;
import ist.palcom.resource.descriptor.InvokeAction;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDAssemblyVer;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.Param;
import ist.palcom.resource.descriptor.SelfTestAction;
import ist.palcom.resource.descriptor.SendMessageAction;
import ist.palcom.resource.descriptor.ServiceAddress;
import ist.palcom.resource.descriptor.ServiceDecl;
import ist.palcom.resource.descriptor.ServiceExp;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.resource.descriptor.ServiceUse;
import ist.palcom.resource.descriptor.TemporaryAddress;
import ist.palcom.resource.descriptor.ThisProviderService;
import ist.palcom.resource.descriptor.ThisProviderServiceAddress;
import ist.palcom.resource.descriptor.ThisService;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.resource.descriptor.VariableDecl;
import ist.palcom.resource.descriptor.VersionPart;
import ist.palcom.services.AbstractService;
import ist.palcom.services.ServiceProxyEvent;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class ScriptedAssembly extends Assembly {

	private AbstractService cp;
	private PRDSubList subList;

	/** This constructor also adds the assembly instance's sublist and services to the device context
	 * 
	 * @param context
	 * @param assemblyInfo
	 * @param period
	 * @throws IOException
	 */
	public ScriptedAssembly(AssemblyDeviceContext context, PRDAssemblyVer assemblyInfo, int period) throws IOException {
        super(context, assemblyInfo, period, getLocalSidIfInInfo(assemblyInfo, context));
        if (!validateScript()) {
        	DebugDispatcher.debug("ScriptedAssembly:failed to validate script", DebugDispatcher.CAT_ASSEMBLY);
        }
        ArrayList services = new ArrayList();
        services.add(this);
        createServiceProvider(services);
        subList = new PRDSubList(assemblyInfo.getName(), context.getNewLocalSID(), PRDSubList.KIND_ASSEMBLY);
		context.addListOfServicesAtDeviceLevel(services, subList);
    }

	/** Fix for stupid java constructor syntax restrictions 
	 **/
	private static LocalSID getLocalSidIfInInfo(PRDAssemblyVer assemblyInfo, DeviceContext c) {
		AbstractResourceAddress ra = assemblyInfo.getCustServiceAddr();
		if(ra instanceof TemporaryAddress) {
			return ((TemporaryAddress)ra).getLocalSID(); 
		} else if(ra instanceof ServiceAddress && ((ServiceAddress)ra).hasPRDService()) {
			return ((ServiceAddress)ra).getPRDService().getLocalSID();
		} else if(ra instanceof ServiceAddress && ((ServiceAddress)ra).hasLocalSID()) {
			return ((ServiceAddress)ra).getLocalSID();
		}
		System.err.println("Warning! ScriptedAssembly.getLocalSidIfInInfo generates new LocalSID");
		return c.getNewLocalSID();

	}		

	/** Create the service provider(s) and add to service list
	 * 
	 * @param l - the list 
	 */
	private void createServiceProvider(List l) {
		if(assemblyVer.hasPRDServiceFMDescription()) {
			ServiceProxy sd = (ServiceProxy) assemblyVer.getPRDServiceFMDescription().createInstance();
			System.err.println("ScriptedAssembly.createServiceProvider: "+sd.getID());
			try {
//			cp = new AbstractService(context, sd.getProviderURN(), sd.getID(), sd);
				cp = new AbstractService(context, assemblyVer.getPRDServiceFMDescription().getID(),
						true, false, PRDService.UNICAST, "P1", false, "1", "Synthesized service", sd);
				assemblyVer.getPRDServiceFMDescription().setLocalSID(cp.getServiceInfo().getLocalSID());
	        	l.add(cp);
	        	AssemblyID version = assemblyVer.getVersion();
	        	VersionPart vp1 = version.getCreatingVersionPart();
				VersionPart vp2 = version.getUpdatingVersionPart();
				VersionPart vp3 = version.getPreviousVersionPart();
				VersionPart vp4 = version.getMergedFromVersionPart();
				if(vp1 != null) {
					vp1 = (VersionPart) vp1.fullCopy();
					vp1.setServicePartID(vp1.getServicePartID()+"p");
				}
				if(vp2 != null) {
					vp2 = (VersionPart) vp2.fullCopy();
					vp2.setServicePartID(vp2.getServicePartID()+"p");
				}
				if(vp3 != null) vp3 = (VersionPart) vp3.fullCopy();
				if(vp4 != null) vp4 = (VersionPart) vp4.fullCopy();
				//XXX HERE BE DRAGONS!
				ServiceID serviceID = new ServiceID(vp1, vp2, vp3, vp4);
//	    		ServiceID serviceID = new ServiceID(
//	            		new DeviceID("X:1"), "TF4", new DeviceID("X:1"), "TF4", null, null, null, null);
	    		cp.setServiceInstanceID(new ServiceInstanceID(context.getDeviceInfo().getDeviceID(), serviceID, 
	    				context.getNewInstanceNumber(serviceID)));

//				super(context, "Gps Service: "+name, true, false, PRDService.UNICAST, "P1", 
//						false, "1", "A GPS service", createServiceProxy(name), 
//						new LocalSID(context.getDeviceInfo().getDeviceID(), context.getNewLocalSIDString()));
				
				
//	        	System.out.println("Creating "+cp.getServiceProxy().getInfo());
//	        	System.out.println("urn="+cp.getServiceProxy().getProviderURN());
//	        	System.out.println("ID="+cp.getServiceProxy().getID());
	        	sd.addOutCommandReceiver(commandThread);
	        	sd.addInCommandReceiver(commandThread);
//	        	addServiceDescription(sd);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
    /** Start the assembly. */
    public void start() throws IOException {
        super.start();
        if(cp != null) {
        	cp.start();
        }
    }

    /** Stop the assembly. */
    public void stop() throws IOException {
    	if(subList != null) {
    		context.removeSubList(subList);
    	}
        if (cp != null) {
            context.getDiscoveryManager().removeLocalServiceDescription(
                    (PRDServiceFMDescription) cp.getServiceProxy().getInfo());
            context.removeService(cp);
            cp.stop();
        }
        super.stop();
    }


	/** Check that bound services have the messages used in the script, and that
	 *  parameter lists match.
	 *  
	 * @return false on errors
	 */	
	protected boolean validateScript() {
		if(assemblyVer.hasEventHandlerScript()) {
			boolean result = true;
			EventHandlerScript ehs = assemblyVer.getEventHandlerScript();
			for(int i=0; i<ehs.getEventHandlers().getNumEventHandlerClause(); i++)
				result = result && validateEventHandlerClause(ehs.getEventHandlers().getEventHandlerClause(i));
			return result;
		} else {
			return true;
		}
	}

	private boolean validateEventHandlerClause(EventHandlerClause ehc) {
//		CommandInfo ci = PalcomBrowserPlugin.doGetCommandInfoFromEHC(ehc);
		if(!ehc.hasCommandInfo()) {
			DebugDispatcher.debug("validateEventHandlerClause: no commandInfo...", DebugDispatcher.CAT_ASSEMBLY);
			return false;
		}
		CommandInfo ci = ehc.getCommandInfo();
		boolean result = true;
		if(ci != null) {
			for(int i=0; i<ehc.getNumAction(); i++) {
				result = result | validateAction(ehc.getAction(i), ci);
			}
			return result;
		} else {
			return false;
		}
	}

	private boolean validateAction(Action a, CommandInfo ci) {
		if(a instanceof AssignAction) {
			return validateAssignAction((AssignAction) a,ci);
		} else if(a instanceof SendMessageAction) {
			return validateSendAction((SendMessageAction) a,ci);
		} 
		return false;
	}

	private boolean validateAssignAction(AssignAction a, CommandInfo ci) {
//		VariableUse v = a.getVariableUse();
//		ParamUse p =a.getParamUse();
//		if(v.)
//		
		return false;
	}

	private boolean validateSendAction(SendMessageAction a, CommandInfo ci) {
		return false;
	}

	/** Handle a CommandEvent received from a ServiceDescription. */
    protected void commandReceived(CommandEvent event) {
    	String cmd = event.getCommand().getID();
    	AbstractResourceAddress src = new ServiceAddress();
    	if(event.getSource() instanceof ServiceProxy) {
//    		System.err.println("ScriptedAssembly broken by API changes");
//    		throw new Error("ScriptedAssembly broken by API changes");
    		src = ((ServiceProxy)event.getSource()).getAddress().validate(context.getResourceAddressingManager());
//    		src = context.getResourceAddressingManager().validateAddress(src);    		
    	}
    	/* Logging */
		System.err.println("ScriptedAssembly("+assemblyVer.getName()+").commandReceived: [IN] "+cmd+" from "+src);

    	logMsg("[IN] "+cmd+" from "+src);

		
    	if(assemblyVer.hasEventHandlerScript()) {
    		EventHandlerList h = assemblyVer.getEventHandlerScript().getEventHandlers();
    		for(int i=0; i < h.getNumEventHandlerClause(); i++) {
    			EventHandlerClause ehc = h.getEventHandlerClause(i);
    			String theCmd = ehc.getCommandName();
//    			String theSrc = ehc.getSenderService();
    			
    			//URN matching does not work with alternatives...
    			//String theSrc = ehc.getServiceExp().getURN();
    			
    			boolean urnMatch = false;
    			try {
    				ServiceExp serviceExp = ehc.getServiceExp();
    				if(serviceExp instanceof ServiceUse) {
//    					System.err.println("... ServiceUse:" + ((ServiceUse)serviceExp).getAssemblyLabel());
    					//XXX This is ugly, but necessary...
    					ServiceDecl decl = serviceExp.getDecl();
    					decl = (ServiceDecl) decl.fullCopy();
    					decl.validateAddress(context.getResourceAddressingManager());
//    	    			System.err.println("ScriptedAssembly.commandReceived: urnMatch: decl = "+decl+"\n----------\n"+src);

    					if(decl != null) {
    						urnMatch = decl.lookupService(src) != null;
    					}
    				} else if (serviceExp instanceof ThisProviderService) {
//    	    			System.err.println("ScriptedAssembly.commandReceived: urnMatch: thisProviderService");
//    	    			System.err.println("Comparing "+src+" and "+((ThisProviderService)serviceExp).getServiceAddress());
    					urnMatch = src.equals( context.getResourceAddressingManager().validateAddress(((ThisProviderService) serviceExp).getServiceAddress()));
    				} else if (serviceExp instanceof ThisService) {
    					//XXX this is probably obsolete
    	    			System.err.println("ScriptedAssembly.commandReceived: urnMatch: thisService. This is an error...");
//    	    			System.err.println("Comparing "+src+" and "+((ThisService)serviceExp).getServiceAddress());
//    					urnMatch = src.equals( ((ThisService) serviceExp).getServiceAddress());
    	    			//XXX for testing: urnMatch = true;
    				}
    			} catch(Exception e){
    				DebugDispatcher.debug("ScriptedAssembly.commandReceived("+cmd+" from "+src+"): "+e, DebugDispatcher.CAT_ASSEMBLY);
    				e.printStackTrace();
    			}

//    			System.err.println("ScriptedAssembly("+assemblyVer.getName()+").commandReceived: urnMatch = "+urnMatch);

//    			if(cmd.equals(theCmd) && src.equals(theSrc)) {
    			if(cmd.equals(theCmd) && urnMatch) {
//    				System.out.println("********found HandlerClause");
    				handleEvent(ehc, event);
    			}
    		}
    	}
    }

    private void handleEvent(EventHandlerClause ehc, CommandEvent event) {
//		DebugDispatcher.debug("ScriptedAssembly("+assemblyVer.getName()+").handleEvent:"+ehc+" : "+event, DebugDispatcher.CAT_ASSEMBLY);
    	for(int i=0; i<ehc.getNumAction();i++) {
    		doAction(ehc.getAction(i), event);
    	}
    	
    }

	private void doAction(Action action, CommandEvent event) {
		//TODO: This should probably be done with a virtual method
		//      in the Action class instead. I.e., action.perform(event, context?, ...)
		if(action instanceof AssignAction) {
			doAssignAction(action, event);
		} else if(action instanceof SendMessageAction) {
			doSendMessageAction(action, event);
		} else if(action instanceof InvokeAction) {
			doInvokeAction(action, event);
		} else if(action instanceof SelfTestAction) {
			doSelfTestAction(action, event);
		}
	}

	private void doSelfTestAction(Action action, CommandEvent event) {

		SelfTestAction a = (SelfTestAction) action;
		byte [] val = a.getParamValue(0).getValue(event);
//		DebugDispatcher.debug("fillParams("+i+"):"+p.getID()+" : "+(val!=null?"length="+val.length:"null"), DebugDispatcher.CAT_ASSEMBLY);
//		DebugDispatcher.debug("fillParams("+i+"):"+p.getID()+" = "+new String(val), DebugDispatcher.CAT_ASSEMBLY);
		if(val != null) {
			String string = new String(val);
			DebugDispatcher.debug("SelfTestAction: val =["+string+"]", DebugDispatcher.CAT_ASSEMBLY);
			setSelfTestResult(string);
		}
	}

	private void doInvokeAction(Action action, CommandEvent event) {
//		DebugDispatcher.debug("ScriptedAssembly("+assemblyVer.getName()+").DoInvokeAction: ["+action+"]", DebugDispatcher.CAT_ASSEMBLY);
	    InvokeAction a = (InvokeAction) action;
	    String commandID = a.getCommand();
	    ServiceProxy sd = cp.getServiceProxy();
	    if(sd != null) {
	        ControlItem tmp = sd.findItem(commandID);
	        if(tmp instanceof Command) {
	            Command command = (Command) tmp;
	            fillParams(command, a, event);
	            logMsg("invoke "+commandID);
	            command.invoke();
	        } else {
	            DebugDispatcher.debug("ScriptedAssembly.doAction():Not a command "+tmp, DebugDispatcher.CAT_ASSEMBLY);
	        }
	    }
	}

	private void doSendMessageAction(Action action, CommandEvent event) {
		SendMessageAction a = (SendMessageAction) action;
		String commandID = a.getCommand();
//			String urn = a.getService();
		AbstractResourceAddress sa = a.getServiceExp().getServiceAddress().validate(context.getResourceAddressingManager());
//		System.err.println("Warning! ScriptedAssembly.doSendMessageAction broken: find LocalSID for SIID!");
//		LocalSID lsid = a.getServiceExp().getLocalSID();
//		System.err.println("... lsid ="+lsid);
//		AbstractResourceAddress tmpAddr = context.getResourceAddressingManager().validateAddress(sa);
		// XXX fix this!
		LocalSID lsid=null;
		if(sa instanceof ServiceAddress) {
			ServiceAddress satmp = (ServiceAddress) sa;
			if(satmp.hasLocalSID())
				lsid = satmp.getLocalSID();
		}
		if (lsid==null) {
			DebugDispatcher.debug("ScriptedAssembly("+assemblyVer.getName()+").doAction():LocalSID not found:"+sa, DebugDispatcher.CAT_ASSEMBLY);
			return;
		}
		DebugDispatcher.debug("SendMessageAction:"+commandID+" -> "+sa,DebugDispatcher.CAT_ASSEMBLY);	
//			ServiceDescription sd = getServiceDescription(Info.urnToServiceURN(urn));
		ServiceProxy sd = getRemoteServiceProxy(lsid);
		
		if (sd != null) {
		    ControlItem tmp = sd.findItem(commandID);
		    if(tmp instanceof Command) {
		        Command command = (Command) tmp;
		        fillParams(command, a, event);
		        logMsg("[OUT] "+commandID+" to "+sa);
//		        logSendMsg(commandID,urn);
		        command.invoke();
		    } else {
		        DebugDispatcher.debug("ScriptedAssembly.doSendMessageAction():Not a command "+tmp, DebugDispatcher.CAT_ASSEMBLY);
		    }
		} else {
			DebugDispatcher.debug("ScriptedAssembly("+assemblyVer.getName()+").doAction():Service not found:"+sa, DebugDispatcher.CAT_ASSEMBLY);
//			Iterator it = proxies.entrySet().iterator();
			Iterator it = remoteServiceProxyIterator();
			while(it.hasNext()) {
				DebugDispatcher.debug(it.next().toString(),DebugDispatcher.CAT_ASSEMBLY);
			}
		}
	}

	private void doAssignAction(Action action, CommandEvent event) {
		AssignAction a = (AssignAction) action;
//			String var = a.getVariableUse().getName();
//			String paramName = a.getParamUse().getName();
//			ControlItem tmp = event.getCommand().findItem(paramName);
//			VariableDecl decl = assemblyInfo.lookupVariable(var);
		VariableDecl decl = a.getVariableUse().getDecl();
		Param tmp = a.getParamUse().getDecl(event);
		if(tmp != null && decl != null) {
			decl.value = tmp.getData();
		}
	}

	private void fillParams(Command command, ActionWithParams a, CommandEvent event) {
		for(int i=0; i<command.getNumParam();i++) {
			Param p = command.getParam(i);
            byte [] val = null;
            try {
                val = a.getParamValue(i).getValue(event);
            } catch (Exception e) {
                e.printStackTrace();
            }
//			DebugDispatcher.debug("fillParams("+i+"):"+p.getID()+" : "+(val!=null?"length="+val.length:"null"), DebugDispatcher.CAT_ASSEMBLY);
//			DebugDispatcher.debug("fillParams("+i+"):"+p.getID()+" = "+new String(val), DebugDispatcher.CAT_ASSEMBLY);
			if(val == null) {
				val = (p.getID()+" uninitialized").getBytes();
			}
			p.setData(val);
		}
	}
}
