/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.assembly.manager;

import ist.palcom.assemblies.Assembly;
import ist.palcom.assemblies.AssemblyDeviceContext;
import ist.palcom.assemblies.IResourceManager;
import ist.palcom.assemblies.ResourceRequestEvent;
import ist.palcom.assemblies.ResourceResponseEvent;
import ist.palcom.collections.LinearMap;
import ist.palcom.debugging.DebugDispatcher;
import ist.palcom.device.DeviceContext;
import ist.palcom.device.HLRemoteConnectHelper;
import ist.palcom.device.RemoteConnectHelper;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.palcomthreads.PeriodicEvent;
import ist.palcom.resource.descriptor.AbstractResourceAddress;
import ist.palcom.resource.descriptor.AbstractServiceDecl;
import ist.palcom.resource.descriptor.AltServiceDecl;
import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.ConnectionDecl;
import ist.palcom.resource.descriptor.ConnectionDeclList;
import ist.palcom.resource.descriptor.ConnectionInfo;
import ist.palcom.resource.descriptor.DeviceAddress;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.DeviceInfo;
import ist.palcom.resource.descriptor.Info;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.Opt;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.Param;
import ist.palcom.resource.descriptor.ServiceAddress;
import ist.palcom.resource.descriptor.ServiceDecl;
import ist.palcom.resource.descriptor.ServiceExp;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.descriptor.ServiceInfo;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.resource.descriptor.ServiceUse;
import ist.palcom.resource.descriptor.SingleServiceDecl;
import ist.palcom.resource.descriptor.ThisService;
import ist.palcom.resourceAddressing.AInfoEvent;
import ist.palcom.service.manager.ServiceManager;
import ist.palcom.services.AbstractService;

import java.io.IOException;
import java.util.Iterator;

/**
 * 
 * A class responsible for providing Assemblies with resources
 * 
 * This is the multi-threaded version, with one thread monitoring 
 * each request.
 * 
 * Currently, only connections are supported: An assembly sends requests
 * for connections, and the ResourceManager responds with an ConnectionInfo,
 * if the connection is possible (i.e., the required services have been
 * discovered)
 * 
 * If established resources go down, dependent assemblies are informed
 * 
 * For alternative resources, if a better alternative appears, dependent assemblies
 * are informed that a change is possible 
 * 
 * @author sven
 *
 */
public class NewResourceManager extends AbstractService implements IResourceManager{

	public static final String SERVICE_NAME = "ResourceManager";

	private ResourceThread resourceThread;

	private DeviceThread deviceThread;
	private InfoThread infoThread;

	private LinearMap connectionDeclToThread;
	private InfoMonitor infoMon;

	private AssemblyDeviceContext context() {
		return (AssemblyDeviceContext) context;
	}

	public NewResourceManager(DeviceContext context, String serviceName) throws IOException {
		super(context, serviceName, true, true, PRDService.UNICAST, "P1", 
				false, "1", "", createServiceProxy(serviceName), new LocalSID(
						context.getDeviceInfo().getDeviceID(), context.getNewLocalSIDString()));
		ServiceID serviceID = new ServiceID(
				new DeviceID("X:1"), "TF2", new DeviceID("X:1"), "TF2", null, null, null, null);
		setServiceInstanceID(new ServiceInstanceID(context.getDeviceInfo().getDeviceID(), serviceID, 
				context.getNewInstanceNumber(serviceID)));
		resourceThread = new ResourceThread();
		deviceThread = new DeviceThread();
		infoThread = new InfoThread();
		connectionDeclToThread = new LinearMap();
		infoMon = new InfoMonitor(context.getScheduler());
		context.getScheduler().scheduleThread(resourceThread);
		context.getScheduler().scheduleThread(deviceThread);
		context.getScheduler().scheduleThread(infoThread);
	}

	/**
	 * This constructor makes it possible to start the ResourceManager from the ServiceManager
	 */
	public NewResourceManager(DeviceContext context) throws IOException {
		this(context, SERVICE_NAME);
	}		
	public void start() throws IOException {
		super.start();
		context().getResourceAddressingManager().listenForDevices(deviceThread);
	}

	public void stop() throws IOException {
		resourceThread.terminate();
		infoThread.terminate();
		deviceThread.terminate();
		super.stop();
	}

	protected void infoReceived(AInfoEvent event) {
		Info i = event.getInfo();
		if(event.infoShouldBeKept()) {
			infoMon.addInfo(i);
		} else {
			infoMon.removeInfo(i);
		}
	}	

	/**
	 * A thread that receives ResourceRequests, and sets up or
	 * terminates handler threads for them 
	 */
	private class ResourceThread extends PalcomPriorityThread {

		public void run() {
			while (true) {
				Event tmpe = waitEvent();
				if(tmpe instanceof ResourceRequestEvent) {
					DebugDispatcher.debug("ResourceManager :"+tmpe,DebugDispatcher.CAT_RESOURCE);
					processResourceRequest((ResourceRequestEvent) tmpe);
				} 
			}
		}

	}

	private class DeviceThread extends PalcomPriorityThread {
		public void run() {
			while (true) {
				AInfoEvent event = (AInfoEvent) waitEvent();
				if(event.getInfo() instanceof DeviceInfo) {
					DeviceAddress da = ((DeviceInfo)event.getInfo()).getDeviceAddress();
					System.err.println("NewResourceManager.DeviceThread: listenForServices on "+da);
					context().getResourceAddressingManager().listenForServices(infoThread, da);
				}
			}
		}
	}
	private class InfoThread extends PalcomPriorityThread {
		public void run() {
			while (true) {
				AInfoEvent event = (AInfoEvent) waitEvent();
				infoReceived(event);
//				if(event.getInfo() instanceof PRDDevice) {
//				DeviceID deviceID = ((PRDDevice)event.getInfo()).getDeviceID();
//				context.getResourceManager2().updatePRDServiceListReceiver(new LocalSID(deviceID, null), this);
//				context.getResourceManager2().updatePRDConnectionReceiver(deviceID, this);
//				}
			}
		}
	}

	/** aux method
	 * 
	 * take care of differences between single and alternative declarations
	 */
	private void doProcessConnectionDecl(ConnectionDecl connectionDecl, ResourceRequestEvent rre) {
		ServiceExp prov = connectionDecl.getProvider();
		ServiceExp cust = connectionDecl.getCustomer();
		try {
			DebugDispatcher.debug("ResourceManager.doProcessConnectionDecl: "+connectionDecl+"("+rre.isActive()+")", DebugDispatcher.CAT_RESOURCE);
			if(rre.isActive()) {
				PalcomThread t;
				if(servicesAreSingle(prov, cust)) {
					SingleConnectionRequest singleConnectionRequest = new SingleConnectionRequest(connectionDecl, rre.getAssembly(), rre.getReceiver());
					t = new ConnectionRequestThread(singleConnectionRequest);
				} else {
					AltConnectionRequest altConnectionRequest = new AltConnectionRequest(connectionDecl, rre.getAssembly(), rre.getReceiver());
					t = new ConnectionRequestThread(altConnectionRequest);
				}
				context.getScheduler().scheduleThread(t);
				connectionDeclToThread.put(connectionDecl, t);
			} else {
				cancelConnectionRequest(connectionDecl);
			} 
		} catch(IllegalArgumentException e) {
			DebugDispatcher.debug("ResourceManager.doProcessConnectionDecl: "+e.getMessage(), DebugDispatcher.CAT_RESOURCE);
		} catch(Exception e) {
			DebugDispatcher.debug("ResourceManager.doProcessConnectionDecl: ("+prov+", "+cust+")"+e, DebugDispatcher.CAT_RESOURCE);
			e.printStackTrace();
		}
	}

	protected void processResourceRequest(ResourceRequestEvent rre) {
		PRDData d = rre.getDeclaration();
		if(d instanceof ConnectionDecl) {
			doProcessConnectionDecl((ConnectionDecl)d, rre);
		} else if( d instanceof ConnectionDeclList) {
//			System.out.println("ResourceManager.addResourceRequest:"+d);
			ConnectionDeclList cdl = (ConnectionDeclList) d;
			for(int i=0; i<cdl.getNumConnectionDecl(); i++){
				ConnectionDecl connectionDecl = cdl.getConnectionDecl(i);
				doProcessConnectionDecl(connectionDecl, rre);
			}
		} else {
			DebugDispatcher.debug("ResourceManager.addResourceRequest: unknown type:"+d, DebugDispatcher.CAT_RESOURCE);
		}
	}

	private void cancelConnectionRequest(ConnectionDecl decl) {
		PalcomThread t = (PalcomThread) connectionDeclToThread.remove(decl);
		if(t != null) {
			DebugDispatcher.debug("ResourceManager.cancelConnectionRequest: Killing thread for "+decl.getAssemblyLabel(), DebugDispatcher.CAT_RESOURCE);
			t.terminate();
		}
	}

	/** Check if a pair of serviceExp:s refer to single (i.e., not alternative) declarations 
	 */
	public boolean servicesAreSingle(ServiceExp prov, ServiceExp cust) {
		boolean provIsSingle = serviceExpRefersToSingle(prov);
		boolean custIsSingle = serviceExpRefersToSingle(cust);
		return provIsSingle && custIsSingle;
	}

	public boolean serviceExpRefersToSingle(ServiceExp se) {
		boolean seIsSingle=false;;
		if(se instanceof ThisService) {
//			DebugDispatcher.debug("ResourceManager.serviceExpRefersToSingle: this=true", DebugDispatcher.CAT_RESOURCE);
			seIsSingle = true;
		} else if(se instanceof ServiceUse) {
			seIsSingle = serviceUseRefersToSingle((ServiceUse)se);
//			DebugDispatcher.debug("ResourceManager.serviceExpRefersToSingle:"+se.toString()+"="+seIsSingle, DebugDispatcher.CAT_RESOURCE);
		} else {
			DebugDispatcher.debug("ResourceManager.serviceExpRefersToSingle:"+se.toString()+" not this or ServiceExp?", DebugDispatcher.CAT_RESOURCE);			
		}
		return seIsSingle;
	}

	private boolean serviceUseRefersToSingle(ServiceUse svc) {
		if(svc != null) {
			ServiceDecl decl = svc.getDecl();
			if(decl != null) {
				AbstractServiceDecl inner = decl.getDecl();
				return inner instanceof SingleServiceDecl;
			} else {
				DebugDispatcher.debug("ResourceManager.serviceUseRefersToSingle: no declaration found for "+svc, DebugDispatcher.CAT_RESOURCE);
				throw new IllegalArgumentException("No declaration found for Service "+svc.getIDforEditing());
			}
		}
		return false;
	}

	private static ServiceProxy createServiceProxy(String controlName) {
		ServiceProxy sd = new ServiceProxy(controlName);
		return sd;
	}

	public void putEvent(ResourceRequestEvent event) {
		resourceThread.putEvent(event);
	}

	private class SingleConnectionRequest extends ConnectionRequest {

		private boolean sat;
		private ConnectionInfo currentConnection;
		SingleConnectionRequest(ConnectionDecl cd, Assembly assembly, PalcomThread recv) {
			super(cd, assembly, recv);
		}		

		public boolean isAlternative() {
			return false;
		}

		public void satisfied(ConnectionInfo i) {
			System.out.println("***SingleConnectionRequest:satisfied:\n"+
					i.getProviderAddress().getDeviceAddress().getPRDDevice().getDeviceID()+
					"-->\n"+
					i.getCustomerAddress().getDeviceAddress().getPRDDevice().getDeviceID()
			);
			getReceiver().putEvent(new ResourceResponseEvent(getDeclaration(), i,ResourceResponseEvent.ADD));
			sat=true;
			currentConnection = i;
		}

		public void unsatisfied() {
			getReceiver().putEvent(new ResourceResponseEvent(getDeclaration(), currentConnection,ResourceResponseEvent.REMOVE));
			sat=false;
			currentConnection=null;
		}

		public boolean isActiveConnection(ConnectionInfo info) {
			return info.equals(currentConnection);
		}

		public boolean checkIfPossible() {
			//XXX
			return false;
		}

		public ConnectionInfo checkAgainstServices(Iterator it) {
			if(sat) return null;
			AbstractResourceAddress sa1 = getDeclaration().getCustomer().getServiceAddress().validate(context.getResourceAddressingManager());
			AbstractResourceAddress sa2 = getDeclaration().getProvider().getServiceAddress().validate(context.getResourceAddressingManager());
//			AbstractResourceAddress sa1 = context.getResourceAddressingManager().validateAddress(tmpSa1);
//			AbstractResourceAddress sa2 = context.getResourceAddressingManager().validateAddress(tmpSa2);
			ServiceInfo s1 = null;
			ServiceInfo s2 = null;
			while(it.hasNext() && (s1==null || s2==null)) {
				ServiceInfo tmp = (ServiceInfo) it.next();
//				AbstractResourceAddress tmp2 = context.getResourceAddressingManager().validateAddress(tmp.getServiceAddress());
				AbstractResourceAddress tmp2 = tmp.getServiceAddress().validate(context.getResourceAddressingManager());
				if(tmp2.equals(sa1)) {
					s1 = tmp;
//					System.err.println("SR.checkAgainstServices..."+getDeclaration().getAssemblyLabel()+" s1:"+sa1+
//					"\n"+sa1.getDeviceAddress().getPRDDevice().getDeviceID());
				} else if(tmp2.equals(sa2)) {
					s2 = tmp;
//					System.err.println("SR.checkAgainstServices..."+getDeclaration().getAssemblyLabel()+" s2:"+sa2+
//					"\n"+sa2.getDeviceAddress().getPRDDevice().getDeviceID());

				} 
			}
			if(s1 != null && s2 != null) {
				return new ConnectionInfo(sa1, sa2, new Opt());
			}
			return null;
		}

		public boolean isSatisfied() {
			return sat;
		}

		public void checkSoftwareManagers(Iterator it) {
			AbstractResourceAddress sa1 = getDeclaration().getCustomer().getServiceAddress().validate(context.getResourceAddressingManager());
			AbstractResourceAddress sa2 = getDeclaration().getProvider().getServiceAddress().validate(context.getResourceAddressingManager());
//			AbstractResourceAddress sa1 = context.getResourceAddressingManager().validateAddress(tmpSa1);
//			AbstractResourceAddress sa2 = context.getResourceAddressingManager().validateAddress(tmpSa2);
			ServiceInfo s1 = null;
			ServiceInfo s2 = null;
			while(it.hasNext() && (s1==null || s2==null)) {
				ServiceInfo tmp = (ServiceInfo) it.next();
//				AbstractResourceAddress tmp2 = context.getResourceAddressingManager().validateAddress(tmp.getServiceAddress());
				AbstractResourceAddress tmp2 = tmp.getServiceAddress().validate(context.getResourceAddressingManager());
				if(isServiceManager(tmp)) {
					if(s1 == null) {
						tryLoadingSoftwareComponent(tmp2, sa1);
					}
					if(s2 == null) {
						tryLoadingSoftwareComponent(tmp2, sa2);
					}
				}
			}
		}
	}

	private class AltConnectionRequest extends ConnectionRequest {

		private AltServiceDecl currentAlternative;
		private ConnectionInfo currentConnectionInfo;
		private boolean sat;

		AltConnectionRequest(ConnectionDecl cd, Assembly assembly, PalcomThread recv) {
			super(cd, assembly, recv);
		}

		public ConnectionInfo checkAgainstServices(Iterator i) {
			return null;
		}

		public ConnectionInfo getMatchingConnectionInfo(Info info) {
			return null;
		}

		public boolean isAlternative() {
			return true;
		}

		public synchronized void satisfied(ConnectionInfo i) {
			sat = true;
		}

		public synchronized void unsatisfied() {
			getReceiver().putEvent(new ResourceResponseEvent(getDeclaration(), currentConnectionInfo,ResourceResponseEvent.REMOVE));
			currentAlternative = null;
			currentConnectionInfo = null;
			sat = false;
		}

		public boolean isActiveConnection(ConnectionInfo info) {
//			return info.isAboutTheSameEntityAs(currentConnectionInfo);
			return currentConnectionInfo.isAboutTheSameEntityAs(info);
		}

		public boolean checkIfPossible() {
			//XXX 
			return false;
//			return currentConnectionInfo != null && remoteInfoHandler.connectionIsPossible(currentConnectionInfo);
		}

		public boolean isSatisfied() {
			// TODO:
			System.err.println("AltServiceDecl.isSatisfied unimplemented");
			return false;
		}
	}

	/** A class responsible for monitoring and acting in response
	 * to changes to the status of a single connection request.
	 * 
	 * @author sven
	 *
	 */
	private class ConnectionRequestThread extends PalcomPriorityThread {

		private static final int PERIOD = 20000;

		ConnectionRequest cr;

		public ConnectionRequestThread(ConnectionRequest connectionRequest) {
			cr = connectionRequest;
		}

		public void run() {
			cr.getDeclaration().getCustomer().listenForDiscovery(context().getResourceAddressingManager(), this, true);
			cr.getDeclaration().getProvider().listenForDiscovery(context().getResourceAddressingManager(), this, true);
			//XXX Periodic checking is not really necessary, but a simple fix for
			// the discoverymechanism being unreliable. Turn on periodic events if
			// there's a problem
			enablePeriodicEvents(PERIOD);
			while(true) {
				Event e = waitEvent();
				if(e instanceof AInfoEvent) {
					AInfoEvent ie = (AInfoEvent)e;
//					System.err.println("CRT: "+cr.getDeclaration().getAssemblyLabel()+"--"+ie.getInfo().getClass().getSimpleName());

					if(ie.infoShouldBeKept()) {
						infoMon.addInfo(ie.getInfo());
						ConnectionInfo result = infoMon.checkRequest(cr);
						if(result != null) {
							cr.satisfied(result);
						}
					}else if(ie.getInfo() instanceof ConnectionInfo && !ie.infoShouldBeKept()) {
						if(cr.isActiveConnection((ConnectionInfo) ie.getInfo())) {
							//			cr.unsatisfied();
						}

					} else if(ie.getInfo() instanceof DeviceInfo && !ie.infoShouldBeKept()) {
						//This is a fix for DiscoveryManager failing to send
						//events for disconnection caused by removal of a service						
//						System.err.println("**** Device removed");
						if(!cr.checkIfPossible()) {
//							System.err.println("***********NOT POSSIBLE!");
							cr.unsatisfied();
						}
					}
				} 
				if (e instanceof PeriodicEvent) {
//					// Check all known infos periodically
//					ConnectionInfo result = infoMon.checkRequest(cr);
//					if(result != null) {
//					cr.satisfied(result);
//					} else {
//					cr.unsatisfied();
//					}
					if(!cr.isSatisfied())
						infoMon.tryLoadingSoftwareComponent(cr);
				}
			}
		}

		public void terminate() {
			cr.getDeclaration().getCustomer().listenForDiscovery(context().getResourceAddressingManager(), this, false);
			cr.getDeclaration().getProvider().listenForDiscovery(context().getResourceAddressingManager(), this, false);
//			context.getResourceManager2().removePRDConnectionReceiver(cr.getDeclaration().getCustomer().getServiceAddress().getServiceInstanceID().getDeviceID(), this);
//			context.getResourceManager2().removePRDConnectionReceiver(cr.getDeclaration().getProvider().getServiceAddress().getServiceInstanceID().getDeviceID(), this);

			super.terminate();
		}
	}

	protected void tryLoadingSoftwareComponent(AbstractResourceAddress smadr, AbstractResourceAddress wanted) {
		if(wanted instanceof ServiceAddress && smadr.getDeviceAddress().equals(wanted.getDeviceAddress())) {
//			System.err.println("NewResourceManager.tryLoadingSoftwareComponent");
			ServiceAddress sa = (ServiceAddress)wanted;
			SoftwareComponentLoader t = new SoftwareComponentLoader(smadr, sa);
			t.perform();
		}
	}


	private class SoftwareComponentLoader {// extends PalcomPriorityThread {
		private AbstractResourceAddress smaddr;
		private ServiceAddress wanted;
		private HLRemoteConnectHelper rch;
		private ServiceAddress thisAddr;

		public SoftwareComponentLoader(AbstractResourceAddress smaddr, ServiceAddress wanted) {
			this.smaddr = smaddr;
			this.wanted = wanted;
			rch = new HLRemoteConnectHelper(context);
			ServiceInstanceID thisSiid = NewResourceManager.this.getServiceInstanceID();
			thisAddr = new ServiceAddress(thisSiid, new Opt(), new Opt());
		}

		public void perform() {
//			System.err.println("NewResourceManager.SCLT connecting...");
			ConnectionInfo ci;
			if((ci =connect()) != null) {
				System.err.println("NewResourceManager.SCLT trying to send load command...");
				if(smaddr.equals(ci.getProviderAddress())) 
					sendLoadCommand(ci.getProviderAddress(), wanted.getServiceInstanceID());
				else
					sendLoadCommand(ci.getCustomerAddress(), wanted.getServiceInstanceID());
				disconnect(ci);
			}
		}

		private void sendLoadCommand(AbstractResourceAddress address, ServiceInstanceID serviceInstanceID) {
			LocalSID lsid = address.tryGetLocalSID();
			if(lsid != null) {
//				System.err.println("NewResourceManager.SCLT sending load command...");
//				System.err.println("... lsid = "+lsid);
				ServiceProxy sp = getRemoteServiceProxy(lsid);
				if(sp != null) {
					Command c = (Command) sp.findItem(ServiceManager.COMMAND_LOAD);
					Param p = c.getParam(0);
					p.setData(serviceInstanceID.getXMLData());
					c.invoke();
				} else {
					System.err.println("WARNING! NewResourceManager.SCLT no serviceProxy found...");
				}
			}
		}

		private ConnectionInfo connect() {
			ConnectionInfo ci = new ConnectionInfo(smaddr, thisAddr, new Opt());
			return rch.doConnect(ci, 5000, 15000);
		}

		private void disconnect(ConnectionInfo ci) {
			context.getResourceAddressingManager().requestDisconnection(ci);
		}

	}

	private boolean isServiceManager(ServiceInfo tmp) {
		ServiceID serviceID = tmp.getServiceAddress().getServiceInstanceID().getServiceID();
		//		System.err.println("*** NewResourceManager.isServiceManager looking at "+tmp.getServiceAddress().getServiceInstanceID().getServiceID());
//		System.err.println("... and comparing to "+ServiceManager.getServiceIDStatic());
		return ServiceManager.isSameService(serviceID);
	}
}
