/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.base;

public class Coroutine {

    private boolean active;

    private CoroutineThread thread;

    public Coroutine() {
        setActive(false);
        thread = new CoroutineThread();
        thread.start();
    }

    public void run() {
    }

    public synchronized void attach() {
        setActive(true);
        notifyAll();
        while (active) {
            try {
                wait();
            } catch (InterruptedException e) {
                throw new CoroutineInterruptedError();
            }
        }
    }

    public synchronized void suspend() {
        if (!(Thread.currentThread() instanceof CoroutineThread)) {
            throw new RuntimeException(
                    "suspend was called by an \"ordinary\" Java thread. "
                            + "Only coroutines may call suspend.");
        }

        setActive(false);
        notifyAll();
        while (!active) {
            try {
                wait();
            } catch (InterruptedException e) {
                throw new CoroutineInterruptedError();
            }
        }
    }

    public synchronized void terminate() {
        if (!(Thread.currentThread() instanceof CoroutineThread)) {
            throw new RuntimeException(
                    "terminate was called by an \"ordinary\" Java thread. "
                            + "Only coroutines may call terminate.");
        }

        if (active) {
            throw new CoroutineInterruptedError();
        } else {
            thread.interrupt();
        }
    }

    private synchronized void setActive(boolean active) {
        this.active = active;
        notifyAll();
    }

    private class CoroutineThread extends Thread {
        public CoroutineThread() {
            // Use daemon threads, so the program gets terminated when the
            // main program ends
            setDaemon(true);

            // Set name for debugging
            setName("Coroutine "
                    + Integer.toHexString(Coroutine.this.hashCode()));
        }

        public void run() {
            try {
                Coroutine.this.suspend();
                synchronized (Coroutine.this) {
                    Coroutine.this.run();
                }
            } catch (CoroutineInterruptedError e) {
                // Terminated
            }
            setActive(false);
        }
    }

    /**
     * A special error that is thrown when a Coroutine thread gets interrupted
     * (for debugging purposes).
     */
    private static class CoroutineInterruptedError extends Error {
    }
}
