/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.connections.test;

import ist.palcom.communication.CommunicationManager;
import ist.palcom.connections.Connection;
import ist.palcom.connections.ConnectionEvent;
import ist.palcom.connections.ConnectionsManager;
import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.mal.DeviceCacheObject;
import ist.palcom.mal.Message;
import ist.palcom.mal.udp.UDPMediaManager;
import ist.palcom.mal.udp.test.UDPCommunicationTest2;
import ist.palcom.palcomthreads.Monitor;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.routing.RoutingManager;
import ist.palcom.routing.util.SimpleRoutingManager;

import java.util.ArrayList;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class ConnectionsTest {

    public static void main(String[] args) {
        // ist.palcom.base.System.instance().enableConstantGC();
        ConnectionsTest test = new ConnectionsTest();
        
        test.testConnectionMethods();
        System.out.println();
    }

    private void testConnectionMethods() {
        System.out.println("--- testConnectionMethods ---");

        final PalcomScheduler scheduler = new PalcomScheduler();
        PalcomThread t = new PalcomPriorityThread() {
            public void run() {
                AbstractMediaManager mediaManager = new UDPMediaManager(scheduler, null);
                ArrayList mediaManagers = new ArrayList();
                mediaManagers.add(mediaManager);
               
//              Generate hopefully unique mark ID for router.
        		String mark = "abc";
                
        		RoutingManager routingManager = new SimpleRoutingManager(mediaManagers, scheduler, mark);
                CommunicationManager communicationManager = new CommunicationManager(scheduler, 
                		routingManager);
                ConnectionsManager connectionsManager = new ConnectionsManager(
                        scheduler, communicationManager);
                
                final DeviceID providerDeviceID = new DeviceID("X:provider");

                ConnectionMonitor providerSideMonitor = new ConnectionMonitor(
                        scheduler);
                ConnectionThread providerSideThread = new ConnectionThread(
                        providerSideMonitor);
                scheduler.scheduleThread(providerSideThread);
                Selector connectionSelector = connectionsManager
                        .startListening2(providerSideThread);

                ConnectionMonitor customerSideMonitor = new ConnectionMonitor(
                        scheduler);
                ConnectionThread customerSideThread = new ConnectionThread(
                        customerSideMonitor);
                scheduler.scheduleThread(customerSideThread);

                // Manually register device URL, so the media manager knows
                // about the provider URN. In real use, this information is
                // spread via broadcast messages
                mediaManager.registerDevice(providerDeviceID.toString(), new DeviceCacheObject(mediaManager
                        .getDefaultUnicastURL(), providerDeviceID.toString(), "W2", "D2", new Selector("1")));

                connectionsManager.open(providerDeviceID,
                        connectionSelector, customerSideThread);

                // ist.palcom.base.System.instance().disableConstantGC();
                // ist.palcom.base.System.instance().setTraceAfter(0);

                // Wait for connection to be established at both sides
                Connection providerSideConnection = providerSideMonitor
                        .waitUntilConnected();
                Connection customerSideConnection = customerSideMonitor
                        .waitUntilConnected();

                System.out.println("Remote DeviceID at provider side: "
                        + providerSideConnection.getRemoteDeviceID());
                System.out.println("Remote DeviceID at customer side: "
                        + customerSideConnection.getRemoteDeviceID());

                providerSideConnection.addMessageReceiver(this);
                customerSideConnection.addMessageReceiver(this);

                byte[] data1 = "Data1".getBytes();
                Message message1 = new Message(data1, 0, data1.length);
                providerSideConnection.sendMessage(message1);
                UDPCommunicationTest2.waitForMessageAndPrintIt(this);

                byte[] data2 = "Data2".getBytes();
                Message message2 = new Message(data2, 0, data2.length);
                customerSideMonitor.connection.sendMessage(message2);
                UDPCommunicationTest2.waitForMessageAndPrintIt(this);

                providerSideMonitor.connection
                        .addDisconnectReceiver(providerSideThread);
                customerSideMonitor.connection
                        .addDisconnectReceiver(customerSideThread);

                customerSideMonitor.connection.close();

                // Wait for connection to be disconnected at both sides
                providerSideMonitor.waitUntilDisconnected();
                customerSideMonitor.waitUntilDisconnected();

                System.out.println("Connection disconnected");

                connectionsManager.stopListening(connectionSelector);

                communicationManager.stop();
                routingManager.stop();
            }
        };
        scheduler.scheduleThread(t);
        scheduler.run();
    }

    private class ConnectionThread extends PalcomPriorityThread {
        private ConnectionMonitor monitor;

        ConnectionThread(ConnectionMonitor monitor) {
            this.monitor = monitor;
        }

        public void run() {
            while (true) {
                ConnectionEvent event = (ConnectionEvent) waitEvent();
                if (event.isConnected()) {
                    monitor.reportConnected(event.getConnection());
                } else {
                    monitor.reportDisconnected(event.getConnection());
                }
            }
        }
    }

    private class ConnectionMonitor extends Monitor {

        private Connection connection;

        private Condition connected = new Condition();

        private Condition disconnected = new Condition();

        public ConnectionMonitor(PalcomScheduler scheduler) {
            super(scheduler);
        }

        void reportConnected(Connection c) {
            enterMonitor();
            try {
                System.out.println("Connected");
                connection = c;
                connected.causeAll();
            } finally {
                exitMonitor();
            }
        }

        void reportDisconnected(Connection c) {
            enterMonitor();
            try {
                System.out.println("Disconnected");
                connection = null;
                disconnected.causeAll();
            } finally {
                exitMonitor();
            }
        }

        Connection waitUntilConnected() {
            enterMonitor();
            try {
                while (connection == null) {
                    connected.await();
                }
                return connection;
            } finally {
                exitMonitor();
            }
        }

        void waitUntilDisconnected() {
            enterMonitor();
            try {
                while (connection != null) {
                    disconnected.await();
                }
            } finally {
                exitMonitor();
            }
        }
    }
}
