/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.discovery;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.communication.CommunicationManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.LocalServiceIDReply;
import ist.palcom.resource.descriptor.LocalServiceIDRequest;
import ist.palcom.resource.descriptor.Opt;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDConnectionList;
import ist.palcom.resource.descriptor.PRDConnectionListReply;
import ist.palcom.resource.descriptor.PRDConnectionListRequest;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDDeviceReply;
import ist.palcom.resource.descriptor.PRDDeviceRequest;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDServiceFMDescriptionReply;
import ist.palcom.resource.descriptor.PRDServiceFMDescriptionRequest;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.PRDServiceListItem;
import ist.palcom.resource.descriptor.PRDServiceListReply;
import ist.palcom.resource.descriptor.PRDServiceListRequest;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.Request;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.SelectorReply;
import ist.palcom.resource.descriptor.SelectorRequest;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceInstanceIDReply;
import ist.palcom.resource.descriptor.ServiceInstanceIDRequest;
import ist.palcom.resource.descriptor.ServiceStatusReply;
import ist.palcom.resource.descriptor.ServiceStatusRequest;
import ist.palcom.xml.XMLException;
import ist.palcom.xml.XMLFactory;
import ist.palcom.xml.XMLRepresentable;

import java.util.Iterator;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class DiscoveryManager implements IHgraphNode{

	protected HgraphNode _node;

	/** The pub/sub topic to use for discovery. */
	public static final String TOPIC = "discovery";

	public static final String DISCOVERY_VERSION = "D2";

	private PalcomScheduler scheduler;

	private PointToPointManager pointToPointManager;

	private PRDDevice localDeviceInfo;

	private CommunicationManager communicationManager;

	/** List with InfoNotices about the local services and connections. */
	private final LinkedList localInfos;

	/** List with SelectorNotices about the local selectors. */
	private final LinkedList selectors;

	/** List with ServiceInstanceIDNotices about the local serviceInstanceIDs. */
	private final LinkedList serviceInstanceIDs;

	/** List with services by localSIDs and their status. */
	private final LinkedList serviceStatusList;

	private DiscoveryManagerThread discoveryManagerThread;

	/** XMLFactory used for creating InfoEvents from received data. */
	private XMLFactory factory;

	private PRDConnectionList connectionInfoList;

	/** Cache number for connections, changes when a connection is added or removed. */
	private long connectionCache;

	/** Cache number for services, changes when a serviceList is added or removed. */
	private long serviceCache;

	/** Cache number for selectors, changes when a selector changes. */
	private long selectorCache;

	/** Cache number for service status, changes when the status of a service changes. */
	private long serviceStatusCache;

	/**
	 * Create a DiscoveryManager. localDeviceinfo is the info about the local
	 * device.
	 */
	public DiscoveryManager(PalcomScheduler scheduler,
			PointToPointManager pointToPointManager,
			CommunicationManager communicationManager, 
			PRDDevice localDeviceInfo) {

		this.scheduler = scheduler;
		this.pointToPointManager = pointToPointManager;
		this.communicationManager = communicationManager;
		this.localDeviceInfo = localDeviceInfo;
		communicationManager.setLocalDeviceID(localDeviceInfo.getDeviceID());
		communicationManager.setLocalDiscoveryVersion(DiscoveryManager.DISCOVERY_VERSION);

		localInfos = new LinkedList();
		selectors = new LinkedList();
		serviceInstanceIDs = new LinkedList();
		serviceStatusList = new LinkedList();
		discoveryManagerThread = new DiscoveryManagerThread();

		factory = new DiscoveryInfoXMLFactory();

		connectionInfoList = new PRDConnectionList();
		localInfos.insertLast(new InfoNotice(connectionInfoList));

		/******hgraph impl*******/
		_node = new HgraphNode("discovery manager");

		connectionCache = 0;
		serviceCache = 0;
		selectorCache = 0;
		serviceStatusCache = 0;
	}

	/**
	 * Start the DiscoveryManager. Listening for remote info requests will
	 * begin, and an initial announcement about this device will be sent out
	 * (the DeviceInfo sent out in an InfoEvent with keepinfo=true).
	 * 
	 * <p>
	 * The PubSubManager that was passed to the constructor must be started
	 * before this call.
	 */
	public void start() {
		scheduler.scheduleThread(discoveryManagerThread);

		Selector selector = pointToPointManager
		.startReceiving2(discoveryManagerThread);

		localDeviceInfo.setDiscoverySelector(selector);
		localDeviceInfo.setDeviceStatus((byte)-1);

		//TODO This value should be changed when the firmware of the device is changed.
		localDeviceInfo.setDeviceVersion("0.1");

		communicationManager.setDiscoverySelector(selector);
	}

	/**
	 * Stop the DiscoveryManager. A HeartAttack message will be
	 * sent out, so other devices can remove information
	 * about this device.
	 * 
	 * <p>
	 * The PubSubManager that was passed to the constructor will not be stopped.
	 */
	public void stop() {
		discoveryManagerThread.terminate();
		communicationManager.sendHeartAttack();

		pointToPointManager.stopReceiving(localDeviceInfo.getDiscoverySelector());

		localInfos.clear();

	}

	/**
	 * Add a local ServiceList to the DiscoveryManager. The list will be sent
	 * out in response to future requests from other devices.
	 */
	public void updateLocalServiceList(PRDServiceList list) {
		serviceCache++;
		updateLocalInfo(list);
		for(int i = 0; i < list.getNumPRDServiceListItem(); ++i) {
			if(list.getPRDServiceListItem(i) instanceof PRDService) {
				updateServiceStatus(list.getPRDServiceListItem(i).getLocalSID(), PRDData.FULLY_OPERATIONAL, "");
			}
		}
	}

	/**
	 * Remove a local ServiceList from the DiscoveryManager. The ServiceList
	 * will no longer be sent out in response to requests.
	 */
	public void removeLocalServiceList(PRDServiceList list) {
		serviceCache++;
		removeLocalInfo(list);
		for(int i = 0; i < list.getNumPRDServiceListItem(); ++i) {
			if(list.getPRDServiceListItem(i) instanceof PRDService) {
				removeServiceStatus(list.getPRDServiceListItem(i).getLocalSID());
			}
		}
	}

	/**
	 * Remove a local Service or SubList from the service list in the DiscoveryManager. The item
	 * will no longer be sent out in response to requests.
	 */
	public boolean removeLocalServiceOrSubList(PRDServiceListItem listItem) {
		boolean result = false;
		serviceCache++;
		if(listItem instanceof PRDService) {
			removeServiceStatus(listItem.getLocalSID());
		}
		InfoNotice notice = (InfoNotice) localInfos.getFirst();
		while (notice != null) {
			if(notice != null && notice.info instanceof PRDServiceList) {
				if (listItem instanceof PRDSubList && 
						notice.info.getLocalSID().equals(listItem.getLocalSID())) {
					removeLocalServiceList((PRDServiceList)notice.info);
				}
				PRDServiceList list = (PRDServiceList)notice.info;
				PRDServiceListItem tmpItem = list.getPRDServiceListItem(listItem.getLocalSID());
				if(tmpItem != null) {
					result = result | tmpItem.remove();
				}
				updateLocalServiceList(list);
			}
			notice = (InfoNotice) notice.getNext();
		}
		changeDeviceCache();
		return result;
	}

	/**
	 * Add a local ServiceDescription to the DiscoveryManager. The
	 * ServiceDescription will be sent out in response to future requests from
	 * other devices.
	 */
	public void updateLocalServiceDescription(PRDServiceFMDescription sd) {
		serviceCache++;
		updateLocalInfo(sd);
	}

	/**
	 * Remove a local ServiceDescription from the DiscoveryManager. The
	 * ServiceDescription will no longer be sent out in response to requests.
	 */
	public void removeLocalServiceDescription(PRDServiceFMDescription sd) {
		serviceCache++;
		removeLocalInfo(sd);
	}

	/**
	 * Add a local ConnectionInfo to the DiscoveryManager. The list will be sent
	 * out in response to future requests from other devices.
	 */
	public void updateLocalConnectionInfo(PRDConnection ci) {
		connectionInfoList.addPRDConnection(ci);
		connectionCache++;
		changeDeviceCache();
	}

	/**
	 * Remove a local ConnectionInfo from the DiscoveryManager. The
	 * ConnectionInfo will no longer be sent out in response to requests.
	 */
	public void removeLocalConnectionInfo(PRDConnection ci) {
		PRDConnectionList newList = new PRDConnectionList();
		for(int i = 0; i < connectionInfoList.getNumPRDConnection(); ++i) {
			PRDConnection cifl = connectionInfoList.getPRDConnection(i);
			if(!cifl.equals(ci)) {
				newList.addPRDConnection(cifl);
			} 
		}
		connectionInfoList = newList;
		connectionCache++;
		changeDeviceCache();
	}

	/**
	 * Add or change a local Selector to the DiscoveryManager. The Selector will be sent
	 * out in response to future requests from other devices.
	 */
	public void updateLocalSelector(LocalSID localSID, Selector selector) {
		SelectorNotice notice = getLocalSelectorNotice(localSID);
		if(notice == null) {
			selectors.insertLast(new SelectorNotice(localSID, selector));
		} else {
			notice.selector = selector;
			selectorCache++;
			changeDeviceCache();
		}
	}

	/**
	 * Remove a local Selector for Service with ID localSID from the DiscoveryManager. The Selector
	 * will no longer be sent out in response to requests.
	 */
	public void removeLocalSelector(LocalSID localSID) {
		SelectorNotice notice = getLocalSelectorNotice(localSID);
		if(notice == null) {
			selectors.remove(notice);
		}
	}

	/**
	 * Add or change a local ServiceInstanceID to the DiscoveryManager. The ServiceInstanceID will 
	 * be sent out in response to future requests from other devices.
	 */
	public void updateLocalServiceInstanceID(LocalSID localSID, ServiceInstanceID serviceInstanceID) {
		ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(localSID);
		if(notice == null) {
			serviceInstanceIDs.insertLast(new ServiceInstanceIDNotice(localSID, serviceInstanceID));
		} else {
			notice.instanceID = serviceInstanceID;
//			serviceCache++;
//			changeDeviceCache();
		}
	}

	/**
	 * Remove a local ServiceInstanceID for Service with ID localSID from the DiscoveryManager. 
	 * The ServiceInstanceID will no longer be sent out in response to requests.
	 */
	public void removeLocalServiceInstanceID(LocalSID localSID) {
		ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(localSID);
		if(notice == null) {
			serviceInstanceIDs.remove(notice);
		}
	}

	/**
	 * Add or change a LocalSID to the DiscoveryManager. The LocalSID will 
	 * be sent out in response to future requests from other devices.
	 */
	public void updateLocalSID(ServiceInstanceID serviceInstanceID, LocalSID localSID) {
		ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(serviceInstanceID);
		if(notice == null) {
			serviceInstanceIDs.insertLast(new ServiceInstanceIDNotice(localSID, serviceInstanceID));
		} else {
			notice.localSID = localSID;
//			serviceCache++;
//			changeDeviceCache();
		}
	}

	/**
	 * Remove a LocalSID for Service with ID ServiceInstanceID from the DiscoveryManager. 
	 * The LocalSID will no longer be sent out in response to requests.
	 */
	public void removeLocalSID(ServiceInstanceID serviceInstanceID) {
		ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(serviceInstanceID);
		if(notice == null) {
			serviceInstanceIDs.remove(notice);
		}
	}

	/**
	 * Add or change the status of a service. The status will be sent
	 * out in response to future requests from other devices.
	 */
	public void updateServiceStatus(LocalSID localSID, byte status, String helpText) {
		ServiceStatusNotice notice = getServiceStatusNotice(localSID);
		if(notice == null) {
			serviceStatusList.insertLast(new ServiceStatusNotice(localSID, status, helpText));
		} else {
			notice.status = status;
			notice.helpText = helpText;
			serviceStatusCache++;
			changeDeviceCache();
		}
	}

	/**
	 * 
	 */
	private void changeDeviceCache() {
		communicationManager.setLocalDeviceCache("" + 
				(Integer.parseInt(communicationManager.getLocalDeviceCache())+1));
	}

	/**
	 * Remove the status info for Service with ID localSID from the DiscoveryManager. The status
	 * will no longer be sent out in response to requests.
	 */
	public void removeServiceStatus(LocalSID localSID) {
		ServiceStatusNotice notice = getServiceStatusNotice(localSID);
		if(notice == null) {
			serviceStatusList.remove(notice);
		}
	}

	/**
	 * Add a local Info to the DiscoveryManager. The Info will be sent out in
	 * response to future info requests from other devices.
	 */
	private void updateLocalInfo(PRDData info) {
		InfoNotice notice = getLocalInfoNotice(info);
		if (notice == null) {
			localInfos.insertLast(new InfoNotice(info));
		} else {
			notice.info = info;
		}
		changeDeviceCache();
	}

	/**
	 * Remove a local Info from the DiscoveryManager. The Info will no longer be
	 * sent out in response to info requests.
	 */
	private void removeLocalInfo(PRDData info) {
		InfoNotice notice = getLocalInfoNotice(info);
		if (notice != null) {
			localInfos.remove(notice);
		}
		changeDeviceCache();
	}

	/** @return The InfoNotice for the local info info, or null. */
	private InfoNotice getLocalInfoNotice(PRDData info) {
		InfoNotice notice = (InfoNotice) localInfos.getFirst();
		while (notice != null && !info.isAboutTheSameEntityAs(notice.info)) {
			notice = (InfoNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The SelectorNotice for the localSID, or null. */
	private SelectorNotice getLocalSelectorNotice(LocalSID localSID) {
		SelectorNotice notice = (SelectorNotice) selectors.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (SelectorNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The ServiceInstanceIDNotice for the localSID, or null. */
	private ServiceInstanceIDNotice getLocalServiceInstanceIDNotice(LocalSID localSID) {
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) serviceInstanceIDs.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The ServiceInstanceIDNotice for the ServiceInstanceID, or null. */
	private ServiceInstanceIDNotice getLocalServiceInstanceIDNotice(ServiceInstanceID serviceInstanceID) {
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) serviceInstanceIDs.getFirst();
		while(notice != null && !notice.instanceID.equals(serviceInstanceID)) {
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The ServiceStatusNotice for the localSID, or null. */
	private ServiceStatusNotice getServiceStatusNotice(LocalSID localSID) {
		ServiceStatusNotice notice = (ServiceStatusNotice) serviceStatusList.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (ServiceStatusNotice) notice.getNext();
		}
		return notice;
	}

	private void sendInfo(byte[] data, String remoteDeviceID,
			Selector remoteSelector, boolean useReliable) {
		Selector localSelector = localDeviceInfo.getDiscoverySelector();
		byte[] messageData = data;
		Message infoMessage = new Message(messageData);
		infoMessage.setSenderSelector(localSelector);
		boolean sent = pointToPointManager.send(infoMessage,
				localSelector, new DeviceID(remoteDeviceID), remoteSelector, useReliable);
	}

	private class DiscoveryManagerThread extends PalcomPriorityThread {

		public void run() {
			while (true) {
				Event event = waitEvent();

				if (event instanceof MessageEvent) {
					MessageEvent messageEvent = (MessageEvent) event;
					Message message = (Message) messageEvent
					.getMessage();
					boolean useReliable = false;
					MessageHeaderNode hNode = (MessageHeaderNode) 
					message.getHeaderNodes().getFirst();
					// Find header node with function reliable
					while (hNode != null
							&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_RELIABLE) {
						hNode = (MessageHeaderNode) hNode.getNext();
					}
					//If found, switch to reliable communication
					if (hNode != null) {
						useReliable = true;
					}
					try {
						XMLRepresentable representable = factory
						.createFromXMLData(message.getData(), 0,
								message.getDataLength());

						if (representable instanceof Request) {
							if(representable instanceof PRDDeviceRequest) {
								PRDDevice deviceInfo = new PRDDevice(null, localDeviceInfo.getName());
								deviceInfo.setRemoteConnectSelector(localDeviceInfo.getRemoteConnectSelector());
								deviceInfo.setDeviceStatus((byte)-1);
								deviceInfo.setDeviceVersion(localDeviceInfo.getDeviceVersion());
								sendInfo(new PRDDeviceReply("" + serviceCache, "" + connectionCache , "1" , 
										"" + selectorCache, "" + serviceStatusCache, deviceInfo).getXMLData(), 
										message.getSenderDeviceID(), message.getSenderSelector(), useReliable);
							} else if(representable instanceof PRDConnectionListRequest) {
								sendInfo(new PRDConnectionListReply(connectionInfoList).getXMLData(),
										message.getSenderDeviceID(), message
										.getSenderSelector(), useReliable);
							} else if(representable instanceof SelectorRequest) {
								SelectorRequest request = (SelectorRequest) representable;
								request.getLocalSID().setDeviceID(localDeviceInfo.getDeviceID());
								SelectorNotice notice = getLocalSelectorNotice(request.getLocalSID());
								if(notice != null) {
									sendInfo(new SelectorReply(notice.localSID, notice.selector)
									.getXMLData(), message.getSenderDeviceID(), message
									.getSenderSelector(), useReliable);
								}
							} else if(representable instanceof ServiceInstanceIDRequest) {
								ServiceInstanceIDRequest request = (ServiceInstanceIDRequest) representable;
								request.getLocalSID().setDeviceID(localDeviceInfo.getDeviceID());
								ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(request.getLocalSID());

								if(notice != null) {
									ServiceInstanceID siid = new ServiceInstanceID(new Opt(), notice.instanceID.getServiceID(),
											notice.instanceID.getInstanceNumber());
									sendInfo(new ServiceInstanceIDReply(notice.localSID, siid)
									.getXMLData(), message.getSenderDeviceID(), message
									.getSenderSelector(), useReliable);
								}
							} else if(representable instanceof LocalServiceIDRequest) {
								LocalServiceIDRequest request = (LocalServiceIDRequest) representable;
								request.getServiceInstanceID().setDeviceID(localDeviceInfo.getDeviceID());
								ServiceInstanceIDNotice notice = getLocalServiceInstanceIDNotice(request.getServiceInstanceID());
								if(notice != null) {
									ServiceInstanceID siid = new ServiceInstanceID(new Opt(), notice.instanceID.getServiceID(),
											notice.instanceID.getInstanceNumber());
									sendInfo(new LocalServiceIDReply(siid, notice.localSID)
									.getXMLData(), message.getSenderDeviceID(), message
									.getSenderSelector(), useReliable);
								}
							} else if(representable instanceof ServiceStatusRequest) {
								ServiceStatusRequest request = (ServiceStatusRequest) representable;
								request.getLocalSID().setDeviceID(localDeviceInfo.getDeviceID());
								ServiceStatusNotice notice = getServiceStatusNotice(request.getLocalSID());
								if(notice != null) {
									sendInfo(new ServiceStatusReply(notice.localSID, notice.status, notice.helpText)
									.getXMLData(), message.getSenderDeviceID(), message
									.getSenderSelector(), useReliable);
								}
							} else {
								InfoNotice notice = (InfoNotice) localInfos
								.getFirst();
								while (notice != null) {
									PRDData info = notice.info;
									if (representable instanceof PRDServiceListRequest
											&& info instanceof PRDServiceList) {
										((PRDServiceListRequest)representable).getParentLocalSID().setDeviceID(localDeviceInfo.getDeviceID());
										if(((PRDServiceListRequest)representable).getParentLocalSID().equals(info.getLocalSID())) {
											PRDServiceList list = (PRDServiceList)info;
											sendInfo(new PRDServiceListReply(
													"1", list).getXMLData(),
													message.getSenderDeviceID(), message
													.getSenderSelector(), useReliable);
										}
									}
									if (representable instanceof PRDServiceFMDescriptionRequest &&
											info instanceof PRDServiceFMDescription
											&& (((PRDServiceFMDescriptionRequest) representable)
													.getLocalSID().getServicePartID().equals(
															((PRDServiceFMDescription)info).getLocalSID().getServicePartID()))) {
										sendInfo(new PRDServiceFMDescriptionReply(
												(PRDServiceFMDescription)info).getXMLData(),
												message.getSenderDeviceID(), message
												.getSenderSelector(), useReliable);
									}
									notice = (InfoNotice) notice.getNext();
								}
							}
						} 
					} catch (XMLException e) {
//						e.printStackTrace();
//						throw new IllegalArgumentException(
//						"Error in discovery XML data: " + e);
//						TODO An event should be sent to upper layers announcing this information
						System.out.println("DiscoveryManager: Device " + message.getSenderDeviceID() + 
						" is not using a compatible discovery version.");
					}
				}
			}
		}
	}

	private class InfoNotice extends Link {

		PRDData info;

		InfoNotice(PRDData info) {
			this.info = info;
		}
	}

	private class SelectorNotice extends Link {
		LocalSID localSID;

		Selector selector;

		SelectorNotice(LocalSID localSID, Selector selector) {
			this.localSID = localSID;
			this.selector = selector;
		}
	}

	private class ServiceInstanceIDNotice extends Link {
		LocalSID localSID;

		ServiceInstanceID instanceID;

		ServiceInstanceIDNotice(LocalSID localSID, ServiceInstanceID instanceID) {
			this.localSID = localSID;
			this.instanceID = instanceID;
		}
	}

	private class ServiceStatusNotice extends Link {
		LocalSID localSID;

		byte status;

		String helpText;

		ServiceStatusNotice(LocalSID localSID, byte serviceStatus, String statusHelpText) {
			this.localSID = localSID;
			status = serviceStatus;
			helpText = statusHelpText;
		}
	}

	/****************hgraph impl below********************/

	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}

}
