/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.discovery.test;

import ist.palcom.communication.CommunicationManager;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.discovery.InfoEvent;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.udp.UDPMediaManager;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.Monitor;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.manager.ResourceManager;
import ist.palcom.routing.BasicRoutingManager;
import ist.palcom.routing.RoutingManager;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class DiscoveryTest {

    public static void main(String[] args) {
        DiscoveryTest test = new DiscoveryTest();

        test.testDiscovery();
        System.out.println();
    }

    private void testDiscovery() {
        System.out.println("--- testDiscovery ---");

        final PalcomScheduler scheduler = new PalcomScheduler();
        PalcomThread t = new PalcomPriorityThread() {
            public void run() {
            	DeviceID deviceID = new DeviceID("X:mydevice");
                final PRDDevice deviceInfo = new PRDDevice(deviceID);
                deviceInfo.setRemoteConnectSelector(new Selector("1034"));
                deviceInfo.setName("My device");

                MediaManager mediaManager = new UDPMediaManager(scheduler, null);
                mediaManager.setPeriod(2000);
                RoutingManager routingManager = new BasicRoutingManager(mediaManager, scheduler);
                CommunicationManager communicationManager = new CommunicationManager(
                		scheduler, routingManager);
                PointToPointManager pointToPointManager = new PointToPointManager(
                        communicationManager);
                final DiscoveryManager discoveryManager = new DiscoveryManager(
                        scheduler, pointToPointManager,
                        communicationManager, deviceInfo);
                final ResourceManager resourceManager = new ResourceManager(
                        scheduler, pointToPointManager,
                        communicationManager, deviceInfo);
                
                PRDServiceList serviceList = new PRDServiceList(new LocalSID(deviceInfo.getDeviceID(), null));
                
                PRDService serviceInfo = new PRDService("Image service", 
                        new LocalSID(deviceID, "images"), PRDService.UNICAST, false, false, "P1", false, null, null);
                
                communicationManager.setLocalDeviceID(deviceID);
                
                serviceList.addPRDService(serviceInfo);
                discoveryManager.updateLocalServiceList(serviceList);
                
                discoveryManager.start();
                resourceManager.start();

                final InfoMonitor infoMonitor = new InfoMonitor(scheduler);
                PalcomThread infoThread = new InfoThread(resourceManager,
                        infoMonitor);
                
                resourceManager.updatePRDDeviceReceiver(infoThread);
                scheduler.scheduleThread(infoThread);
                
                PRDDevice receivedDeviceInfo = infoMonitor
                        .waitUntilDeviceInfo();
                System.out.println("Device discovered");
                System.out.println("  Device ID: "
                        + receivedDeviceInfo.getDeviceID());
                System.out.println("  Device name: "
                        + receivedDeviceInfo.getName());
                PRDServiceList receivedServiceList = infoMonitor
                        .waitUntilServiceList();
                System.out.println("Got service list");
                System.out.println("  Device ID: " + receivedServiceList.getParentLocalSID().getDeviceID());
                System.out.println("  Parent LocalSID: "
                        + receivedServiceList.getParentLocalSID().toString());
                System.out.println("  Services:");
                for (int i = 0; i < receivedServiceList.getNumPRDServiceListItem(); i++) {
                    PRDService si = (PRDService)receivedServiceList.getPRDServiceListItem(i);
                    System.out.println("    " + si.getName() + ", "
                            + receivedServiceList.getParentLocalSID().getDeviceID() + ", " + si.getLocalSID().toString());
                }
                
                serviceInfo.setName("Image service extra");
                PRDServiceList list = new PRDServiceList(new LocalSID(deviceInfo.getDeviceID(), null));
                list.addPRDService(serviceInfo);
                discoveryManager.updateLocalServiceList(list);
                
                receivedServiceList = infoMonitor
                .waitUntilServiceList();
                System.out.println("Got service list");
                System.out.println("  Device ID: " + receivedServiceList.getParentLocalSID().getDeviceID());
                System.out.println("  Parent LocalSID: "
                		+ receivedServiceList.getParentLocalSID().toString());
                System.out.println("  Services:");
                for (int i = 0; i < receivedServiceList.getNumPRDServiceListItem(); i++) {
                	PRDService si = (PRDService)receivedServiceList.getPRDServiceListItem(i);
                	System.out.println("    " + si.getName() + ", "
                			 + receivedServiceList.getParentLocalSID().getDeviceID() + ", " + si.getLocalSID().toString());
                }
                

                infoThread.terminate();
                discoveryManager.stop();
                resourceManager.stop();
                communicationManager.stop();
                routingManager.stop();
            }
        };
        scheduler.scheduleThread(t);
        scheduler.run();
    }

    private class InfoMonitor extends Monitor {

        private PRDDevice deviceInfo = null;

        private PRDServiceList serviceList = null;

        private Condition received = new Condition();

        public InfoMonitor(PalcomScheduler scheduler) {
            super(scheduler);
        }

        void infoReceived(InfoEvent event) {
            enterMonitor();
            try {
                if (event.getInfo() instanceof PRDDevice) {
                    deviceInfo = (PRDDevice) event.getInfo();
                } else if (event.getInfo() instanceof PRDServiceList) {
                    serviceList = (PRDServiceList) event.getInfo();
                }
                received.causeAll();
            } finally {
                exitMonitor();
            }
        }

        PRDDevice waitUntilDeviceInfo() {
            enterMonitor();
            try {
                while (deviceInfo == null) {
                    received.await();
                }
                return deviceInfo;
            } finally {
                exitMonitor();
            }
        }

        PRDServiceList waitUntilServiceList() {
            enterMonitor();
            PRDServiceList tempList = null;
            try {
                while (serviceList == null) {
                    received.await();
                }
                tempList = serviceList;
                serviceList = null;
                return tempList;
            } finally {
                exitMonitor();
            }
        }
    }

    private class InfoThread extends PalcomPriorityThread {

        private ResourceManager resourceManager;

        private InfoMonitor infoMonitor;

        InfoThread(ResourceManager resourceManager, InfoMonitor infoMonitor) {
            this.resourceManager = resourceManager;
            this.infoMonitor = infoMonitor;
        }

        public void run() {
            while (true) {
                InfoEvent event = (InfoEvent) waitEvent();
                if(event.getInfo() instanceof PRDDevice) {
                	DeviceID deviceID = ((PRDDevice)event.getInfo()).getDeviceID();
                    resourceManager.updatePRDServiceListReceiver(new LocalSID(deviceID, null), this);
                }
                infoMonitor.infoReceived(event);
            }
        }
    };
}
