/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import java.util.Iterator;

import ist.palcom.base.HashMap;
import ist.palcom.base.StringBuilder;

/**
 * The default implementation of H-graph Nodes as described by the IHgraphNode interface
 */
public class HgraphNode implements IHgraphNode{
	
	private IHgraphNode parent;
	private String name;
	private Path path;
	
	private HashMap children;
	
	private PrintHandler printHandler;
	private NodeHandler nodeHandler;
	private ValueHandler valueHandler;
	private FuncHandler funcHandler;
	private GenericInvokator genericInvokator;

	private SiblingFunc sf;
	private ValueFunc vf;
	private FuncFunc ff;
	

	/**
	 * Constructs an anonymous HgraphNode. It uses Object.toString() as name and prefixes it with "ANONYMOUS:"
	 */
	public HgraphNode(){
		this.name = "ANONYMOUS:" + super.toString();
		init(name);
	}
	
	/**
	 * Constructs a named HgraphNode
	 * @param name The name of the node
	 */
	public HgraphNode(String name){
		this.name = name;
		init(name);
	}
	
	public IHgraphNode getNodeParent(){
		return parent;
	}
	
	private void init(String name){
		children = new HashMap();
		sf = new SiblingFunc(this);
		vf = new ValueFunc(this);
		ff = new FuncFunc(this);
		printHandler = new PrintHandler(this);
		ff.addHandler(printHandler, null);
		nodeHandler = new NodeHandler(this);
		ff.addHandler(nodeHandler, null);
		valueHandler = new ValueHandler(this);
		ff.addHandler(valueHandler, null);
		funcHandler = new FuncHandler(this);
		ff.addHandler(funcHandler, null);
		path = new Path(name);
		genericInvokator = new GenericInvokator(this);
		ff.addInvokator(genericInvokator, null);
	}
	
	public SiblingFunc getSiblingFunc(){
		return sf;
	}
	public ValueFunc getValueFunc(){
		return vf;
	}
	public FuncFunc getFuncFunc(){
		return ff;
	}

	public String getNodeName(){
		return name;
	}
	
	public int setNodeName(String name, Path callerPath){
		if((parent != null) && parent.hasChildNamed(name)){
			return CHILD_NAME_CLASH;
		}
		String oldName = this.name;
		this.name = name;
		if(parent != null){
			parent.getSiblingFunc().childSatName(oldName, name);
		}
		updatePath();
		ff.trigger(NodeHandler.CMD_SETNAME, new Value("name", name), callerPath);
		return STATUS_OK;
	}

	public void updatePath(){
		if(parent != null){
			path = new Path(name, parent.getPath().getPathArray());
		}else{
			path = new Path(name);
		}
		Iterator iter = children.values().iterator();
		while(iter.hasNext()){
			IHgraphNode n = (IHgraphNode)iter.next();
			n.updatePath();
		}
	}
	
	public Path getPath(){
		return path;
	}
	
	public int setNodeParent(IHgraphNode newParent, Path callerPath){
		if(this.isAncestorOf(newParent)){
			return CIRCULARITY_CLASH;
		}
		if((parent != null) && (parent.getFuncFunc() != newParent.getFuncFunc())){ //totally crazy hack because wrapping auto generated code SUCKS!
			parent.removeNodeChild(name, callerPath);			
		}
		parent = newParent;
		//path = new Path(name, parent.getPath().getPathArray(), ff.getSignatures(), this);
		updatePath();
		return STATUS_OK;
	}

	public boolean hasChildNamed(String name){
		return children.containsKey(name);
	}
		
	public int addNodeChild(IHgraphNode node, Path callerPath){
		if(children == null){
			children = new HashMap();
			children.put(node.getNodeName(), node);
		}else{
			if(children.containsKey(node.getNodeName())){
				return CHILD_NAME_CLASH;				
			}
			if(isAncestorOf(node)){
				return CIRCULARITY_CLASH;
			}
			children.put(node.getNodeName(), node);
		}
		node.setNodeParent(this, callerPath);
		sf.updateChildSiblingReferences(node.getNodeName(), node);
		ff.trigger(NodeHandler.CMD_ADDCHILD, new Value("node", node), callerPath);
		return STATUS_OK;
	}
	
	public boolean isAncestorOf(IHgraphNode node){
		if(this.equals(node)){
			return true;
		}
		if(children == null){
			return false;
		}
		Iterator iter = children.values().iterator();
		while (iter.hasNext()){
			IHgraphNode n = (IHgraphNode)iter.next();
			if(n.isAncestorOf(node)){
				return true;
			}
		}		
		return false;
	}
			
	public int removeNodeChild(String name, Path callerPath){
		if(children == null){
			return STATUS_OK;
		}
		if(((IHgraphNode)children.get(name)) != null){
			if((((IHgraphNode)children.get(name)).getPath() != null) && (getPath() != null)){
				//System.out.println("Remo: "+((IHgraphNode)children.get(name)).getPath() + " from "+getPath());				
			}
		}
		children.remove(name);
		//TODO: siblingrefs
		//TODO: TRIGGER
		return STATUS_OK;
	}
	
	public IHgraphNode getChildNamed(String name){
		return (IHgraphNode)children.get(name);
	}

	public String toXML(){
		return XML.toXML(this);
	}
	
	public String toXMLFull(){
		return XML.toXMLFull(this);
	}
	
	public String toXMLRecursively(){
		return XML.toXMLRecursively(this);		
	}

	public String toXMLFullRecursively(){
		return XML.toXMLFullRecursively(this);
	}	
		
	public Iterator getChildIterator(){
		return children.values().iterator();
	}
}
