/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import java.util.Iterator;

/**
 * The IHgraphNode is the key identity in relation to Hgraphs. For more information, see the package specification
 */
public interface IHgraphNode {

	public static final int ERROR = -1;
	public static final int STATUS_OK = 0;
	public static final int CHILD_NAME_CLASH = 1;
	public static final int NODE_NOT_CHILD = 3;
	public static final int CIRCULARITY_CLASH = 4;
	public static final int COMMAND_OR_NODE_MISSING = 11;
	public static final int VALUE_NOT_PRESENT = 5;
	public static final int VALUE_ALREADY_EXISTS = 6;
	public static final int SIBLING_NONEXISTENT = 7;
	public static final int SIBLING_ALREADY_EXISTS = 8;
	public static final int SIBLING_REFERENCE_ALREADY_EXISTS = 9;
	public static final int SIBLING_REFERENCE_NONEXISTENT = 10;
	public static final int PATH_MISSING = 12;
	public static final int LISTENER_MISSING = 13;
	public static final int HANDLER_MISSING = 14;
	public static final int INVOKATOR_MISSING = 15;
	
	/**
	 * @return The name of the node.
	 */
	public String getNodeName();
	
	/**
	 * Sets the name of the node. If a sibling already has the name, an error control message is returned
	 * @param name The new name
	 * @return IHgraphNode.STATUS_OK if the change succeeded. IHgraphNode.CHILD_NAME_CLASH if a sibling with the given name already exists.
	 */
	public int setNodeName(String name, Path callerPath);

	/**
	 * @return the Path object pertaining to the path of the node
	 */
	public Path getPath();

	/**
	 * Updates the path of the H-Graph node along with its offspring recursively.
	 * The path is updated when the node name and the node parent is sat.
	 */
	public void updatePath();
	
	/**
	 * @return The parent of the node. null if the parent is root
	 */
	public IHgraphNode getNodeParent();
	
	/**
	 * Sets the parent of the HgraphNode. This triggers a recalculation of the path of the node and its offspring
	 * @param newParent the new parent
	 * @return IHgraphNode.STATUS_OK if the operation succeeds. IHgraphNode.CIRCULARITY_CLASH if the operation causes a circularity.
	 */
	public int setNodeParent(IHgraphNode newParent, Path callerPath);
	
	/**
	 * @return An Iterator over the children of the node
	 */
	public Iterator getChildIterator();
	
	/**
	 * Adds a child to the node's private list of children.
	 * Ensures that names amongst children are unique
	 * @param node The new child to be added
	 * @return IHgraphNode.STATUS_OK if the change succeeded. IHgraphNode.CHILD_NAME_CLASH if a sibling with the given name already exists.
	 */
	public int addNodeChild(IHgraphNode node, Path callerPath);
	
	/**
	 * Removes a child based on its name.
	 * @param name The name of the child to be removed
	 * @return IHgraphNode.STATUS_OK if the operation succeeds
	 */
	public int removeNodeChild(String name, Path callerPath);

	/**
	 * Checks whether the node has a child with a given name
	 * @param name The name to check
	 * @return True iff the node has an actual child with the given name. False if not.
	 */
	public boolean hasChildNamed(String name);
	
	/**
	 * Checks whether this node is an ancestor of the parameter node. 
	 * It does so with a depth-first sharch of its offspring (children etc.)
	 * @return true if this node is an ancestor of the parameter node. false if not.
	 */
	public boolean isAncestorOf(IHgraphNode node);

	/**
	 * @return The SiblingFunc Sibling functionality object for the given node.
	 */
	public SiblingFunc getSiblingFunc();	
	/**
	 * @return The ValueFunc Value functionality object for the given node.
	 */
	public ValueFunc getValueFunc();
	/**
	 * @return The FuncFunc functional functionality (ie. handlers, invokators and listeners) object for the given node.
	 */
	public FuncFunc getFuncFunc();
	
	/**
	 * @return A simple xml'ified String of the node (type, name, path, children)
	 */
	public String toXML();

	/**
	 * @return A complete xml'ified String of the node (same as simple + handlers, invokators, listeners, values)
	 */
	public String toXMLFull();
	
	/**
	 * @return A simple recursive xml'ified String of the node (type, name, path, children)
	 */
	public String toXMLRecursively();

	/**
	 * @return A complete recursive xml'ified String of the node (same as simple + handlers, invokators, listeners, values)
	 * HEAVY!!!
	 */
	public String toXMLFullRecursively();
	
}
