/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import ist.palcom.base.StringBuilder;

/**
 * The Path class is for representing paths in an Hgraph.
 * These are represented as delimited text identifiers.
 * Paths are also used for manifesting functional and structural dependencies in the hgraph
 */
public class Path {

	/**
	 * The delimiter string for paths
	 */
	public static final String PATH_DELIMITER = "#";
	
	/**
	 * The internal representation of a path is a string array
	 */
	private String[] path;
	
	/**
	 * A stringified version of the path
	 */
	private String stringPath;
	
	/**
	 * Creates a path of length 1
	 * @param name The sole identifier in the path
	 */
	public Path(String name){
		path = new String[1];
		path[0] = name;
	}

	/**
	 * Creates a child path by copying the supplied super path and appending a name identifier
	 * @param name The name of the node to which this path belongs
	 * @param superPath The path of the parent
	 */
	public Path(String name, String[] superPath){
		path = new String[superPath.length + 1];
		System.arraycopy(superPath, 0, path, 0, superPath.length);
		path[superPath.length] = name;
	}

	/**
	 * Creates a path from an already existing string array
	 * @param path
	 */
	public Path(String[] path){
		this.path = path;
		stringPath = toString();
	}
	
	/**
	 * @return The internal representation String array
	 */
	public String[] getPathArray(){
		return path;
	}
	
	/**
	 * Returns a stringified version of the path
	 */
	public String toString(){
		StringBuilder build = StringBuilder.instantiate();
		build.append(PATH_DELIMITER);
		for(int i = 0; i<path.length; i++){
			build.append(path[i]);
			build.append(PATH_DELIMITER);
		}
		return build.toString();
	}
	
	/**
	 * @return the length of the path
	 */
	public int getLength(){
		return path.length;
	}
	
	/**
	 * Compares the parameter path with this path
	 * @param eqPath 
	 * @return True if they are equally long and all entries are equal. False if not
	 */
	public boolean equals(Path eqPath){
		if(eqPath.getLength() == path.length){
			for(int i = 0; i<path.length; i++){
				if(!path[i].equals(eqPath.getPathArray()[i])){
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	/**
	 * @param path a string which supposedly is a path
	 * @return true if the parameter string matches the stringified version of this path
	 */
	public boolean equals(String path){
		return stringPath.equals(path);
	}
	
	/**
	 * Creates a Path object from a string
	 * @return a Path based on the string parameter
	 */
	public static Path fromString(String stringPath){
		String sPath = stringPath;
		if(stringPath.startsWith(Path.PATH_DELIMITER)){
			sPath = stringPath.substring(Path.PATH_DELIMITER.length());
		}
		return new Path(sPath.split(PATH_DELIMITER));
	}
	
	/**
	 * Creates a child path from a string and an identifier string
	 * @param name 
	 * @param parentStringPath
	 * @return a Path based on the parameters
	 */
	public static Path fromString(String name, String parentStringPath){
		String sPath = parentStringPath;
		if(parentStringPath.startsWith(Path.PATH_DELIMITER)){
			sPath = parentStringPath.substring(Path.PATH_DELIMITER.length());
		}
		return new Path(name, sPath.split(PATH_DELIMITER));
	}
}
