/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import java.util.Iterator;

import ist.palcom.base.StringBuilder;

/**
 * Sibling functionality for H-Graph nodes.
 * This class provides functionality for holding and maintaining a set of Sibling References.
 */
public class SiblingFunc {

	public static final String SIBLINGREFERENCES_TAG = "null.siblingreferences";
	public static final String CHILDSIBLINGREFERENCES_TAG = "null.childsiblingreferences";
	
	private SiblingReference[] siblingReferences;
	private ChildSiblingReference[] childSiblingReferenceMap;
	
	private IHgraphNode node;
	
	public SiblingFunc(IHgraphNode node){
		this.node = node;
	}

	/**
	 * Called when a child name is sat. 
	 * The method traverses the children referring to the node and updates their naming of it
	 */
	public int childSatName(String oldName, String newName){
		if(childSiblingReferenceMap != null){
			for(int i = 0; i < childSiblingReferenceMap.length; i++){
				if(childSiblingReferenceMap[i].getReferer().equals(oldName)){
					childSiblingReferenceMap[i].setReferer(newName);
				}
				if(childSiblingReferenceMap[i].getReferee().equals(oldName)){
					if(node.hasChildNamed(childSiblingReferenceMap[i].getReferer())){
						Iterator iter = node.getChildIterator();
						while(iter.hasNext()){
							IHgraphNode nn = (IHgraphNode)iter.next();
							if(nn.getNodeName().equals(childSiblingReferenceMap[i].getReferer())){
								nn.getSiblingFunc().changeGlobalName(oldName, newName);
							}
						}
					}
					childSiblingReferenceMap[i].setReferee(newName);
				}
			}			
		}		
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * Called when the node with some given global name changes
	 * @param name The global name of the child whose reference is changed
	 * @param ref The new reference of the child
	 */
	public void updateChildSiblingReferences(String name, IHgraphNode ref){
		if(childSiblingReferenceMap != null){
			if(ref == null){
				recursiveRemoveReferences(name, 0);
			}
			for(int i = 0; i < childSiblingReferenceMap.length; i++){
				if(childSiblingReferenceMap[i].getReferee().equals(name)){
					if(node.hasChildNamed(childSiblingReferenceMap[i].getReferer())){
						Iterator iter = node.getChildIterator();
						while(iter.hasNext()){
							IHgraphNode nn = (IHgraphNode)iter.next();
							if(nn.getNodeName().equals(childSiblingReferenceMap[i].getReferer())){
								nn.getSiblingFunc().changeReference(name, ref);
							}
						}
					}
				}
			}
		}
	}
		
	/**
	 * Creates a new child sibling reference and adds it to the local map.
	 * These are the primary means for maintaining the sibling references of the children
	 * @param referer
	 * @param referee
	 * @return an int indicating the status of the method
	 */
	public int addChildSiblingReference(String referer, String referee){
		if(childSiblingReferenceMap == null){
			childSiblingReferenceMap = new ChildSiblingReference[1];
			childSiblingReferenceMap[0] = new ChildSiblingReference(referer, referee);
			return IHgraphNode.STATUS_OK;
		}
		for(int i = 0; i<childSiblingReferenceMap.length; i++){
			if(childSiblingReferenceMap[i].getReferer().equals(referer) && childSiblingReferenceMap[i].getReferee().equals(referee)){
				return IHgraphNode.SIBLING_REFERENCE_ALREADY_EXISTS;
			}
		}
		ChildSiblingReference[] tmp = new ChildSiblingReference[childSiblingReferenceMap.length + 1];
		System.arraycopy(childSiblingReferenceMap, 0, tmp, 0, childSiblingReferenceMap.length);
		tmp[childSiblingReferenceMap.length] = new ChildSiblingReference(referer, referee);
		childSiblingReferenceMap = tmp;
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * Removes a child sibling reference from the local map
	 * @param referer
	 * @param referee
	 * @return a status message
	 */
	private int removeChildSiblingReference(String referer, String referee){
		if(childSiblingReferenceMap ==  null){
			return IHgraphNode.SIBLING_REFERENCE_NONEXISTENT;
		}
		for(int i = 0; i<childSiblingReferenceMap.length; i++){
			if(childSiblingReferenceMap[i].getReferer().equals(referer) && childSiblingReferenceMap[i].getReferee().equals(referee)){
				if(childSiblingReferenceMap.length == 1){
					childSiblingReferenceMap = null;
					return IHgraphNode.STATUS_OK;
				}
				ChildSiblingReference[] tmp = new ChildSiblingReference[childSiblingReferenceMap.length - 1];
				System.arraycopy(childSiblingReferenceMap, 0, tmp, 0, i);
				System.arraycopy(childSiblingReferenceMap, i+1, tmp, i, childSiblingReferenceMap.length - i - 1);
				childSiblingReferenceMap = tmp;
				return IHgraphNode.STATUS_OK;				
			}
		}		
		return IHgraphNode.SIBLING_REFERENCE_NONEXISTENT;		
	}
	
	/**
	 * Adds a sibling reference to the local sibling reference map.
	 * These are maintained across both global name and reference changes
	 * @see SiblingReference
	 * @return a status message
	 */
	public int addSiblingReference(SiblingReference ref){
		if(siblingReferences == null){
			siblingReferences = new SiblingReference[1];
			siblingReferences[0] = ref;
		}else{
			if(ref.hasLocalName()){
				for(int i = 0; i<siblingReferences.length; i++){
					if(siblingReferences[i].hasLocalName() && siblingReferences[i].getLocalName().equals(ref.getLocalName())){
						return IHgraphNode.SIBLING_ALREADY_EXISTS;
					}
				}
			}
			SiblingReference[] tmp = new SiblingReference[siblingReferences.length + 1];
			System.arraycopy(siblingReferences, 0, tmp, 0, siblingReferences.length);
			tmp[siblingReferences.length] = ref;
			siblingReferences = tmp;
		}
		node.getNodeParent().getSiblingFunc().addChildSiblingReference(node.getNodeName(), ref.getGlobalName());
		return IHgraphNode.STATUS_OK;
	}

	/**
	 * Removes a sibling reference based on its name.
	 * This name is firstly assumed to be a local name, but if none is found, 
	 * the method commences searching assuming it to be a global name
	 * @param name
	 * @return a status message
	 */
	public int removeSiblingReference(String name){
		return removeSiblingReference(getSiblingReferenceObject(name));
	}
	
	/**
	 * Removes a specific sibling reference
	 * @param ref
	 * @return a status message
	 */
	public int removeSiblingReference(SiblingReference ref){
		if((siblingReferences == null) || (ref == null)){
			return IHgraphNode.SIBLING_NONEXISTENT;
		}
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].equals(ref)){
				SiblingReference[] tmp = new SiblingReference[siblingReferences.length - 1];
				System.arraycopy(siblingReferences, 0, tmp, 0, i);
				System.arraycopy(siblingReferences, i + 1, tmp, i, siblingReferences.length - i - 1);					
				siblingReferences = tmp;
				for(int j = 0; j < siblingReferences.length; j++){
					if(siblingReferences[j].getGlobalName().equals(ref.getGlobalName())){
						return IHgraphNode.STATUS_OK;						
					}
				}
				node.getNodeParent().getSiblingFunc().removeChildSiblingReference(node.getNodeName(), ref.getGlobalName());
				return IHgraphNode.STATUS_OK;
			}
		}
		return IHgraphNode.SIBLING_NONEXISTENT;
	}
	
	/**
	 * @return A sibling reference object with some given name
	 */
	private SiblingReference getSiblingReferenceObject(String name){
		if(siblingReferences == null){
			return null;
		}
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].hasLocalName() && siblingReferences[i].getLocalName().equals(name)){
				return siblingReferences[i];
			}
			if(siblingReferences[i].getGlobalName().equals(name)){
				return siblingReferences[i];
			}
		}
		return null;		
	}
	
	/**
	 * @return the sibling reference with the given name. Naming is scoped, taking frst local then global names. null if none exists
	 */
	public IHgraphNode getSiblingReference(String name){
		if(siblingReferences == null){
			return null;
		}
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].hasLocalName() && siblingReferences[i].getLocalName().equals(name)){
				return siblingReferences[i].getRef();
			}
			if(siblingReferences[i].getGlobalName().equals(name)){
				return siblingReferences[i].getRef();
			}
		}
		return null;
	}

	/**
	 * @return the sibling reference with the given local name. null if none exists
	 */
	public Object getSiblingReferenceFromLocalName(String name){		
		if(siblingReferences == null){
			return null;
		}
		
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].hasLocalName() && siblingReferences[i].getLocalName().equals(name)){
				return siblingReferences[i].getRef();
			}
		}
		return null;
	}

	/**
	 * @return the sibling reference with the given global name. null if none exists
	 */
	public Object getSiblingReferenceFromGlobalName(String name){
		if(siblingReferences == null){
			return null;
		}
		
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].getGlobalName().equals(name)){
				return siblingReferences[i];
			}
		}
		return null;
	}
	
	/**
	 * Removes all child sibling references pertaining to a named referer.
	 * @param name
	 * @param startindex
	 * @return a status message
	 */
	private int recursiveRemoveReferences(String name, int startindex){
		for(int i = 0; i<childSiblingReferenceMap.length; i++){
			if(childSiblingReferenceMap[i].getReferer().equals(name)){
				ChildSiblingReference[] tmp = new ChildSiblingReference[childSiblingReferenceMap.length - 1];
				System.arraycopy(childSiblingReferenceMap, 0, tmp, 0, i);
				System.arraycopy(childSiblingReferenceMap, i+1, tmp, i, childSiblingReferenceMap.length - i - 1);
				childSiblingReferenceMap = tmp;
				return recursiveRemoveReferences(name, i);
			}
		}
		return IHgraphNode.STATUS_OK;
	}

	/**
	 * Changes the Ojbect reference of a sibling reference. This is called from the parent node when
	 * the referred sibling changes (i.e. when the sibling is replaced by one with the same name)
	 * @param globalName
	 * @param newRef
	 * @return and int indicating whether the chance has succeeded
	 */
	private int changeReference(String globalName, IHgraphNode newRef){
		if(siblingReferences == null){
			return IHgraphNode.SIBLING_NONEXISTENT;
		}
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].getGlobalName().equals(globalName)){
				siblingReferences[i].setRef(newRef);
			}
		}
		return IHgraphNode.STATUS_OK;		
	}
	
	/**
	 * Changes the global name of a SiblingReference. This is called from the parent node if the name of the
	 * referred sibling changes
	 * @param oldName
	 * @param newName
	 * @return An int indicating whether the change has succeeded
	 */
	private int changeGlobalName(String oldName, String newName){
		if(siblingReferences == null){
			return IHgraphNode.SIBLING_NONEXISTENT;
		}
		for(int i = 0; i < siblingReferences.length; i++){
			if(siblingReferences[i].getGlobalName().equals(oldName)){
				siblingReferences[i].setGlobalName(newName);
			}
		}
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * @return an hgraphMXL representation of the sibling references of the given node. If no node are present, the string is empty
	 */
	public String siblingReferencesToXML(){
		if(siblingReferences == null){
			return "";
		}
		StringBuilder build = StringBuilder.instantiate();
		build.append("<");
		build.append(SIBLINGREFERENCES_TAG);
		build.append(">\n");
		for(int i = 0; i<siblingReferences.length; i++){
			build.append(siblingReferences[i].toXML());
		}
		build.append("</");
		build.append(SIBLINGREFERENCES_TAG);
		build.append(">\n");
		return build.toString();
	}
	
	/**
	 * @return an hgraphXML representation of the sibling references of the given nodes children. If node are present, the string is empty
	 */
	public String childSiblingReferenceMapToXML(){
		if(childSiblingReferenceMap == null){
			return "";
		}
		StringBuilder build = StringBuilder.instantiate();
		build.append("<");
		build.append(CHILDSIBLINGREFERENCES_TAG);
		build.append(">\n");

		for(int i = 0; i<childSiblingReferenceMap.length; i++){
			build.append(childSiblingReferenceMap[i].toXML());
		}
		build.append("</");
		build.append(CHILDSIBLINGREFERENCES_TAG);
		build.append(">\n");
		return build.toString();		
	}
}
