/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import ist.palcom.base.StringBuilder;

/**
 * Sibling references act as a mainteneance structure for mapping between regular objec references
 * and H-graph meta references.
 * Sibling references have a global name, and they may have a local name.
 * Global names change, if the referred IHgraphNode's name change.
 * Local names take precedence over global names, and they are as such overriding when the same
 */
public class SiblingReference {

	private final String localName;
	private String globalName;
	private IHgraphNode siblingRef;
	
	/**
	 * Creates a Sibling Reference from the global name and object reference to the sibling 
	 * @param siblingRef The node to be referred
	 */
	public SiblingReference(IHgraphNode siblingRef){
		this.globalName = siblingRef.getNodeName();
		this.siblingRef = siblingRef;
		localName = null;
	}

	/**
	 * Creates a Sibling Reference from a local name and the global name an object reference to the sibling 
	 * @param localName A local name for the sibling reference. The local name is static and cannot be changed after
	 * creation time. It remains the same although the global name is changed.
	 * @param siblingRef the node to be referred
	 */
	public SiblingReference(String localName, IHgraphNode siblingRef){
		this.localName = localName;
		this.siblingRef = siblingRef;
		this.globalName = siblingRef.getNodeName();
	}
	
	/**
	 * @return True if the sibling reference has a local name. False if not
	 */
	public boolean hasLocalName(){
		return localName != null;
	}
	
	/**
	 * @return The local name of the sibling reference
	 */
	public String getLocalName(){
		return localName;
	}
	
	/**
	 * @return The global name of the sibling reference
	 */
	public String getGlobalName(){
		return globalName;
	}
	
	/**
	 * @return the Object reference.
	 */
	public IHgraphNode getRef(){
		return siblingRef;
	}
	
	/**
	 * Sets the global name. This possibility is required as node names are subject to change
	 * @param globalName
	 */
	public void setGlobalName(String globalName){
		this.globalName = globalName;
	}
	
	/**
	 * Sets the object reference. This possibiliy is required in order to keep the sibling reference
	 * up to date if a node is removed and another with the same name is inserted.
	 * @param ref
	 */
	public void setRef(IHgraphNode ref){
		this.siblingRef = ref;
	}
	
	/**
	 * @return an HgraphXML version of the reference
	 */
	public String toXML(){
		StringBuilder build = StringBuilder.instantiate();
		build.append("<");
		build.append(getClass().toString());
		if(localName != null){
			build.append(" localname=\"");
			build.append(localName);
			build.append("\"");
		}
		build.append(" globalname=\"");
		build.append(globalName);
		build.append("\" ref=\"");
		if(siblingRef == null){
			build.append("null");
		}else{
			build.append(siblingRef.hashCode());
		}
		build.append("\"/>\n");
		return build.toString();
	}
}
