/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph.manager;

import java.util.Iterator;

import ist.palcom.base.HashMap;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.FuncHandler;
import ist.palcom.hgraph.HgraphEvent;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.NodeHandler;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.PrintHandler;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.Value;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.palcomthreads.Monitor;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;

/**
 * The HgraphManager provides a single point of entry to the whole Hgraph.
 * It provides both a synchronous (via regular method invokation) as well as an (potentially) asynchronous (via Hgraph invokations) interface
 */
public class HgraphManager implements IHgraphNode{

	private HgraphManagerReturnHandler managerReturnHandler;
	private HgraphManagerQueryHandler managerQueryHandler;
	private HgraphManagerInvokator managerInvokator;
	private PalcomScheduler scheduler;
	private HashMap monitorMap;
	private HashMap queryMap;
	
	/*****hgraph impl*****/
	protected HgraphNode _node;
	/*********************/
	
	public HgraphManager(PalcomScheduler scheduler){
		
		/******hgraph impl*******/
		_node = new HgraphNode("hgraphmanager");
		/************************/
		
		this.scheduler = scheduler;
		monitorMap = new HashMap();
		queryMap = new HashMap();
		managerReturnHandler = new HgraphManagerReturnHandler(this);
		getFuncFunc().addHandler(managerReturnHandler, null);
		managerQueryHandler = new HgraphManagerQueryHandler(this);
		getFuncFunc().addHandler(managerQueryHandler, null);
		managerInvokator = new HgraphManagerInvokator(this);
		getFuncFunc().addInvokator(managerInvokator, null);
	}
	
	/**
	 * Checks whether a given path exists
	 * @param path the path
	 * @return true if it does, false if not.
	 */
	public boolean cd(String path){
		String sPath = path;
		if(path.equals("##") || path.equals("#")){
			return true;
		}
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		if(sPath.length() == 0){
			return true;
		}else{
			String[] splitPath = sPath.split(Path.PATH_DELIMITER);
			System.out.println("len: "+splitPath.length);
			if(splitPath.length == 1){
				if(sPath.equals(getPath().getPathArray()[0])){
					return true;
				}else{
					return false;
				}
			}else{
				String name = splitPath[splitPath.length - 1];
				String[] newPath = new String[splitPath.length - 1];
				System.arraycopy(splitPath, 0, newPath, 0, splitPath.length - 1);

				HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
				Path tmp = new Path(newPath);
				String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_CD;
				System.out.println("Setting return key: "+key);
				monitorMap.put(key, mon);
				HgraphEvent event = mon.invoke(tmp, NodeHandler.CMD_HASCHILD, new Value("name", name), HgraphManagerReturnHandler.CMD_RETURN_CD);
				monitorMap.remove(key);
				if(event != null){
					if(((Boolean)event.getValue().getValue()).booleanValue() == true){
						return true;
					}					
				}
				return false;
			}
		}		
	}
	
	public void setLiveView(Path whichPath, Path returnPath){
		getFuncFunc().getGenericInvokator().invoke(FuncHandler.CMD_ADDLISTENER, whichPath, new Value("", new LiveViewListener(returnPath)));
	}
	public void unsetLiveView(Path whichPath, Path returnPath){
		getFuncFunc().getGenericInvokator().invoke(FuncHandler.CMD_REMOVELISTENER, whichPath, new Value("", LiveViewListener.name));
	}
	
	/**
	 * Returns the info for some given dir
	 * @param path the path of the node to dir
	 * @return A string giving the dir
	 */
	public String dir(String path){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = Path.fromString(sPath);
		String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_DIR;
		System.out.println("Setting return key: "+key);
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, PrintHandler.CMD_PRINT, HgraphManagerReturnHandler.CMD_RETURN_DIR);
		monitorMap.remove(key);
		if(event != null){
			return (String)event.getValue().getValue();
		}
		return "Error reported from hgraphmanager";
	}
	
	/**
	 * Returns the recursive account of some given dir and its subdirs
	 * @param path the path of the node to dir
	 * @return A string giving the recuesive account
	 *
	 */
	public String dirRecursively(String path){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = Path.fromString(sPath);
		String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_DIR;
		System.out.println("Setting return key: "+key);
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, PrintHandler.CMD_PRINT_RECURSIVELY, HgraphManagerReturnHandler.CMD_RETURN_DIR);
		monitorMap.remove(key);
		if(event != null){
			return (String)event.getValue().getValue();
		}
		return "Error reported from hgraphmanager";
	}

	/**
	 * Returns the full info for some given dir
	 * @param path the path of the node to dir
	 * @return A string giving the full dir
	 */
	public String dirFull(String path){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = Path.fromString(sPath);
		String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_DIR;
		System.out.println("Setting return key: "+key);
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, PrintHandler.CMD_PRINT_FULL, HgraphManagerReturnHandler.CMD_RETURN_DIR);
		monitorMap.remove(key);
		if(event != null){
			return (String)event.getValue().getValue();
		}
		return "Error reported from hgraphmanager";
	}	
	
	/**
	 * Creates a new node
	 * @param path the path on which to install the node
	 * @param name the name of the new node
	 * @return status indicating the status of the command
	 */
	public int newNode(String path, String name){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = Path.fromString(sPath);
		String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_NEWNODE;
		System.out.println("Setting return key: "+key);
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, NodeHandler.CMD_NEWCHILD, new Value("name", name), HgraphManagerReturnHandler.CMD_RETURN_NEWNODE);
		monitorMap.remove(key);
		if(event != null){
			return ((Integer)event.getValue().getValue()).intValue();
		}
		return IHgraphNode.ERROR;
	}
	
	/**
	 * Removes a given node
	 * @param path parent the path of the node
	 * @param name the name of the node to be removed
	 * @return status indicating the status of the command
	 */
	public int removeNode(String path, String name){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = Path.fromString(sPath);
		String key = tmp + HgraphManagerReturnHandler.CMD_RETURN_REMOVENODE;
		System.out.println("Setting return key: "+key);
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, NodeHandler.CMD_REMOVECHILD, new Value("name", name), HgraphManagerReturnHandler.CMD_RETURN_REMOVENODE);
		monitorMap.remove(key);
		if(event != null){
			return ((Integer)event.getValue().getValue()).intValue();
		}
		return IHgraphNode.ERROR;
	}
	
	/**
	 * Sets the name of a node
	 * @param path The path of the node whose name is to be set
	 * @param oldName the old name
	 * @param newName the new name
	 * @return status indicating the status of the command
	 */
	public int setNodeName(String path, String oldName, String newName){
		String sPath = path;
		if(path.startsWith("##")){
			sPath = getPath().getPathArray()[0] + Path.PATH_DELIMITER + path.substring(2);
		}else if(path.startsWith(Path.PATH_DELIMITER)){
			sPath = path.substring(Path.PATH_DELIMITER.length());
		}
		HgraphRetValMonitor mon = new HgraphRetValMonitor(scheduler);
		Path tmp = new Path(oldName, Path.fromString(sPath).getPathArray());
		Path ret = new Path(newName, Path.fromString(sPath).getPathArray());
		ret.getPathArray()[ret.getPathArray().length - 1] = newName;
		String key = ret + HgraphManagerReturnHandler.CMD_RETURN_SETNAME;
		monitorMap.put(key, mon);
		HgraphEvent event = mon.invoke(tmp, NodeHandler.CMD_SETNAME, new Value("name", newName), HgraphManagerReturnHandler.CMD_RETURN_SETNAME);
		monitorMap.remove(key);
		if(event != null){
			return ((Integer)event.getValue().getValue()).intValue();
		}
		return IHgraphNode.ERROR;
	}
	
	/**
	 * Callback function for use by HgraphManager ReturnHandler
	 * Used for either returning reposts to the original recipient through hgraphs
	 * or for returning regular blocking calls
	 * 
	 * @param event The event holding the return information
	 */
	public void returnEvent(HgraphEvent event){
		String tmp = event.getFromPath() + event.getCommand();
		System.out.println("Searching for return key: "+tmp);
		HgraphRetValMonitor mon = (HgraphRetValMonitor)monitorMap.get(tmp);
		if(mon != null){
			monitorMap.remove(tmp);
			mon.retEvent(event);
		}else{
			HgraphEvent qEvent = (HgraphEvent)queryMap.get(tmp);
			if(qEvent != null){
				queryMap.remove(tmp);
				getFuncFunc().getGenericInvokator().invoke(qEvent.getReturnCommand(), qEvent.getFromPath(), event.getValue());
			}else{
				qEvent = (HgraphEvent)queryMap.get("##"+event.getCommand());
				if(qEvent != null){
					queryMap.remove(tmp);
					getFuncFunc().getGenericInvokator().invoke(qEvent.getReturnCommand(), qEvent.getFromPath(), event.getValue());
				}else{
					System.out.println("Hgraph Manager cannot find returner for event: "+event);						
				}
			}
		}
	}

	/**
	 * Callback function for use by HgraphManager QueryHandler.
	 * Used for reposting hgraph queries onto the overall hgraph
	 * 
	 * @param event The caught event holding reposting information
	 */
	public void queryEvent(HgraphEvent event){
		
		Object o = event.getValue().getValue();
		if(o instanceof IHgraphNode){
			IHgraphNode n = (IHgraphNode)o;
			Path p = (Path)n.getValueFunc().getValue("path").getValue();
			if(p != null){
				if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_CD)){
					String key = p + HgraphManagerReturnHandler.CMD_RETURN_CD;
					queryMap.put(key, event);
					managerInvokator.invokeHasChild(p);
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_DIR)){
					String key = p + HgraphManagerReturnHandler.CMD_RETURN_DIR;
					queryMap.put(key, event);
					managerInvokator.invokePrint(p);
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_DIRFULL)){
					String key = p + HgraphManagerReturnHandler.CMD_RETURN_DIR;
					queryMap.put(key, event);
					managerInvokator.invokePrintFull(p);
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_DIRRECURSIVE)){
					String key = p + HgraphManagerReturnHandler.CMD_RETURN_DIR;
					queryMap.put(key, event);
					managerInvokator.invokePrintFullRecursively(p);
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_NEWNODE)){
					String name = (String)n.getValueFunc().getValue("name").getValue();
					if(name != null){
						String key = p + HgraphManagerReturnHandler.CMD_RETURN_NEWNODE;
						queryMap.put(key, event);
						managerInvokator.invokeNewChild(p, new Value("name", name));
					}
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_REMOVENODE)){
					String name = (String)n.getValueFunc().getValue("name").getValue();
					if(name != null){
						String key = p + HgraphManagerReturnHandler.CMD_RETURN_REMOVENODE;
						queryMap.put(key, event);
						managerInvokator.invokeRemoveChild(p, new Value("name", name));
					}
				}else if(event.getCommand().equals(HgraphManagerQueryHandler.CMD_MG_SETNAME)){
					String name = (String)n.getValueFunc().getValue("name").getValue();
					if(name != null){
						String key = p + HgraphManagerReturnHandler.CMD_RETURN_SETNAME;
						queryMap.put(key, event);
						managerInvokator.invokeSetName(p, new Value("name", name));
					}
				}
			}
		}
	}
	
	/**
	 * Monitor class for use with blocking (synchronous) calls to hgraph invoke
	 */
	private class HgraphRetValMonitor extends Monitor{
		
		private Condition valueAvailable = new Condition();
		private HgraphEvent retEvent = null;
		private HelperThread ht;
		 
		public HgraphRetValMonitor(PalcomScheduler scheduler){
			super(scheduler);
		}
		
		public HgraphEvent invoke(Path path, String command, Value val, String retCommand){
			enterMonitor();
			ht = new HelperThread(path, command, val, retCommand);
			scheduler.scheduleThread(ht);
			valueAvailable.await();
			exitMonitor();
			ht.terminate();
			return retEvent;
		}
		
		public HgraphEvent invoke(Path path, String command, String retCommand){
			enterMonitor();
			ht = new HelperThread(path, command, retCommand);
			scheduler.scheduleThread(ht);
			valueAvailable.await();
			exitMonitor();
			ht.terminate();
			return retEvent;
		}

		public void retEvent(HgraphEvent event){
			enterMonitor();
			try{
				this.retEvent = event;
				valueAvailable.causeAll();
			}finally{
				exitMonitor();
			}
		}
	}
	private class HelperThread extends PalcomPriorityThread{
		
		private Path path;
		private String command;
		private String retCommand;
		private Value val;
		
		public HelperThread(Path path, String command, String retCommand){
			this.path = path;
			this.command = command;
			this.retCommand = retCommand;
		}
		public HelperThread(Path path, String command, Value val, String retCommand){
			this.path = path;
			this.command = command;
			this.retCommand = retCommand;
			this.val = val;
		}
		
		public void run(){
			if(val == null){
				getFuncFunc().getGenericInvokator().invoke(command, path, retCommand);
			}else{
				getFuncFunc().getGenericInvokator().invoke(command, path, retCommand, val);
			}
			waitEvent();
		}
	}
	
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}

}