/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hmap;

import ist.palcom.base.StringBuilder;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

public class HmapDir extends HmapNode {
	
	public static final String HMAPDIR_COMMAND_ADDNODE = "addNode";
	public static final String HMAPDIR_COMMAND_REMOVENODE = "removeNode";

	/**
	 * Collection of entries contained by this h-map dir.
	 */
	protected Collection entries = new ArrayList();

	public HmapDir(){
		super(HmapNode.HET_DIR, null, "HmapDir", "1.0", "ist.palcom.hmap.HmapDir");
		addHandler(new EntryHandler());		
	}
	/**
	 * @param name The name of the directory
	 */
	public HmapDir(String name) {
		super(HmapNode.HET_DIR, name, "HmapDir", "1.0", "ist.palcom.hmap.HmapDir");
		addHandler(new EntryHandler());
	}
	
	protected HmapDir(String name, String typeName, String typeVersion, String typeFullyQualifiedVersion){
		super(HmapNode.HET_DIR, name, typeName, typeVersion, typeFullyQualifiedVersion);
		addHandler(new EntryHandler());
	}
		
	public HmapNode resolveLocally(HmapKey key, int type){
		Iterator iter = entries.iterator();
		while(iter.hasNext()){
			HmapNode node = (HmapNode)iter.next();
			if(node.getKey().equals(key)){
				if((node.isDir() && (type == HET_DIR)) || (node.isLeaf() && (type == HET_LEAF)) || (type == HET_UNKNOWN)){
					return node;
				}
			}
		}
		return null;
	}
	
	/**
	 * Resolves the node a the given path
	 * @return The resolved node if it exists. Null if not
	 */
	public HmapNode resolve(HmapPath resolvePath, int type){
		if(path == null){
			recalculatePath();
		}
		if((resolvePath.getLength() == 1) && (path.getLength() == 1)){ //if we are both root nodes
			if(getKey().equals(resolvePath.getKeyAt(0))){ //if we have equal keys
				return this;	
			}
			return null;
		}
		int ctr = 0;
		int max = 0;
		if(path.getLength() > resolvePath.getLength()){
			max = resolvePath.getLength();
		}else{
			max = path.getLength();
		}
		while((ctr < max)){
			if(path.getKeyAt(ctr).equals(resolvePath.getKeyAt(ctr))){
				ctr ++;
			}else{
				break;
			}
		}
		if(ctr == path.getLength()){
			return resolveDown(resolvePath, type);
		}
		return resolveUp(resolvePath, type, ctr);
	}
	
	/**
	 * Relays the invocation of a handler in accordance with its destination given in the event
	 * @param event The event to invoke the handlers with
	 */
	protected void invokeHandlers(HmapEvent event){
		if(path == null){
			recalculatePath();
		}
		if((event.getToPath().getLength() == 1) && (path.getLength() == 1)){ //if we are both root nodes
			if(event.getToPath().getKeyAt(0).equals(getKey())){ //if we have same keys
				Iterator iter = handlerIterator();
				while(iter.hasNext()){
					Handler handler = (Handler)iter.next();
					handler.handle(event);
				}
			}
		}
		int ctr = 0;
		int max = 0;
		if(path.getLength() > event.getToPath().getLength()){
			max = event.getToPath().getLength();
		}else{
			max = path.getLength();
		}
		while((ctr < max)){
			if(path.getKeyAt(ctr).equals(event.getToPath().getKeyAt(ctr))){
				ctr ++;
			}else{
				break;
			}
		}
		if(ctr == 0){
			return;
		}
		if(ctr == path.getLength()){
			invokeHandlersDown(event);
		}else{
			invokeHandlersUp(event, ctr);
		}
	}

	/**
	 * Resolves a node on an upwards path with background in a precalculated common point
	 * @param resolvPath The path to resolve
	 * @param type The type of the node to resolve
	 * @param commonPoint The precalculated common point between the current path and the path to be resolved
	 * @return
	 */
	private HmapNode resolveUp(HmapPath resolvPath, int type, int commonPoint){
		if(path == null){
			recalculatePath();
		}
		if(path.getLength() == commonPoint){
			return resolveDown(resolvPath, type);
		}else{
			return getParent().resolveUp(resolvPath, type, commonPoint);
		}
	}

	/**
	 * Invokes a handler on an upwards path with background in a precalculated common point
	 * @param event The HmapEvent
	 * @param commonPoint The precalculated common point between the current path and the path to be resolved
	 */
	protected void invokeHandlersUp(HmapEvent event, int commonPoint){
		if(path == null){
			recalculatePath();
		}
		if(path.getLength() == commonPoint){
			invokeHandlersDown(event);
		}else{
			getParent().invokeHandlersUp(event, commonPoint);
		}		
	}

	/**
	 * Resolves a node on a downwards path.
	 * @param resolvPath The path of the node to be resolved
	 * @param type The type of the node to be resolved
	 * @return The resolved node. Null if no such exists
	 */
	private HmapNode resolveDown(HmapPath resolvPath, int type){
		if(path == null){
			recalculatePath();
		}
		if(resolvPath.getLength() <= path.getLength()){ //less-than should never happen..
			return null;
		}
		if(!resolvPath.getKeyAt(path.getLength()).hasName()){
			return null;
		}
		Iterator iter = entries.iterator();
		while(iter.hasNext()){
			HmapNode node = (HmapNode)iter.next();
			if(node.hasName()){
				if(node.getName().equals(resolvPath.getKeyAt(path.getLength()).getName())){
					if(resolvPath.getLength() == (path.getLength() + 1)){
						if((node.isDir() && (type == HET_DIR)) || (node.isLeaf() && (type == HET_LEAF)) || (type == HET_UNKNOWN)){
							return node;
						}else{
							return null;
						}
					}else{
						if(!node.isDir()){
							return null;
						}
						return node.asDir().resolveDown(resolvPath, type);
					}
				}	
			}
		}
		return null;
	}
	
	/**
	 * Invokes a handler on a downwards path.
	 * @param event The HmapEvent
	 */
	protected void invokeHandlersDown(HmapEvent event){
		if(path == null){
			recalculatePath();
		}
		if(event.getToPath().getLength() == path.getLength()){
			Iterator iter = handlerIterator();
			while(iter.hasNext()){
				Handler handler = (Handler)iter.next();
				handler.handle(event);
			}
			return;
		}
		if(event.getToPath().getLength() <= path.getLength()){ //less-than should never happen..
			return;
		}
		if(!event.getToPath().getKeyAt(path.getLength()).hasName()){
			return;
		}		
		Iterator iter = entries.iterator();
		while(iter.hasNext()){
			HmapNode node = (HmapNode)iter.next();
			if(node.hasName()){
				if(node.getKey().equals(event.getToPath().getKeyAt(path.getLength()))){
					if(node.isDir()){
						node.asDir().invokeHandlersDown(event);
						return;
					}else{
						node.invokeHandlers(event);
					}
				}	
			}
		}		
	}

	/**
	 * Adds a hmap node to this dir
	 * Ensures that the child name is unique
	 * @param node The node to add
	 */
	public void addNode(HmapNode node) {
		_addNode(node);
		invokeListeners(new HmapEvent(null, getPath(), HMAPDIR_COMMAND_ADDNODE, node));
	}
	
	/**
	 * Adds a hmap node to this dir
	 * Ensures that the child name is unique
	 * @param event The event containing node to add
	 */
	public void addNode(HmapEvent event){
		HmapNode pars = event.getParameters();
		if(pars != null){
			_addNode(pars);
		}
		invokeListeners(event);
	}

	protected void _addNode(HmapNode node){
		if(!node.isAncestorOf(this)){
			if(node.hasName()){
				Iterator iter = entries.iterator();
				while(iter.hasNext()){
					HmapNode tmpNode = (HmapNode)iter.next();
					if(tmpNode.getKey().equals(node.getKey())){
						//TODO: Throw Exception
						return;
					}
				}
			}
			this.entries.add(node);
			node.setParent(this);
		}else{
			//TODO: Throw Exception
		}		
	}
	
	/**
	 * @return True if this node is an ancestor of the given node. False if not
	 */
	protected boolean isAncestorOf(HmapNode node){
		if(this.equals(node)){
			return true;		
		}
		Iterator iter = entries.iterator();
		while(iter.hasNext()){
			if(((HmapNode)iter.next()).isAncestorOf(node)){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Removes a node from this dir
	 * @param node The entry to remove
	 */
	public void removeNode(HmapNode node) {
		this.entries.remove(node);
		invokeListeners(new HmapEvent(null, getPath(), HMAPDIR_COMMAND_REMOVENODE, node));
	}
	
	
	public void removeNode(HmapEvent event){
		if(event.getParameters() != null){
			this.entries.remove(event.getParameters());
		}
		invokeListeners(event);
	}
	
	/**
	 * Removes the first found child node of this dir named <name>
	 * @param name The name of the child node to remove
	 * @return True if a node has been removed. False if no node named <name> was found
	 */
	public boolean removeNode(String name){
		Iterator iter = entries.iterator();
		HmapNode node;
		while(iter.hasNext()){
			node = (HmapNode)iter.next();
			if(node.hasName() && node.getName().equals(name)){
				removeNode(node);
				return true;
			}
		}
		return false;
	}
		
	/**
	 * @return The iterator of the internal datastructure holding the child entries
	 */
	public Iterator iterator() {
		return entries.iterator();
	}

	protected void recalculatePath(HmapPath parentPath) {
		this.path = new HmapPath(parentPath, getKey());
		Iterator iter = entries.iterator();
		while(iter.hasNext()){
			HmapNode node = (HmapNode)iter.next();
			node.recalculatePath(this.path);
		}
	}

	/**
	 * @return A stringified version of the dir and its children
	 */
	public String toString() {
		return toString(0);
	}

	/**
	 * @return A stringified version of the dir
	 */
	public String toStringFlat() {
		StringBuilder tmp = StringBuilder.instantiate();
		
		tmp.append(getKey().toString() + "#\r\n");
		
		tmp.append(printHandlers(3));
		tmp.append(printInvokators(3));
		tmp.append(printListeners(3));
		
		Iterator iter = entries.iterator();
		while (iter.hasNext()) {
			HmapNode n = (HmapNode) iter.next();
			tmp.append("   ");
			tmp.append(n.getKey().toString() + "#\r\n");
		}
		return tmp.toString();
	}

	/**
	 * @param indent Start the returned string with a number of whitespaces
	 * @return A stringified version of the dir and its children
	 */
	public String toString(int indent) {
		StringBuilder tmp = StringBuilder.instantiate();
		for (int i = 0; i < indent; i++){
			tmp.append(" ");
		}
		if(getKey().hasName()){
			tmp.append(getKey().getName().toString());			
		}else{
			tmp.append("Anonymous");
		}
		tmp.append("#\r\n");
		
		Iterator iter = entries.iterator();
		while (iter.hasNext()) {
			HmapNode n = (HmapNode) iter.next();
			tmp.append(n.toString(indent + 3));
		}
		return tmp.toString();
	}	

	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node with all handlers, invokators and listernes also
	 */
	public String toStringAll(int indent) {
		StringBuilder tmp = StringBuilder.instantiate();
		for (int i = 0; i < indent; i++){
			tmp.append(" ");
		}
		tmp.append(getKey().toString());
		tmp.append("#\r\n");
		
		tmp.append(printHandlers(indent + 3));
		tmp.append(printInvokators(indent + 3));
		tmp.append(printListeners(indent + 3));
		
		Iterator iter = entries.iterator();
		while (iter.hasNext()) {
			HmapNode n = (HmapNode) iter.next();
			tmp.append(n.toStringAll(indent + 3));
		}
		return tmp.toString();
	}	
	
	public class EntryHandler extends AbstractHandler{
		
		private static final String name = "Dir Entry Handler";
		private static final String description = "Handles the adding and removing of entries in this dir";
		
		public void handle(HmapEvent event){
			if(event.getCommand().equals(HMAPDIR_COMMAND_ADDNODE)){
				addNode(event);
			}else if(event.getCommand().equals(HMAPDIR_COMMAND_REMOVENODE)){
				HmapNode pars = event.getParameters();
				if(pars != null){
					if(pars.isLeaf() && (((HmapLeaf)pars).getValue() instanceof String)){
						removeNode((String)((HmapLeaf)pars).getValue());
					}else{
						removeNode(pars);
					}
				}
			}
		}
		
		public String getName(){
			return name;
		}
		
		public String getDescription(){
			return description;
		}
	}	
}
