/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hmap;

import ist.palcom.base.StringBuilder;

/**
 * An HmapKey represents the identity for any HmapNode.
 * I is a (name, type) pair, but only one of the two need be present.
 * Thus, as types are not sibling unique, neither can keys be guaranteed to be.
 * @author tonz
 *
 */
public class HmapKey {

	/**
	 * The name of this key's node
	 */
	private String name;

	private final String TYPE_NAME;
	private final String TYPE_VERSION;
	private final String TYPE_FULLYQUALIFIEDNAME;

	/**
	 * @param name The name of the key (node)
	 * @param type The type object corresponding to the type of the node
	 */
	public HmapKey(String name, String typeName, String typeVersion, String typeFullyQualifiedName){
		this.name = name;
		this.TYPE_NAME = typeName;
		this.TYPE_VERSION = typeVersion;
		this.TYPE_FULLYQUALIFIEDNAME = typeFullyQualifiedName;
	}
	
	public HmapKey(String name){
		this.name = name;
		this.TYPE_NAME = null;
		this.TYPE_VERSION = null;
		this.TYPE_FULLYQUALIFIEDNAME = null;
	}
	
	public static HmapKey fromString(String encodedString){
		String[] tmp = encodedString.split(";");
		int j = 0;
		if(tmp[0].length() == 0){
			j = 1;
		}
		
		String name = null;
		String typeName = null;
		String typeVersion = null;
		String typeFullyQualifiedName = null;
		
		for(int i = j ; i < tmp.length; i++){
			String[] entry = tmp[i].split("::");
			if(entry[0].equals("n")){
				name = entry[1];
			}else if(entry[0].equals("t")){
				typeName = entry[1];
			}else if(entry[0].equals("v")){
				typeVersion = entry[1];
			}else if(entry[0].equals("f")){
				typeFullyQualifiedName = entry[1];
			}
		}
		if((name == null) && (typeName == null) && (typeVersion == null) && (typeFullyQualifiedName == null)){
			name = encodedString;
		}
		return new HmapKey(name, typeName, typeVersion, typeFullyQualifiedName);
	}
	
	public static String toString(HmapKey key){
		StringBuilder tmp = StringBuilder.instantiate();
		boolean first = true;
		if(key.hasName()){
			tmp.append("n::");
			tmp.append(key.getName());
			first = false;
		}
		if(key.hasTypeName()){
			if(!first){
				tmp.append(";");
			}else{
				first = false;
			}
			tmp.append("t::");
			tmp.append(key.getTypeName());
		}
		if(key.hasTypeVersion()){
			if(!first){
				tmp.append(";");
			}else{
				first = false;
			}
			tmp.append("v::");
			tmp.append(key.getTypeVersion());
		}
		if(key.hasTypeFullyQualifiedName()){
			if(!first){
				tmp.append(";");
			}else{
				first = false;
			}
			tmp.append("f::");
			tmp.append(key.getTypeFullyQualifiedName());
		}
		return tmp.toString();
	}
	
	/**
	 * @return A stringified version of this key. The representation is stored for later use.
	 */
	public String toString(){
		return HmapKey.toString(this);
	}
		
	/**
	 * Checks another key for equity with this one. 
	 * As names are guaranteed to be unique, the name is the only thing necessary to chek,
	 * unless both keys only have types. In that case, it is purely a type comparison mechanism.
	 * @param key The key to be compared with
	 * @return true if the keys have the same name and/or type. False is not
	 */
	public boolean equals(HmapKey key){
		
		if(hasName() && key.hasName()){ // XOR
			if(name != null){
				if(name.equals(key.getName())){
					return true;
				}else{
					return false;
				}
			}
		}
		return false;
	}

	/**
	 * @return true if the key has a name. False if not
	 */
	public boolean hasName(){
		return name != null;
	}
	/**
	 * @return true if the key has a type name. False if not
	 */
	public boolean hasTypeName(){
		return TYPE_NAME !=null;
	}
	/**
	 * @return true of the key has a type version. False if not
	 */
	public boolean hasTypeVersion(){
		return TYPE_VERSION != null;
	}
	/**
	 * @return true if the key has a type fully qualified name. False if not.
	 */
	public boolean hasTypeFullyQualifiedName(){
		return TYPE_FULLYQUALIFIEDNAME != null;
	}
	/**
	 * @return The name of the key. Null if it has none
	 */
	public String getName(){
		return name;
	}
	/**
	 * @return The name of the type of the key.
	 */
	public String getTypeName(){
		return TYPE_NAME;
	}
	/**
	 * @return A version string of the type of the key
	 */
	public String getTypeVersion(){
		return TYPE_VERSION;
	}
	/**
	 * @return The fully qualified name of the key
	 */
	public String getTypeFullyQualifiedName(){
		return TYPE_FULLYQUALIFIEDNAME;
	}
	/**
	 * Sets the name of the key. Resets the internal string representation of the key
	 * @param name The new name of the key
	 */
	public void setName(String name){
		this.name = name;
	}	
}
