/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hmap;

import ist.palcom.base.StringBuilder;

/**
 * A TreePath is a path in an h-map tree, represented by an array of PathEntry objects
 * A TreePath can be instantiated with a string encoded as a path of the form:
 * 
 *  #<HmapKey>#<HmapKey>#<HmapKey>
 *  
 * Look to {@link HmapKey} for the encoding of each entry
 * 
 * @author tonz
 *
 */
public class HmapPath {

	/**
	 * An array of HmapKeys corresponding to this path
	 */
	private HmapKey[] path = null;
		
	public HmapPath(HmapKey[] keyArray){
		this.path = keyArray;
	}
	
	/**
	 * Create a new path
	 * @param hmapKey The root entry in a new path
	 */
	public HmapPath(HmapKey hmapKey){
		path = new HmapKey[1];
		path[0] = hmapKey;
	}
		
	/**
	 * @param treePath The path on which to build this path
	 * @param hmapKey The new entry
	 */	
	public HmapPath(HmapPath treePath, HmapKey hmapKey){
		path = new HmapKey[treePath.getLength() + 1];
		for(int i = 0; i<treePath.getLength(); i++){
			path[i] = treePath.getKeyAt(i);
		}
		path[treePath.getLength()] = hmapKey;		
	}

	/**
	 * @return A stringified version of this path.
	 */
	public String toString(){
		StringBuilder tmp = StringBuilder.instantiate();
		for(int i = 0; i<path.length; i++){
			tmp.append("#");
			tmp.append(path[i].toString());
		}
		return tmp.toString();
	}
	
	/**
	 * Returns the key corresponding to a specific entry in the path.
	 * @param i the entry number
	 * @return a correct key if one exists. Null if out of bounds
	 */
	public HmapKey getKeyAt(int i){
		if((i<path.length) && (i>=0)){
			return path[i];
		}
		return null;
	}
	
	/**
	 * @return The length of the path
	 */
	public int getLength(){
		return path.length;
	}
	
	/**
	 * Checks for equality with another path
	 * @param ePath The path to check
	 * @return true of the paths are equal. false if not
	 */
	public boolean equals(HmapPath ePath){
		if(ePath == null){
			return false;
		}
		if(getLength() != ePath.getLength()){
			return false;
		}
		for(int i = 0; i<getLength(); i++){
			if(!path[i].equals(ePath.getKeyAt(i))){
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Returns a subpath of this path. Indices are defined as array counts: from 0 to length-1
	 * @param from The entry to take from; included
	 * @param to The entry to take to; included
	 * @return
	 */
	public HmapPath subPath(int from, int to){
		if((from >= 0) && (to <= getLength())){
			HmapKey[] tmp = new HmapKey[to - from];
			for(int i = from; i < to; i++){
				tmp[i - from] = path[i];
			}
			return new HmapPath(tmp);
		}
		return null;
	}
	
	/**
	 * @param one The new first part of the path
	 * @param two The new second part of the path
	 * @return A merger of two paths into one, with the first parameter as the first part of the path and the second the second
	 */
	public static HmapPath mergePaths(HmapPath one, HmapPath two){
		HmapKey[] tmp = new HmapKey[one.getLength()+two.getLength()];
		for(int i = 0 ; i < one.getLength() ; i++){
			tmp[i] = one.getKeyAt(i);
		}
		for(int i = 0 ; i < two.getLength() ; i++){
			tmp[i + one.getLength()] = two.getKeyAt(i);
		}
		return new HmapPath(tmp);
	}
	
	public static HmapPath fromString(String encodedString){
		String[] tmp = encodedString.split("#");
		int j = 0;
		if(tmp[0].length() == 0){
			j = 1;
			HmapKey[] path = new HmapKey[tmp.length - j];		
			for(int i = j ; i < tmp.length ; i++){
				path[i - 1] = HmapKey.fromString(tmp[i]);
			}
			return new HmapPath(path);
		}else{
			j = 0;
			HmapKey[] path = new HmapKey[tmp.length - j];		
			for(int i = j ; i < tmp.length ; i++){
				path[i] = HmapKey.fromString(tmp[i]);
			}			
			return new HmapPath(path);
		}
	}
}
