/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.localnetworkview;

import ist.palcom.collections.LinearSet;
import ist.palcom.debugging.DebugDispatcher;
import ist.palcom.resource.descriptor.ASTNode;
import ist.palcom.resource.descriptor.Connection;
import ist.palcom.resource.descriptor.ConnectionInfo;
import ist.palcom.resource.descriptor.ConnectionList;
import ist.palcom.resource.descriptor.ControlService;
import ist.palcom.resource.descriptor.Device;
import ist.palcom.resource.descriptor.DeviceInfo;
import ist.palcom.resource.descriptor.List;
import ist.palcom.resource.descriptor.Network;
import ist.palcom.resource.descriptor.NetworkNode;
import ist.palcom.resource.descriptor.ResourceStatusInfo;
import ist.palcom.resource.descriptor.Service;
import ist.palcom.resource.descriptor.ServiceDescription;
import ist.palcom.resource.descriptor.ServiceInfo;
import ist.palcom.resource.descriptor.ServiceInfoList;
import ist.palcom.resource.descriptor.ServiceInfoListItem;
import ist.palcom.resource.descriptor.ServiceListPart;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.resource.descriptor.SubList;
import ist.palcom.resource.descriptor.SubListInfo;
import ist.palcom.resource.descriptor.Universe;

import java.util.Iterator;


/**
 * PalcomDeviceTree describes a device's view of the world. It builds a tree 
 * rooted by the Universe node which has two children:
 * - Net - the tree of devices and services known through discovery
 * - ConnectionList - the list of known connections
 * 
 * The class is responsible for keeping the tree consistent. E.g., if a device 
 * is removed from the tree, all connections having that device as provider 
 * or consumer are also removed.
 * 
 * There are two ways of reacting to changes to the tree:
 * 
 *   - addNode() and removeNode() return true if the request led to an actual
 *     modification to the tree (duplicates and irrelevant messages are ignored)
 *     
 *   - objects implementing DeviceTreeChangeListener can register for notification
 *     when nodes are added, removed, or changed.

 * As much as possible of the functionality (e.g., lookup, etc.) is delegated to
 * the NetworkNode class hierarchy and implemented in the attribute grammar.
 * 
 * @author sven
 *
 */

public class PalcomDeviceTree {
	private Universe universe;
	public PalcomDeviceTree() {
		Network network = new Network();
//		network.setName("Services");
		network.setName("Devices");
		universe = new Universe(network, new ConnectionList());
	}

	/**
	 * @return the root of this tree
	 */
	public Universe getUniverse() {
		return universe;
	}

	/**
	 * addNode is inserts newly discovered services and connections into the
	 * tree. It is also responsible for linking new nodes to existing ones --
	 * e.g., a connection node to its service nodes.
	 * 
	 * The public method delegates the actual work to private methods, one for
	 * each type of node.
	 * 
	 * @author sven
	 * 
	 * @return true if the tree was actually modified
	 */
	public boolean addNode(Object i) {
//		System.err.println("addNode: "+i+" --- "+i.getClass());
		synchronized(getUniverse()) {
			boolean result = false;
			if (i instanceof DeviceInfo) {
				result = addNode((DeviceInfo) i);
			} else if (i instanceof ServiceInfoList) {
				result = addNode((ServiceInfoList) i);
			} else if (i instanceof ServiceDescription) {
				result = addNode((ServiceDescription) i);
			} else if (i instanceof ConnectionInfo) {
				result = addNode((ConnectionInfo) i);
			} else if (i instanceof ServiceProxy) {
				result = addNode((ServiceProxy) i);
			} else if (i instanceof ResourceStatusInfo) {
				result = changeStatus((ResourceStatusInfo) i);
			} else {
				DebugDispatcher.debug("PalcomDeviceTree.addNode(Info): unknown type :" + i + " class == "
						+ i.getClass(), DebugDispatcher.CAT_SYSTEM);
			}
			if(result) universe.resetTree();
//			DebugDispatcher.debug("addNode(Object) returns "+result, DebugDispatcher.CAT_SYSTEM);				
			return result;
		}
	}


	private boolean changeStatus(ResourceStatusInfo info) {
//		System.err.println("*********** PalcomDeviceTres.changeStatus...");
		synchronized(getUniverse()) {
			boolean result = false;
			NetworkNode n = universe.lookup(info.getAddress());
			if(n != null) {
				n.setLatestStatus(info);
				n.resetTree();
//				System.err.println("*********** PalcomDeviceTres.changeStatus ("+(char)info.getStatus()+") for "+n.getBrowserLabel());
				treeChanged(NOTIFY_CHANGE, n);
				result = true;
			} else {
				System.err.println("*********** PalcomDeviceTres.changeStatus: matching node not found: "+info.getAddress());			
			}
			return result;
		}
	}

	/** Remove the node(s) corresponding to an Info, including dependencies (e.g,
	 * if a device is removed, its connections are also removed) 
	 * 
	 * The public method delegates the actual work to private methods, one for
	 * each type of node.
	 * 
	 * @param i - the Info
	 * @return true if the tree was actually modified
	 */
	public boolean removeNode(Object i) {
		synchronized(getUniverse()) {
			boolean result = false;
//			System.out.println("PalcomDeviceTree.removeNode: "+i);
			if (i instanceof DeviceInfo) {
				result = removeNode((DeviceInfo)i);
			} else if (i instanceof ConnectionInfo) {
				result = removeNode((ConnectionInfo)i);
			} else if (i instanceof ServiceInfoList) {
				result = removeNode((ServiceInfoList) i);
			} else if (i instanceof ServiceProxy) {
				result = removeNode((ServiceProxy) i);		
			} else {
				// TODO: Can services be removed from an active device?
				DebugDispatcher.debug("PalcomDeviceTree.RemoveNode: unknown type :" + i
						+ " class == " + i.getClass(), DebugDispatcher.CAT_SYSTEM);
			}
			if(result) universe.resetTree();
			return result;
		}
	}


	private boolean addNode(DeviceInfo i) {
		boolean result = false;
		if (universe.lookup(((DeviceInfo) i).getDeviceAddress()) == null) {
			Device d = new Device();
			d.setInfo((DeviceInfo) i);
			//XXX This is a hack
			d.setLatestStatus((char) ((DeviceInfo)i).getDeviceAddress().getPRDDevice().getDeviceStatus());
			universe.getNet().addDevice(d);
			result = true;
		}
		return result;
	}

	private boolean addNode(ServiceInfoList sl) {
		boolean result = false;
//		System.err.println("addnode(SIL):"+sl);
//		NetworkNode n = universe.getNet().localLookup(sl.getParentURN());
//		System.err.println("... looking up:"+sl.getParentAddress());
		NetworkNode n = universe.lookup(sl.getParentAddress());

		if (n instanceof Device || n instanceof SubList) {
			// Check if services has been removed
//			System.err.println("...found parent: "+n);
//			List parentSL=n.getServiceListPartList();
//			if(parentSL != null && parentSL.getNumChild() > sl.getNumServiceInfoListItem()) {
			if(n.getNumServiceListPart() > sl.getNumServiceInfoListItem()) {
//				//service(s) has been removed from the list
				for(int i=0; i<n.getNumServiceListPart(); i++) {
					ServiceListPart slp = n.getServiceListPart(i);
					if(sl.getIndexOfChild(slp) < 0) {
						System.err.println("**PalcomDeviceTree.addNode(ServiceList): removed info:"+slp);
						result=removeService(n,slp.getServiceInfoListItem());						
					}
				}
			}
			//XXX FIXME
////			if((parentSL).getNumChild() > sl.getNumServiceInfo()) {
//			for(int i=0; i<parentSL.getNumChild(); i++) {
//				ServiceInfo tmpSi = ((Service)parentSL.getChild(i)).getInfo();
//				//XXX fix the following:
////				if(sl.getPRDDataThatEquals(tmpSi)==null){
//				if(sl.getIndexOfChild(tmpSi) >= 0 ) {
//					System.err.println("**PalcomDeviceTree.addNode(ServiceList): removed info:"+tmpSi);
//					result=removeService(n,tmpSi);
//				} 
////				}
//			}

			// Add or update from the new ServiceList
			for (int k = 0; k < sl.getNumServiceInfoListItem(); k++) {
				ServiceInfoListItem sli = sl.getServiceInfoListItem(k);
				if(sli instanceof ServiceInfoListItem) {
					result |= addOrUpdateService(n, (ServiceInfoListItem) sli);
//					else if(sli instanceof ServiceInfoList) {
//					result |= addNode((ServiceInfoList) sli); 
				} else {
					System.err.println("WARNING! PalcomDeviceTree.addNode() can't handle "+sli);
				}
			}
		} else {
			System.err.println("addNode(ServiceList) parent not found:" + sl);//, DebugDispatcher.CAT_SYSTEM);
//			DebugDispatcher.debug("parentURN==" + sl.getParentURN(), DebugDispatcher.CAT_SYSTEM);
		}

		return result;
	}

	/**
	 * Add or update the ServiceInfo of a device or service
	 * 
	 * @param n -
	 *            The parent Device or Service
	 * @param sli -
	 *            The ServiceInfo to add or update
	 * @return true if the tree was modified
	 */
	private boolean addOrUpdateService(NetworkNode n, ServiceInfoListItem sli) {
		ServiceListPart existingSli;
//		DebugDispatcher.debug("PalcomDeviceTree:addOrUpdateService: "+n+" : "+sli, DebugDispatcher.CAT_SYSTEM);
		boolean result = false;
		// TODO: does the lookup have to be global or is it enough to search n's subtree?
		// existingService = (Service) network.localLookup(si);
		existingSli = (ServiceListPart) n.lookup(sli);

		if (existingSli == null) {
//			DebugDispatcher.debug("... creating a new Service node: "+sli.getName(), DebugDispatcher.CAT_SYSTEM);
			// create new ServiceNode
//			Service tmps = new Service();
//			tmps.setInfo(si);
			ServiceListPart tmps = sli.createServiceNode();
			///// dispatch, of sorts.../////////
			if (n instanceof Device) {
//				DebugDispatcher.debug("... adding to Device: "+((Device)n).getName(), DebugDispatcher.CAT_SYSTEM);
				((Device) n).addService(tmps);
//				DebugDispatcher.debug("... device now has "+((Device) n).getNumService()+" children", DebugDispatcher.CAT_SYSTEM);				
//				DebugDispatcher.debug("... size of child array "+((Device) n).getChildArray().length, DebugDispatcher.CAT_SYSTEM);				
			} else if (n instanceof SubList) {
//				DebugDispatcher.debug("... adding to Service: "+((SubList)n).getName(), DebugDispatcher.CAT_SYSTEM);
				((SubList) n).addService(tmps);
			}
			////////////////////////////////////
			treeChanged(NOTIFY_ADD, tmps);
			result = true;
		} else {
			// check for updates
			if(existingSli instanceof Service) {
				Service existingService = (Service)existingSli;
				if (!existingService.getInfo().equals(sli) && sli instanceof ServiceInfo) {
					existingService.setInfo((ServiceInfo)sli);
					DebugDispatcher.debug("Updating ServiceInfo for"
							+ existingService.getName(), DebugDispatcher.CAT_SYSTEM);
					treeChanged(NOTIFY_CHANGE, existingService);
				} 
			} else {
				if(existingSli instanceof SubList) {
					SubList existingSubList = (SubList)existingSli;
					if (!existingSubList.getInfo().equals(sli) && sli instanceof SubListInfo) {
						existingSubList.setInfo((SubListInfo)sli);
						DebugDispatcher.debug("Updating ServiceInfo for"
								+ existingSubList.getName(), DebugDispatcher.CAT_SYSTEM);
						treeChanged(NOTIFY_CHANGE, existingSubList);
					} 
				}
			}
		}
		return result;
	}

	private boolean addNode(ServiceDescription sdi) {
		boolean result = false;
//		System.out.println("addNode(ServiceDescriptionInfo)"+sdi);
		NetworkNode n = universe.lookup(sdi.getServiceAddress());

		if (n != null && n instanceof ControlService) {
			ControlService s = (ControlService) n;
			// Ignore identical descriptions
			if (!sdi.equals(s.getDescription())) {
//				System.out.println("addNode "+sdi);
				ServiceDescription sdi2 = (ServiceDescription) sdi.fullCopy();
				s.setDescription(sdi2);
				treeChanged(NOTIFY_CHANGE,s);
				result = true;
			} else {
//				System.out.println("Identical description:" + sdi);
			}
		} else {
			// TODO: Is this an error or a legal situation?
			DebugDispatcher.debug("addNode(SDI): Provider " + sdi.getServiceAddress()
					+ " not found: n == "+n, DebugDispatcher.CAT_SYSTEM);
		}
		if(result) universe.resetTree();
		return result;
	}

	private boolean addNode(ServiceProxy proxy) {
		boolean result = false;
//		NetworkNode n = universe.getNet().localLookup(desc.getProviderURN());
//		System.out.println("addNode(ServiceDescription)"+desc);
//		System.out.println("desc.getNumControlItem:"+desc.getNumControlItem());
//		XXX don't use LocalSID
		NetworkNode n = universe.lookup(proxy.getLocalSID());

		if (n != null && n instanceof ControlService) {
			ControlService s = (ControlService) n;
			// Ignore identical descriptions
			if (!s.hasProxy() || !proxy.getInfo().equals(s.getDescription())) {
//				System.out.println("PalcomDeviceTree.addNode "+proxy);
				ServiceProxy tmp = (ServiceProxy) proxy.fullCopy();
				s.setProxy(tmp);
				treeChanged(NOTIFY_CHANGE,s);
				result = true;
			} else {
//				System.out.println("Identical description:" + proxy);
			}
		} else {
			// TODO: Is this an error or a legal situation?
			DebugDispatcher.debug("Provider " + proxy.getInfo().getID() +"("+proxy.getLocalSID()+")"
					+ " not found", DebugDispatcher.CAT_SYSTEM);
		}
//		if(result) universe.resetTree();
		return result;
	}

	private boolean addNode(ConnectionInfo ci) {
		boolean result = false;
		ConnectionList cl = universe.getConnections();
		boolean match = false;
		// Ignore duplicate Infos
		for (int j = 0; !match && j < cl.getNumConnection(); j++) {
			match = match || cl.getConnection(j).getInfo().equals(ci);
		}
		if (!match) {
			/* Link connection to services */
			try {
				NetworkNode svc1 = universe.lookup(ci.getCustomerAddress());
				NetworkNode svc2 = universe.lookup(ci.getProviderAddress());
//				System.err.println("PalcomDeviceTree.addNode(CI) custn("+ci.getCustomerAddress()+"): "+ custn);
//				System.err.println("PalcomDeviceTree.addNode(CI) provn("+ci.getProviderAddress()+"): "+ provn);
				// Check if services exist
				if (svc1 instanceof Service && svc2 instanceof Service) {
					//XXX this will not work for services that are both
					// "customers" and "providers": the service nodes need to
					// be changed to allow having a set of connections for each
					// service
					Connection c = new Connection(ci);
					if(((Service)svc1).getInfo().hasDesc())
						((Service)svc1).setConnection(c);
					if(((Service)svc2).getInfo().hasDesc())
						((Service)svc2).setConnection(c);
					universe.getConnections().addConnection(c);
					treeChanged(NOTIFY_ADD,c);
					result = true;
				} else {
					DebugDispatcher.debug("Got ConnectionInfo for nonexitent service:"
							+ ci, DebugDispatcher.CAT_SYSTEM);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * @param i
	 * @return true if the tree was actually modified
	 */
	private boolean removeNode(ConnectionInfo i) {
		ConnectionList cl = universe.getConnections();
		boolean match = false;
		boolean result = false;
		try {
			for (int j = 0; !match && j < cl.getNumConnection(); j++) {
				Connection c = cl.getConnection(j);
				//if (c.getInfo().equals(i)) {
				if(i.isAboutTheSameEntityAs(c.getInfo())) {
					int idx = c.getParent().getIndexOfChild(c);
					if (idx >= 0) {
						c.getParent().removeChild(idx);
						treeChanged(NOTIFY_REMOVE,c);
						result  = true;
					}
				}
			}
		} catch (Exception e){e.printStackTrace();}
		return result;
	}

	/**
	 * @param i
	 * @return true if the tree was actually modified
	 */
	private boolean removeNode(DeviceInfo i) {
		Device d;
		boolean result = false;
		// lookup URN and not Info, as the info objects will not be the same
		// (they differ at least in the "active" field.)
		if ((d = (Device) universe.lookup(i.getAddress())) != null) {
			//TODO: should we get rid of this hack? (see the method comment)
//			removeConnectionsForDevice(d);
			int idx = d.getParent().getIndexOfChild(d);
			if (idx >= 0) {
				d.getParent().removeChild(idx);
				treeChanged(NOTIFY_REMOVE,d);
				result  = true;
			}
		} 
		return result;
	}

	private boolean removeNode(ServiceProxy proxy) {
		boolean result = false;
//		NetworkNode n = universe.getNet().localLookup(desc.getProviderURN());
//		System.out.println("addNode(ServiceDescription)"+desc);
//		System.out.println("desc.getNumControlItem:"+desc.getNumControlItem());
//		XXX don't use LocalSID
		NetworkNode n = universe.lookup(proxy.getLocalSID());

		if (n != null && n instanceof ControlService) {
			ControlService s = (ControlService) n;
//			System.err.println("RemoveNode(proxy)"+s.getName());
			try {
				if(s.hasProxy()) {
					result = s.getProxy().remove();
					if(!result) {
						System.err.println("WARNING! PalcomDeviceTree.remove(proxy): ASTNode.remove failed?");
					}
				}
			} catch(Exception e) {
				System.err.println("PalcomDeviceTree.removeNode(Proxy) : "+e);
			}
			if(result) treeChanged(NOTIFY_CHANGE,s);
		} else {
			// TODO: Is this an error or a legal situation?
			DebugDispatcher.debug("Provider " + proxy.getInfo().getID() +"("+proxy.getLocalSID()+")"
					+ " not found", DebugDispatcher.CAT_SYSTEM);
		}
//		if(result) universe.resetTree();
		return result;
	}


	/** Obsolete helper method: removes all connections dependent on device d
	 * 
	 *  This method is not strictly necessary, but helps avoid problems in
	 *  the presentation due to race conditions when the device is removed
	 *  before its dependant connections: I.e., the text representation of
	 *  a connection does getProvider().parentDevice().getName()...
	 *  
	 *  This way, a lot of null tests and exception handling is avoided.
	 * 
	 * @param d
	 * @deprecated -- this is probably obsolete
	 */
	private void removeConnectionsForDevice(Device d) {
		ConnectionList cl = d.universe().getConnections();

		// for(int j=0; j < cl.getNumConnectionInfo(); j++){
		int j = 0;
		while (j < cl.getNumConnection()) {
			Connection c = cl.getConnection(j);
			if (c.getProvider().parentDevice().equals(d)
					|| c.getCustomer().parentDevice().equals(d)) {
				int idx = c.getParent().getIndexOfChild(c);
				if (idx >= 0) {
					c.getParent().removeChild(idx);
					treeChanged(NOTIFY_REMOVE,c);
					// Hack! As the child numbering has changed, start
					// over from the beginning. (would j-- work ?) 
					// How is this done properly?
					j = 0; 
					continue;
				}
			}
			j++;
		}
	}

	private boolean removeNode(ServiceInfoList sl) {
		boolean result = false;
		NetworkNode n = universe.lookup(sl.getParentAddress());

		if (n instanceof Device || n instanceof Service) {
			for (int k = 0; k < sl.getNumServiceInfoListItem(); k++) {
				ServiceInfoListItem sli = sl.getServiceInfoListItem(k);
//				PRDService si = (PRDService)sl.getPRDSecondOrder(k);
				if( sli instanceof ServiceInfoListItem) {
					//XXX adapt this to new grammar
					result = removeService(n, sli);
//					} else if(sli instanceof ServiceInfoList) {
//					removeNode( (ServiceInfoList) sli);
				} else {
					System.err.println("Warning! PalcomDeviceTree: Unknown type in removeNode: "+sli.getClass());
					throw new Error("Unknown type in removeNode: "+sli.getClass());
				}
			}
		} else {
			DebugDispatcher.debug("removeNode(ServiceList) parent not found:" + sl, DebugDispatcher.CAT_SYSTEM);
			DebugDispatcher.debug("parentURN==" + sl.getParentAddress(), DebugDispatcher.CAT_SYSTEM);
		}

		return result;
	}

	/**
	 * Remove the Service corresponding to a ServiceInfo from a device or service
	 * 
	 * @param n -
	 *            The parent Device or Service
	 * @param slp -
	 *            The ServiceInfo to of the service to be removed
	 * @return true if the tree was modified
	 */
	private boolean removeService(NetworkNode n, ServiceInfoListItem slp) {
		ServiceListPart existingService;
		boolean result = false;
		// TODO: does the lookup have to be global or is it enough to search n's subtree?
		// existingService = (Service) network.localLookup(si);
		existingService = (ServiceListPart) n.lookup(slp);

		if (existingService != null) {
			ASTNode parent = existingService.getParent();
			int idx = parent.getIndexOfChild(existingService);

			// find parent device or service
//			ASTNode parentNode = parent;
//			while(parentNode != null && ! (parentNode instanceof NetworkNode))
//			parentNode = parentNode.getParent();
//			if(parentNode != n) {
//			System.out.println("parentNode = "+parentNode+", n = "+n);
//			}

			if(idx >= 0) {
//				System.out.println("removing " +existingService+" from " + parent);
				//System.out.println("Notifying "+n);
				parent.removeChild(idx);
				treeChanged(NOTIFY_REMOVE, existingService);
				treeChanged(NOTIFY_CHANGE, n);
				result = true;
			}  else {

				System.out.println("PalcomDeviceTree::removeService: idx < 0???:"+slp);
			}
		} else {
			System.out.println("PalcomDeviceTree::removeService failed. Could not find "+ slp);
		}
		return result;
	}


	/**
	 * A crude way of implementing refresh of the communications list
	 * 
	 */
	public void removeConnections() {
		synchronized(getUniverse()) {
//			Network network;
//			universe.setNet(network=new Network());
//			network.setName("Services");

//			ConnectionList l = universe.getConnections();
			universe.setConnections(new ConnectionList());
			// TODO: This doesn't work: must sleep(timeout) for discovery before doing comparison.
			// change this method to return the old ConnectionList
			// and add a new method "handleObsoleteConnections(ConnectionList oldList)" 
			// with the code below:
//			/* For all connections missing in the new list: notify listeners of removal */
//			for(int i=0; i<l.getNumConnection();i++) {
//			Connection c = l.getConnection(i);
//			boolean found = false;
//			for(int j=0; !found && j < universe.getConnections().getNumConnection(); j++) {
//			found = found | c.equals(universe.getConnections().getConnection(j));
//			}
//			treeChanged(NOTIFY_REMOVE,c);
//			}
		}
	}

	/* The handling of TreeChangeListeners */
	private static final int NOTIFY_ADD = 1;
	private static final int NOTIFY_REMOVE = 2;
	private static final int NOTIFY_CHANGE = 3;

//	private HashSet<DeviceTreeChangeListener> myChangeListeners = new HashSet<DeviceTreeChangeListener>();
	private LinearSet myChangeListeners = new LinearSet();
	public void addChangeListener(DeviceTreeChangeListener l) {
		myChangeListeners.add(l);
	}

	public void removeChangeListener(DeviceTreeChangeListener l) {
		myChangeListeners.remove(l);
	}

	/**
	 * Called by internal operations to notify changeListeners of a change to the tree
	 * @param change - the type of change (add, remove, or change)
	 * @param n - the changed node
	 */
	private void treeChanged(int change, NetworkNode n){
		for (Iterator iter = myChangeListeners.iterator(); iter.hasNext();) {
			DeviceTreeChangeListener l = (DeviceTreeChangeListener) iter.next();
			switch(change) {
			case NOTIFY_ADD:
				l.nodeAdded(n);
				break;
			case NOTIFY_REMOVE:
				l.nodeRemoved(n);
				break;
			case NOTIFY_CHANGE:
				l.nodeChanged(n);
				break;
			}
		}
	}
}
