/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import ist.palcom.collections.LinearMap;
import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.palcomthreads.PeriodicEvent;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.HeartBeat;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;

import java.util.ArrayList;

/**
 * AbstractMediaManager is the central class which hides details about different media
 * and protocols (such as UDP or IR). It is an abstract class. Concrete
 * subclasses implement medium/protocol specifics by instantiating proper
 * subclasses of CommunicationThread.
 * <p>
 * AbstractMediaManager handles two kinds of communication: unicast and broadcast (in
 * the case of UDP, "broadcast" means multicast). Subclasses implement the
 * method isBroadcast, which determines whether an address is unicast or
 * broadcast.
 * <p>
 * Communication between this class and the CommunicationThreads that do the
 * actual sending and receiving of messages is done through events. For a
 * message to be sent, the AbstractMediaManager posts a MessageArrayWithURLEvent in the
 * CommunicationThread's mailbox (putEvent). The MessageArrayWithURLEvent contains
 * the message and the URL to send it to. When a message is received, the
 * CommunicationThread posts a MessageArrayWithURLEvent in the mailbox of the
 * AbstractMediaManager's DistributorThread 
 * (the AbstractMediaManager is not itself a thread).
 * The URL in the MessageArrayWithURLEvent now contains the URL
 * <em>of the device which the message was received from.</em>. The DistributorThread handles
 * the distribution of the message to interested receivers.
 * <p>
 * The public methods of this class are used by clients. The protected abstract
 * methods are implemented by subclasses for supporting a particular combination
 * of protocols/media.
 * 
 * @see ist.palcom.mal.CommunicationThread CommunicationThread
 * @see ist.palcom.mal.udp.UDPMediaManager UDPMediaManager
 * 
 * @author David Svensson, david@cs.lth.se
 */
public abstract class AbstractMediaManager implements MediaManager {

	public static final String PALCOM_URL_PREFIX = "palcom:";

	/**
	 * The number of times a heartBeat (device) can time out before it is removed (during
	 * normal, periodic updating - after initiateDiscovery, heartBeat (devices) are 
	 * removed after the first timeout)
	 */
	private static final int TIMEOUTS_BEFORE_REMOVAL = 5;

	protected final PalcomScheduler scheduler;

	protected DistributorThread distributorThread;

	private LinearMap registeredDevices;

	private LinearMap unknownReceivers;

	private PalcomThread errorHandler;

	private PalcomThread communicationListenerThread;

	private ArrayList heartBeatReceivers;

	private String localDeviceCache = "0";

	private byte localStatusFlag = PRDData.FULLY_OPERATIONAL;

	private String localDeviceID = "urn:palcom://newDevice";

	public static final String WIRE_PROTOCOL_VERSION = "W2";

	private String localDiscoveryVersion = "D0";

	private Selector localDiscoverySelector = new Selector("1");
	
	private boolean active = false;

	private static final long DEFAULT_PERIOD = 2000;
	
	/** The time interval between heartbeats. */
	private long period;

	/**
	 * The time interval between sending a request and the timeout for concerned
	 * infos.
	 */
	private long timeoutDelay;

	private long nextHeartBeatTime;

	/** List of TimeoutNotices with timeouts for infos. */
	private LinkedList timeouts;

	/**
	 * Default AbstractMediaManager constructor. Should be overridden in subclasses
	 * where appropriate, to allow for instantiation using reflection mechanisms.
	 * 
	 * You may specify an errorHandler @see #addErrorHandler(PalcomPriorityThread)
	 * if you wish to know about any errors occuring in the startup phase of the
	 * AbstractMediaManager; If not, simply pass null.
	 */
	public AbstractMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler) {
		if (errorHandler != null) addErrorHandler(errorHandler);
		this.scheduler = scheduler;

		distributorThread = new DistributorThread();

		registeredDevices = new LinearMap();
		communicationListenerThread = null;
		heartBeatReceivers = new ArrayList();

		unknownReceivers = new LinearMap();

		period = DEFAULT_PERIOD;
		
		timeoutDelay = DEFAULT_PERIOD / 2;

		timeouts = new LinkedList();

	}

	/**
	 * Registers as a listener to messages received 
	 * by this manager from the network. 
	 * Currently only one listener at a time is supported.
	 * @param listener The PalcomThread that are 
	 * 		  interested in messages from the media manager.
	 */
	public void registerAsListener(PalcomThread listener) {
		//TODO clean up this method
		communicationListenerThread = listener;
		if (distributorThread.getScheduler() == null) {
			scheduler.scheduleThread(distributorThread);
		}
		URL[] urls = getDefaultGroupAddresses();
		for(int i = 0; i < urls.length; ++i) {
			URL protocolSpecificURL = getProtocolSpecificUrl(urls[i]);
			//Creates and starts the threads
			getReceivingThread(protocolSpecificURL);
		}
	}

	/** 
	 * Register as a listener for heartBeats. 
	 * This method is used by test programs and the RoutingManager. 
	 */
	public void registerForHeartBeats(PalcomThread listener) {
		heartBeatReceivers.add(listener);
	}
	
	/** Set the time interval between heartbeats. 
	 * This method has to be called in order for the device to send out heartbeats requesting answers,
	 * otherwise it will only send out heartbeats in reply to other devices.
	 */
	public void setPeriod(long period) {
		this.period = period;
		active = true;
	}

	public PalcomPriorityThread getDistributorThread(){
		return distributorThread;
	}

	/** Used in the same manner as setSendBufferSize. */
	public abstract void suggestSendBufferSize(int size);

	public abstract int getSendBufferSize();

	public void setLocalDeviceCache(String deviceCache) {
		localDeviceCache = deviceCache;
		// TODO Consider doing sendHeartBeat(true); on every device cache change
	}

	public String getLocalDeviceCache() {
		return localDeviceCache;
	}

	public String getLocalDeviceID() {
		return localDeviceID;
	}
	
	public void setLocalDeviceID(String deviceID) {
		localDeviceID = deviceID;
	}

	public void setLocalStatusFlag(byte statusFlag) {
		localStatusFlag = statusFlag;
	}

	public void setLocalDiscoveryVersion(String discoveryVersion) {
		localDiscoveryVersion = discoveryVersion;
	}

	public void setDiscoverySelector(Selector selector) {
		localDiscoverySelector = selector;
	}

	/**
	 * Send a message to a PalCom URL. The selector of the URL, if present, is
	 * put into the message as receiver selector. The sender selector of the
	 * message must be set before calling this method.
	 */
	private void send(Message message, URL receiverURL) {
		URL protocolSpecificURL = getProtocolSpecificUrl(receiverURL);
		if (protocolSpecificURL == null) {
			throw new IllegalArgumentException(receiverURL
					+ " is not a valid PalCom URL.");
		}

		CommunicationThread thread = getSendingThread(protocolSpecificURL);
		if (thread != null) {
			// Make a copy of the message, so it does not get modified further
			// in the mailbox queue
			Message copy = message.getCopy();
			byte[] copyArray = new byte[copy.getTotalLength()];
			copy.write(copyArray, 0);
			thread.putEvent(new MessageArrayWithURLEvent(copyArray, protocolSpecificURL));
		}
	}

	/**
	 * Send a message as a byte[] to a PalCom URL. The sender and receiver selector of the
	 * message must be set before calling this method if they are required for the message type.
	 */
	private void send(byte[] messageArray, URL receiverURL) {
		URL protocolSpecificURL = getProtocolSpecificUrl(receiverURL);
		if (protocolSpecificURL == null) {
			throw new IllegalArgumentException(receiverURL
					+ " is not a valid PalCom URL.");
		}

		CommunicationThread thread = getSendingThread(protocolSpecificURL);
		if (thread != null) {
			// Make a copy of the message, so it does not get modified further
			// in the mailbox queue
			byte[] copy = new byte[messageArray.length];
			System.arraycopy(messageArray, 0, copy, 0, messageArray.length);
			thread.putEvent(new MessageArrayWithURLEvent(copy, protocolSpecificURL));
		}
	}

	/** Send out a broadcast message. */
	private void sendBroadcast(byte[] messageArray) {
		URL[] broadcastURLs = getDefaultGroupAddresses();
		for (int i = 0; i < broadcastURLs.length; i++) {
			send(messageArray, broadcastURLs[i]);
		}
	}

	/** 
	 * Send out a broadcast HeartBeat message. 
	 * This method is used by the ResourceManager.
	 * Sending HeartBeatAck messages are handled automatically by the AbstractMediaManager.
	 * 
	 * @param isHeartBeatAck is the heartBeat an ack or 
	 * 		a request for others to answer with heartbeatAcks.
	 * 
	 * @param messageHeaderNodes used when for instance a mark node should be sent 
	 * 		with the heartBeat. 
	 * */
	public void sendHeartBeat(boolean isHeartBeatAck, LinkedList messageHeaderNodes) {
		MessageHeaderNode responseNode = null;
		if(!isHeartBeatAck) {
			responseNode = new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_HEARTBEAT, 
					new String[]{localDeviceCache, new String(new byte[]{localStatusFlag})});
		} else {
			responseNode = new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_HEARTBEATACK, 
					new String[]{localDeviceCache, new String(new byte[]{localStatusFlag})});
		}
		Message responseMessage = new Message();
		messageHeaderNodes.insertFirst(responseNode);
		responseMessage.setHeaderNodes(messageHeaderNodes);
		sendBroadcast(responseMessage);
	}

	/** 
	 * Send out a broadcast HeartAttack message. 
	 * This method is used by the DiscoveryManager and tells the receiving 
	 * devices that this devices is about to close down.
	 * */
	public void sendHeartAttack() {
		MessageHeaderNode responseNode = new MessageHeaderNode(
				MessageHeaderNode.NODE_FORMAT_HEARTATTACK);
		Message responseMessage = new Message();
		LinkedList headerNodes = new LinkedList();
		headerNodes.insertFirst(responseNode);
		responseMessage.setHeaderNodes(headerNodes);
		sendBroadcast(responseMessage);
	}

	/**
	 * Stop the manager. This will stop all listening for messages.
	 */
	public void stop() {
		distributorThread.terminate();
	}

	/**
	 * Subscribe to error events from this media manager.
	 * Unhandled errors within the AbstractMediaManager will be send to the handler
	 * in the form of {@link MALErrorEvent} events.
	 * Notice: Currently, only one error handler at a time is supported
	 */
	public void addErrorHandler(PalcomThread handler) {
		errorHandler = handler;
	}

	/**
	 * Send a message to a receiver at a given deviceID.
	 * @param message The message to send.
	 * @param receiverDeviceID The deviceID of the receiver.
	 * @return True if the AbstractMediaManager knows about the device, and has sent the
	 *         message.
	 */
	public boolean send(Message message, String receiverDeviceID) {
		DeviceCacheObject device = getRegisteredDevice(receiverDeviceID);
		if(device != null) {
			URL deviceURL = device.getDeviceURL();
			if (deviceURL != null) {
				URL receiverURL = new URL(PALCOM_URL_PREFIX + deviceURL.getString());
				//Routing message, insert RoutingR node.
				if(device.getShortID() != null) {
					message.insertHeaderNodeLast(new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_ROUTINGR, 
							new String[]{device.getShortID()}));
				}
				send(message, receiverURL);
				return true;
			} 
		}
		return false;
	}

	/** Send out a broadcast message, 
	 * with required message header nodes inserted before. */
	public void sendBroadcast(Message message) {

		URL[] broadcastURLs = getDefaultGroupAddresses();
		for (int i = 0; i < broadcastURLs.length; i++) {
			send(message, broadcastURLs[i]);
		}
	}

	/**
	 * This method is public so it can be used by test programs, but it is
	 * usually not called from other classes.
	 */
	public void registerDevice(String deviceID, DeviceCacheObject device) {
		registeredDevices.put(deviceID, device);
	}

	/**
	 * This method removes a device and its information 
	 * from the list with registered devices.
	 * @param deviceID The ID of the device to be removed.
	 * @return True if the device was removed from the list.
	 */
	public boolean unRegisterDevice(String deviceID) {
		DeviceCacheObject removedDevice = (DeviceCacheObject)registeredDevices.remove(deviceID);
		return removedDevice!=null;
	}

	private DeviceCacheObject getRegisteredDevice(String deviceID) {
		if(deviceID != null) {
			return (DeviceCacheObject) registeredDevices.get(deviceID);
		} else {
			return null;
		}
	}

	/** Get all devices registered at this media manager except itself. 
	 *  @return ArrayList with DeviceCacheObjects
	 */
	public ArrayList getRegisteredDevices() {
		ArrayList deviceValues = registeredDevices.values();
		for(int i = 0; i < deviceValues.size(); ++i) {
			DeviceCacheObject device = (DeviceCacheObject)deviceValues.get(i);
			if(device.getDeviceID().equals(localDeviceID) || 
					device.getDeviceCache() == null ||
					device.getDeviceStatus() == null) {
				deviceValues.remove(i);
				--i;
			}
		}
		return deviceValues;
	}

	/**
	 * Checks if the device with ID deviceID is known to the media manager.
	 * @param deviceID  must contain deviceID but nothing more.
	 * @return true if device is registered
	 */
	public boolean isDeviceRegistered(String deviceID) {
		if(deviceID != null) {
			return registeredDevices.containsKey(deviceID);
		}
		return false;
	}

	private boolean isDeviceRegistered(DeviceCacheObject device) {
		ArrayList devices = registeredDevices.values();
		return devices.contains(device);
	}

	private void updateDeviceInfo(String deviceID, String deviceCache, String deviceStatus) {
		DeviceCacheObject device = getRegisteredDevice(deviceID);
		if(device != null) {
			device.updateDeviceInfo(deviceCache, deviceStatus);
			registeredDevices.put(deviceID, device);
		}
	}

	private DeviceCacheObject getRegisteredDevice(URL deviceURL, String shortID) {
		deviceURL = getProtocolSpecificUrl(deviceURL);
		ArrayList devices = registeredDevices.values();
		DeviceCacheObject lookupObject = new DeviceCacheObject(deviceURL, shortID);
		int index = devices.indexOf(lookupObject);
		if(index >= 0) {
			return (DeviceCacheObject)devices.get(index);
		} else {
			return null;
		}
	}

	/**
	 * Returns the URL that the manager listens for messages at by default.
	 * 
	 * This method has been made public so it can be used from test programs,
	 * but it is not otherwise used by classes other than AbstractMediaManager and its
	 * subclasses.
	 * 
	 * @return The default URL. This URL is listened on unless another URL is
	 *         specified in a call to startReceiving(String, PalcomThread).
	 */
	public abstract URL getDefaultUnicastURL();

	/**
	 * @return CommunicationThread for receiving at url. If it does not exist, a
	 *         new one is created and returned.
	 */
	protected abstract CommunicationThread getReceivingThread(URL url);

	/** Stop receiving at url, and remove the receiving thread. */
	protected abstract void removeReceivingThread(URL url);

	/**
	 * @return A CommunicationThread that can be used for sending a message to
	 *         the given URL.
	 */
	protected abstract CommunicationThread getSendingThread(URL url);

	/**
	 * @return true if the URL is a broadcast address (or multicast, depending
	 *         on the protocol), false if not.
	 */
	protected abstract boolean isBroadcast(URL url);

	/**
	 * This method is called every time a message is received. It may be
	 * overridden by subclasses that want to do more than just let the message
	 * be distributed to registered receivers.
	 */
	protected void messageReceived(Message message) {
	}

	/**
	 * @return The default broadcast addresses for group communication.
	 */
	public abstract URL[] getDefaultGroupAddresses();

	/**
	 * When unhandled errors occur, concrete implementations of this class should
	 * call this method for notifying error handlers about the events.
	 */
	protected void raiseErrorEvent(MALErrorEvent mee) {
		if (mee.getSource() == null) mee.setSource(this);
		if (errorHandler != null) {
			errorHandler.putEvent(mee);
		} else {
			System.err.println("MediaManager error: " + mee);
			mee.getError().printStackTrace();
		}
	}

	public static String getProtocol(URL url) {
		if (url == null) {
			throw new NullPointerException();
		}
		url = getProtocolSpecificUrl(url);
		int firstColonIndex = url.getString().indexOf(":");
		return url.getString().substring(0, firstColonIndex);
	}

	public static String getHost(URL url) {
		if (url == null) {
			throw new NullPointerException();
		}
		url = getProtocolSpecificUrl(url);
		int firstColonIndex = url.getString().indexOf(":");
		String threeChars = url.getString().substring(firstColonIndex,
				firstColonIndex + 3);
		if (!threeChars.equals("://")) {
			throw new IllegalArgumentException(url + " is not a valid URL");
		}
		int secondColonIndex = url.getString()
		.indexOf(":", firstColonIndex + 1);
		return url.getString().substring(firstColonIndex + 3, secondColonIndex);
	}

	/**
	 * @return The port of the protocol-specific part of the URL, or -1 if not
	 *         specified.
	 */
	public static int getPort(URL url) {
		url = getProtocolSpecificUrl(url);
		int firstColonIndex = url.getString().indexOf(":");
		int secondColonIndex = url.getString()
		.indexOf(":", firstColonIndex + 1);

		if (secondColonIndex == -1) {
			return -1;
		} else {
			String portString = url.getString().substring(secondColonIndex + 1,
					url.getString().length());
			return Integer.parseInt(portString);
		}
	}

	/**
	 * @return The protocol-specific part of the URL (the URL that can be used
	 *         for addressing the device).
	 */
	public static URL getProtocolSpecificUrl(URL url) {
		if (url == null) {
			throw new NullPointerException();
		}
		if (!url.getString().startsWith(PALCOM_URL_PREFIX)) {
			return url; // Considered a protocol-specific URL
		}
		int startIndex = PALCOM_URL_PREFIX.length();

		int colonSlashSlashIndex = url.getString().indexOf("://");
		if (colonSlashSlashIndex == -1) {
			throw new IllegalArgumentException("Wrong URL format");
		}
		int slashIndex = url.getString().indexOf("/", colonSlashSlashIndex + 3);

		if (slashIndex == -1) {
			return new URL(url.getString().substring(startIndex,
					url.getString().length()));
		} else {
			return new URL(url.getString().substring(startIndex, slashIndex));
		}
	}

	/** @return Selector part of the URL, or null if it is not set. */
	public static Selector getSelector(URL url) {
		if (url == null) {
			throw new NullPointerException();
		}

		if (!url.getString().startsWith(PALCOM_URL_PREFIX)) {
			return null;
		}

		int colonSlashSlashIndex = url.getString().indexOf("://");
		if (colonSlashSlashIndex == -1) {
			return null;
		}
		int slashIndex = url.getString().indexOf("/", colonSlashSlashIndex + 3);

		if (slashIndex == -1) {
			return null;
		} else {
			String selectorString = url.getString().substring(slashIndex + 1,
					url.getString().length());
			return new Selector(selectorString);
		}
	}

	private void setNewHeartBeatTime() {
		nextHeartBeatTime = ist.palcom.base.System.instance().time()
		+ period;
	}

	private long getNextHeartBeatTime() {
		return nextHeartBeatTime;
	}

	/** @return The TimeoutNotice for heartBeat, or null if not set. */
	private TimeoutNotice getTimeout(String deviceID) {
		TimeoutNotice notice = (TimeoutNotice) timeouts.getFirst();
		while (notice != null) {
			if(!notice.deviceID.equals(deviceID)) {
				notice = (TimeoutNotice) notice.getNext();
			} else {
				break;
			}
		}
		return notice;
	}

	private void setTimeout(String deviceID, long timeout, int timeoutsLeft) {
		TimeoutNotice notice = getTimeout(deviceID);
		if (notice == null) {
			timeouts.insertLast(new TimeoutNotice(deviceID, timeout, timeoutsLeft));
		} else {
			notice.nextTimeout = timeout;
			notice.timeoutsLeft = timeoutsLeft;
		}
	}

	private void removeTimeout(String deviceID) {
		TimeoutNotice notice = getTimeout(deviceID);
		if (notice != null) {
			timeouts.remove(notice);
		}
	}

	/**
	 * Keeps track of the next timeout for a heartBeat, and the number of timeouts
	 * left for it.
	 */
	private class TimeoutNotice extends Link {
		String deviceID;

		long nextTimeout;

		int timeoutsLeft;

		TimeoutNotice(String deviceID, long nextTimeout, int timeoutsLeft) {
			this.deviceID = deviceID;
			this.nextTimeout = nextTimeout;
			this.timeoutsLeft = timeoutsLeft;
		}
	}

	/** Thread for receiving MessageArrayWithURLEvents and distributing them 
	 * to interested listeners. It also handles the intial discovery of other 
	 * devices. */
	public class DistributorThread extends PalcomPriorityThread {
		public void run() {

			long threadPeriod = period / 20;
			enablePeriodicEvents((int) threadPeriod);
			boolean firstPeriod = true;
			while (true) {
				Event event = waitEvent();

				if (event instanceof PeriodicEvent) {
					long now = ist.palcom.base.System.instance().time();

					if (now > getNextHeartBeatTime() && active && !firstPeriod) {
						sendHeartBeat(false, new LinkedList());
						setNewHeartBeatTime();
					}
					
					if (firstPeriod) {
						sendHeartBeat(!active, new LinkedList());
						firstPeriod = false;
					}

					ArrayList devices = registeredDevices.values();
					for(int index = 0; index < devices.size(); ++index) {
						DeviceCacheObject device = (DeviceCacheObject)devices.get(index);
						String deviceID = device.getDeviceID();

						TimeoutNotice timeoutNotice = getTimeout(deviceID);
						if (timeoutNotice != null
								&& now > timeoutNotice.nextTimeout) {
							if (timeoutNotice.timeoutsLeft <= 1) {
								removeTimeout(deviceID);
								unRegisterDevice(deviceID);
								for(int i = 0; i < heartBeatReceivers.size(); ++i) {
									((PalcomThread)heartBeatReceivers.get(i)).putEvent(
											new HeartAttackEvent(deviceID, AbstractMediaManager.this, new ArrayList()));
								}
							} else {
								setTimeout(deviceID, timeoutNotice.nextTimeout
										+ period,
										timeoutNotice.timeoutsLeft - 1);
							}
						}
					}

				} else if (event instanceof MessageArrayWithURLEvent) {
					MessageArrayWithURLEvent mawue = (MessageArrayWithURLEvent) event;
					byte[] messageArray = mawue.getMessageArray();
					URL senderURL = mawue.getDeviceURL();

					//check what type of message it is and then deside what to do with it
					Message message = processMessageEvent(messageArray, senderURL);

					if(message != null) {
						handleProcessedMessage(senderURL, message);
					}

//					TODO remove this else if block when all subManagers and 
					//communicationThreads have stopped using MessageWithURLEvent.
				} else if (event instanceof MessageWithURLEvent) {
					System.out.println("DistributorThread in MediaManager received MessageWithURLEvent, " +
					"switch to MessageArrayWithURLEvent.");
					MessageWithURLEvent mawue = (MessageWithURLEvent) event;
					Message message = mawue.getMessage();
					byte[] messageArray = new byte[message.getTotalLength()];
					message.write(messageArray, 0);
					URL senderURL = mawue.getURL();

					//check what type of message it is and then deside what to do with it
					message = processMessageEvent(messageArray, senderURL);

					if(message != null) {
						handleProcessedMessage(senderURL, message);
					}
				} else if (event instanceof MALErrorEvent) {
					MALErrorEvent mee = (MALErrorEvent) event;
					raiseErrorEvent(mee);
				}
			}
		}

		/**
		 * Checks if the message was a new device registering and in that case if
		 * there are any previously saved messages from when this new device was unknown.
		 * @param senderURL The url from where the message came from
		 * @param message The received message
		 */
		private void handleProcessedMessage(URL senderURL, Message message) {
			MessageHeaderNode hNode = (MessageHeaderNode)message.getHeaderNodes().getFirst();

			//Check if this new device has any previous messages saved from 
			//when it was unknown due to race conditions. Otherwise distribute it.
			if(hNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_HBINFOREPLY){
				if(unknownReceivers.containsKey(senderURL)) {
					ArrayList list = (ArrayList) unknownReceivers.get(senderURL);
					for(int i = 0; i < list.size(); ++i) {
						Message savedMessage = (Message)list.get(i);

						MessageHeaderNode searchNode = (MessageHeaderNode) 
						savedMessage.getHeaderNodes().getFirst();
						while(searchNode != null && 
								searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGS) {
							searchNode = (MessageHeaderNode) searchNode.getNext();
						}
						String shortID = null;
						if(searchNode!=null) {
							shortID = searchNode.getData()[0];
						}

						DeviceCacheObject device = getRegisteredDevice(senderURL, shortID);
						if(device != null) {
							savedMessage.setSenderDeviceID(device.getDeviceID());
							distribute(savedMessage, (MessageHeaderNode) savedMessage.getHeaderNodes().getFirst());
							list.remove(i);
						}
					}
					if(list.isEmpty()) {
						unknownReceivers.remove(senderURL);
					}
				}
				//If it is a message to be routed, distribute it
				MessageHeaderNode searchNode = (MessageHeaderNode) 
				message.getHeaderNodes().getFirst();
				while(searchNode != null && 
						searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGR) {
					searchNode = (MessageHeaderNode) searchNode.getNext();
				}
				if(searchNode!=null) {
					distribute(message, (MessageHeaderNode) message.getHeaderNodes().getFirst());
				}
			} else {
				distribute(message, hNode);
			}
		}

		/** Send the message to interested listeners. */ 
		private void distribute(Message message, MessageHeaderNode hNode) {
			messageReceived(message);
//			Send message to the registered listener.
			if(communicationListenerThread != null && hNode != null
					&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_HEARTBEAT
					&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_HEARTBEATACK
					&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_HEARTATTACK) {
				communicationListenerThread.putEvent(new MessageEvent(message, AbstractMediaManager.this));
			} else if(!heartBeatReceivers.isEmpty() && hNode != null) {
				ArrayList marks = new ArrayList();
				MessageHeaderNode searchNode = (MessageHeaderNode) message.getHeaderNodes().getFirst();
				while(searchNode != null) {
					if(searchNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_MARK) {
						marks.add(searchNode.getData());
					}
					searchNode = (MessageHeaderNode) searchNode.getNext();
				}
				if(hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_HEARTATTACK) {
					MessageHeaderNode versionNode = (MessageHeaderNode) hNode.getNext();
					HeartBeat heartBeat = new HeartBeat(new DeviceID(message.getSenderDeviceID()), 
							hNode.getFormatIdentifier()==MessageHeaderNode.NODE_FORMAT_HEARTBEATACK, 
							hNode.getData()[0], versionNode.getData()[1], hNode.getData()[1].getBytes()[0], 
							message.getSenderSelector());
					for(int i = 0; i < heartBeatReceivers.size(); ++i) {
						((PalcomThread)heartBeatReceivers.get(i)).putEvent(new HeartBeatEvent(heartBeat, AbstractMediaManager.this, marks));
					}
				} else {
					for(int i = 0; i < heartBeatReceivers.size(); ++i) {
						((PalcomThread)heartBeatReceivers.get(i)).putEvent(new HeartAttackEvent(message.getSenderDeviceID(), AbstractMediaManager.this, marks));
					}
				}
			}
		}

		/**
		 * Processes a MessageArrayWithURLEvent by unwrapping some of the contents.
		 * @param messageArray 
		 * 				The array containing the entire message
		 * @param senderURL
		 * 				The address from where the message came.
		 */
		private Message processMessageEvent(byte[] messageArray, URL senderURL) {
			Message message = Message.parse(messageArray, 0, messageArray.length);
			LinkedList headerNodes = message.getHeaderNodes();
			MessageHeaderNode hNode = new MessageHeaderNode(messageArray, 0, messageArray.length);
//			Message message = new Message(hNode, messageArray);
			byte formatIdentifier = hNode.getFormatIdentifier();

			LinkedList responseList = new LinkedList();
			Message responseMessage = new Message();
			String shortID = null;

			//Look for header node RoutingS and if found insert RoutingR node in possible response message.
			MessageHeaderNode searchNode = (MessageHeaderNode) headerNodes.getFirst();
			while(searchNode != null && 
					searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGS) {
				searchNode = (MessageHeaderNode) searchNode.getNext();
			}
			if(searchNode!=null) {
				if(!senderURL.equals(getDefaultUnicastURL())) {
					shortID = searchNode.getData()[0];
					responseList.insertLast(new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_ROUTINGR, searchNode.getData()));
				} else {
					return null;
				}
			}

			boolean routingRExists = false;
			//Look for header node RoutingR
			searchNode = (MessageHeaderNode) headerNodes.getFirst();
			while(searchNode != null && 
					searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGR) {
				searchNode = (MessageHeaderNode) searchNode.getNext();
			}
			if(searchNode!=null) {
				routingRExists = true;
			}

			//Set deviceID if the device is registered.
			if(isDeviceRegistered(new DeviceCacheObject(getProtocolSpecificUrl(senderURL), shortID))) {
				DeviceCacheObject device = getRegisteredDevice(senderURL, shortID);
				message.setSenderDeviceID(device.getDeviceID());
				message.setSenderSelector(device.getDiscoverySelector());
			} else {
				if(formatIdentifier != MessageHeaderNode.NODE_FORMAT_HEARTBEAT && 
						formatIdentifier != MessageHeaderNode.NODE_FORMAT_HEARTBEATACK &&
						formatIdentifier != MessageHeaderNode.NODE_FORMAT_HBINFOREPLY &&
						formatIdentifier != MessageHeaderNode.NODE_FORMAT_HBINFOREQUEST) {
					//Save message from unknown host until it can be identified. (Result of race conditions)
					if(!unknownReceivers.containsKey(senderURL)) {
						ArrayList list = new ArrayList();
						list.add(message);
						unknownReceivers.put(senderURL, list);
					} else {
						((ArrayList) unknownReceivers.get(senderURL)).add(message);
					}
					return null;
					//Can not route message from unknown host.
				} else if(routingRExists) {
					return null;
				}
			}

			//HeartBeat(Ack) message
			if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_HEARTBEAT || 
					formatIdentifier == MessageHeaderNode.NODE_FORMAT_HEARTBEATACK) {

				//New device found? If so send HBInfoRequest asking for more info.
				if(!isDeviceRegistered(new DeviceCacheObject(getProtocolSpecificUrl(senderURL), shortID))) {
					responseList.insertFirst(new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_HBINFOREQUEST));
					responseMessage.setHeaderNodes(responseList);
					send(responseMessage, senderURL);
					message = null; //We don't want to send this message any further before 
					//we know more about the device.
				} else {
					DeviceCacheObject device = getRegisteredDevice(senderURL, shortID);
					message.insertHeaderNodeAfter(new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_VERSION, 
							new String[]{device.getWireVersion(), device.getDiscoveryVersion()}), 
							(MessageHeaderNode) headerNodes.getFirst());

					updateDeviceInfo(device.getDeviceID(), hNode.getData()[0], hNode.getData()[1]);

					setTimeout(device.getDeviceID(),
							getNextHeartBeatTime()
							+ timeoutDelay,
							TIMEOUTS_BEFORE_REMOVAL);
				}
				if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_HEARTBEAT) {
					setNewHeartBeatTime();
					//If heartBeat, send a response to the requesting device.
					MessageHeaderNode response = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_HEARTBEATACK, 
							new String[]{localDeviceCache, new String(new byte[]{localStatusFlag})});
					sendBroadcast(response.toByteArray());
				}
			//HeartAttack message
			} else if (formatIdentifier == MessageHeaderNode.NODE_FORMAT_HEARTATTACK) {

				String deviceID = getRegisteredDevice(getProtocolSpecificUrl(senderURL), shortID).getDeviceID();
				if(deviceID != null) {
					unRegisterDevice(deviceID);
					removeTimeout(deviceID);
				}
				unknownReceivers.remove(senderURL);

				//HBInfoRequest message
			} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_HBINFOREQUEST) {

				if(!routingRExists) {
					//Send a HBInfoReply to the device that requested it containing the deviceID 
					//and DiscoverySelector used on this device.
					MessageHeaderNode hbInfoReplyMessage = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_HBINFOREPLY, 
							new String[]{localDeviceID, localDiscoverySelector.toString()});
					MessageHeaderNode versionMessage = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_VERSION, 
							new String[]{WIRE_PROTOCOL_VERSION, localDiscoveryVersion});
					//TODO remove routingR
					MessageHeaderNode routingNode = (MessageHeaderNode)responseList.getLast();
					String routingR = "";
					if(routingNode != null) {
						routingR = routingNode.toString();
					}
					byte[] responseArray = (hbInfoReplyMessage.toString() + versionMessage.toString() + 
							routingR).getBytes();
					send(responseArray, senderURL);
					message = null; //Message should not be sent any further up.
				}
				//HBInfoReply message
			} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_HBINFOREPLY) {

				String[] data = hNode.getData();
				if(data.length == 2) {
					String deviceID = data[0]; //aka senderDeviceID
					message.setSenderDeviceID(deviceID);
					Selector discoverySelector = new Selector(data[1]);
					//check next header node for version information
					int offset = hNode.getTotalLength();
					MessageHeaderNode versionNode = new MessageHeaderNode(messageArray, 
							offset, messageArray.length);
					if(versionNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_VERSION) {
						if(versionNode.getData().length==2) {
							//Add a mapping for the device URL
							URL senderDeviceURL = getProtocolSpecificUrl(senderURL);
							DeviceCacheObject device = new DeviceCacheObject(
									senderDeviceURL, deviceID, versionNode.getData()[0], 
									versionNode.getData()[1], discoverySelector);
							if(shortID != null) {
								device.setShortID(shortID);
							}
							registerDevice(deviceID, device);
						} else {
							//TODO fix errorHandler? Syntax error in the message header
							System.out.println("Error in message header node syntax!");
						}
					} else {
						//TODO fix errorHandler? Syntax error in the message, missing version node
						System.out.println("Error in message header node syntax!");
					}
				} else {
					//TODO fix errorHandler? Syntax error in the message header
					System.out.println("Error in message header node syntax!");
				}


			} //else Other message that can't be handled in this layer.

			//Only messages that support the same wire protocol version is accepted.
			if(message != null && isDeviceRegistered(message.getSenderDeviceID())) {
				DeviceCacheObject device = getRegisteredDevice(message.getSenderDeviceID());
				if(device!= null && !WIRE_PROTOCOL_VERSION.equals(device.getWireVersion())) {
//					TODO An event should be sent to upper layers announcing this information
					System.out.println("Device " + message.getSenderDeviceID() + 
					" is not using a compatible wire version.");
					message = null;
				}
			}
			return message;
		}
	}
}
