/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class implements a media manager that can have other media managers as
 * submanagers, handling the actual communication.
 * <p>
 * The structure is that of the Composite design pattern - a
 * CompoundMediaManager has the same interface as "ordinary" MediaManagers, and
 * can be used wherever a MediaManager is needed.
 * <p>
 * Submanagers are added using the {@link #add(String, MediaManager) add}
 * method.
 * 
 * @see ist.palcom.mal.MediaManager MediaManager
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class CompoundMediaManager extends AbstractMediaManager {

    private ArrayList protocols;

    private ArrayList subManagers;

    private URL[] defaultGroupAddresses;

    private PalcomPriorityThread errorHandler;

    public CompoundMediaManager(PalcomScheduler scheduler,
            String defaultProtocol, MediaManager defaultManager, PalcomThread errorHandler) {
        super(scheduler, errorHandler);
        protocols = new ArrayList();
        subManagers = new ArrayList();

        protocols.add(defaultProtocol);
        subManagers.add(defaultManager);
    }
    
//  TODO: this needs to be implemented
	public void suggestSendBufferSize(int size) {
		
	}
	
//	TODO: this needs to be implemented
	public int getSendBufferSize() {
		return -1;
	}

    public boolean send(Message message, String senderDeviceID,
            Selector senderSelector, String receiverDeviceID, Selector receiverSelector) {
        boolean sent = false;
        int index = 0;
        while (!sent && index < subManagers.size()) {
            MediaManager manager = (MediaManager) subManagers.get(index);
            if (manager.send(message, receiverDeviceID)) {
                return true;
            }
            index++;
        }
        return false;
    }

    public void sendBroadcast(Message message, String senderDeviceID,
            Selector senderSelector) {
        for (int i = 0; i < subManagers.size(); i++) {
            MediaManager manager = (MediaManager) subManagers.get(i);
            manager.sendBroadcast(message);
        }
    }

    /**
     * Adds the error handler to all registered managers, and cached for further
     * reference
     */
    public void addErrorHandler(PalcomPriorityThread handler) {
        errorHandler = handler;
        for (int i = 0; i < subManagers.size(); i++) {
            MediaManager manager = (MediaManager) subManagers.get(i);
            manager.addErrorHandler(handler);
        }
    }

    public void stop() {
        for (int i = 0; i < subManagers.size(); i++) {
            MediaManager manager = (MediaManager) subManagers.get(i);
            manager.stop();
        }
        protocols.clear();
        subManagers.clear();
        defaultGroupAddresses = null;

        super.stop();
    }

    /**
     * Add a manager handling a specific protocol, The protocol is the protocol
     * part of the URLs, such as "datagram" for UDP or "http" for HTTP.
     */
    public void add(String protocol, MediaManager manager) {
        protocols.add(protocol);
        subManagers.add(manager);
        defaultGroupAddresses = null; // Recompute next time
        if (errorHandler != null)
            manager.addErrorHandler(errorHandler);
    }

    public URL getDefaultUnicastURL() {
        return getDefaultManager().getDefaultUnicastURL();
    }

    protected CommunicationThread getReceivingThread(URL url) {
        return getManager(url).getReceivingThread(url);
    }

    protected void removeReceivingThread(URL url) {
        getManager(url).removeReceivingThread(url);
    }

    protected CommunicationThread getSendingThread(URL url) {
        return getManager(url).getSendingThread(url);
    }

    protected boolean isBroadcast(URL url) {
        return getManager(url).isBroadcast(url);
    }

    public URL[] getDefaultGroupAddresses() {
        if (defaultGroupAddresses == null) {
            // Need to (re)compute it
            ArrayList list = new ArrayList();
            Iterator subManagerIterator = subManagers.iterator();
            while (subManagerIterator.hasNext()) {
                AbstractMediaManager subManager = (AbstractMediaManager) subManagerIterator
                        .next();
                URL[] addresses = subManager.getDefaultGroupAddresses();
                for (int i = 0; i < addresses.length; i++) {
                    list.add(addresses[i]);
                }
            }
            defaultGroupAddresses = new URL[list.size()];
            for (int i = 0; i < list.size(); i++) {
                defaultGroupAddresses[i] = (URL) list.get(i);
            }
        }
        return defaultGroupAddresses;
    }

    private AbstractMediaManager getDefaultManager() {
        return (AbstractMediaManager) subManagers.get(0);
    }

    private AbstractMediaManager getManager(URL url) {
        String protocol = getProtocol(url);
        int index = protocols.indexOf(protocol);
        if (index == -1) {
            System.err.println("URL " + url + " not supported.");
            return null;
        } else {
            return (AbstractMediaManager) subManagers.get(index);
        }
    }
}
