/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import java.util.ArrayList;

import ist.palcom.collections.LinkedList;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;

/** Interface for implementing a media manager. 
 * @author Thomas Forsstrm, thomasf@cs.lth.se 
 * */
public interface MediaManager {

	/**
	 * Registers as a listener to messages received 
	 * by this manager from the network. 
	 * Currently only one listener at a time is supported.
	 * @param listener The PalcomThread that are 
	 * 		  interested in messages from the media manager.
	 */
	public void registerAsListener(PalcomThread listener);
	
	/** 
	 * Register as a listener for heartbeats and heartattacks.  
	 */
	public void registerForHeartBeats(PalcomThread listener);
	
	public PalcomPriorityThread getDistributorThread();
	
	/** Used in the same manner as setSendBufferSize. */
	public abstract void suggestSendBufferSize(int size);
	
	public abstract int getSendBufferSize();
	public void setLocalDeviceCache(String deviceCache);
	public String getLocalDeviceCache();
	public String getLocalDeviceID();
	public void setLocalDeviceID(String deviceID);
	public void setLocalStatusFlag(byte statusFlag);
	public void setLocalDiscoveryVersion(String discoveryVersion);
	public void setDiscoverySelector(Selector selector);
	
	/** 
	 * Send out a broadcast HeartBeat message. 
	 * This method is used by the ResourceManager.
	 * Sending HeartBeatAck messages are handled automatically by the AbstractMediaManager.
	 * 
	 * @param isHeartBeatAck is the heartBeat an ack or 
	 * 		a request for others to answer with heartbeatAcks.
	 * 
	 * @param messageHeaderNodes used when for instance a mark node should be sent 
	 * 		with the heartBeat. 
	 * */
	public void sendHeartBeat(boolean isHeartBeatAck, LinkedList messageHeaderNodes);
	
	/** 
	 * Send out a broadcast HeartAttack message. 
	 * This method is used by the DiscoveryManager and tells the receiving 
	 * devices that this devices is about to close down.
	 * */
	public void sendHeartAttack();
	
	/**
	 * Stop the manager. This will stop all listening for messages.
	 */
	public void stop();
	
	/**
	 * Subscribe to error events from this media manager.
	 * Unhandled errors within the MediaManager will be send to the handler
	 * in the form of {@link MALErrorEvent} events.
	 * Notice: Currently, only one error handler at a time is supported
	 */
	public void addErrorHandler(PalcomThread handler);
	
	/**
	 * Send a message to a receiver at a given URN.
	 * @param message The message to send.
	 * @param receiverURN The URN of the receiver.
	 * @return True if the MediaManager knows about the URN, and has sent the
	 *         message.
	 */
	public boolean send(Message message, String receiverDeviceID);
	
	/** Send out a broadcast message, 
	 * with required message header nodes inserted before. */
	public void sendBroadcast(Message message);
	
	/**
	 * This method is public so it can be used by test programs, but it is
	 * usually not called from other classes.
	 */
	public void registerDevice(String deviceID, DeviceCacheObject device);
	
	/**
	 * This method removes a device and its information 
	 * from the list with registered devices.
	 * @param urn The urn of the device to be removed.
	 * @return True if the device was removed from the list.
	 */
	public boolean unRegisterDevice(String deviceID);
	
	/**
	 * Checks if the device with URN deviceURN is known to the media manager.
	 * @param deviceURN  must contain URN but nothing more.
	 * @return true if device is registered
	 */
	public boolean isDeviceRegistered(String deviceID);
	
	/** Get all devices registered at this media manager except itself. 
	 *  @return ArrayList with DeviceCacheObjects
	 */
	public ArrayList getRegisteredDevices(); 
	
	/** Set the time interval between heartbeats. */
	public void setPeriod(long period);
}
