/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;

/**
 * Message implement the common message format.
 * 
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class Message {

	/** Special value used when there is no data */
	public static final byte TYPE_NONE = -1;

	public static final byte TYPE_SINGLE = 'd';

	public static final byte TYPE_MULTIPART = '+';

	private byte dataType;

	private String senderDeviceID; //deviceID in some cases

	/** Selector of sender. */
	private Selector senderSelector; //discoverySelector in some cases

	private String receiverDeviceID;

	/** Selector of receiver. */
	private Selector receiverSelector;
	
	/**
	 * URL of the sender of this message, if it has been received over the
	 * network.
	 */
	private URL senderURL;

	/**
	 * URL of the receiver of this message, if it has been received over the
	 * network.
	 */
	private URL receiverURL;

	/** Data array used when data type is TYPE_SINGLE. */
	private byte[] data;

	/** Parts list used when data type is TYPE_MULTIPART. */
	private LinkedList parts = new LinkedList();

	/** Nodes list for the format idenfifiers in the message*/
	private LinkedList headerNodes = new LinkedList();

	/** Data array for the unhandled data. Should always be the entire message as a byte array if used*/
	private byte[] rawData;

	/*protected Message() {
        senderURN = null;
        senderSelector = null;
        receiverURN = null;
        receiverSelector = null;
        senderURL = null;
        receiverURL = null;
    }*/

	public Message() {
		//connectionFunction = FUNCTION_NONE;
		dataType = TYPE_NONE;
		rawData = new byte[0];
	}

	/**Create a message of type TYPE_NONE 
	 * containing a message header node and 
	 * a byte array with all the data unhandled.
	 */
	public Message(MessageHeaderNode node, byte[] rawData) {
		dataType = TYPE_NONE;
		headerNodes.insertLast(node);
		this.rawData = rawData;
	}

	/**Create a message of type TYPE_NONE containing only a message header node.
	 */
	public Message(MessageHeaderNode node) {
		dataType = TYPE_NONE;
		headerNodes.insertLast(node);
		rawData = node.toByteArray();
	}

	/** Create a message of TYPE_SINGLE with the given data. */
	public Message(byte[] data) {
		this(TYPE_SINGLE, data);
	}

	/** Create a message of TYPE_SINGLE with the given data. */
	public Message(byte[] data, int offset, int dataLength) {
		this(TYPE_SINGLE, data, offset, dataLength);
	}

	public Message(byte dataType) {
		this(dataType, new byte[0]);
	}

	public Message(byte dataType, byte[] data) {
		this(dataType, data, 0, data.length);
	}

	public Message(byte dataType, byte[] data, int offset, int dataLength) {
		this.dataType = dataType;
		if (dataType == TYPE_SINGLE) {
			this.data = new byte[dataLength];
			System.arraycopy(data, offset, this.data, 0, dataLength);
		} else if (dataType == TYPE_MULTIPART) {
			int bytesRead = 0;
			while (bytesRead < dataLength) {
				Message part = parseWithOutHeaderNodes(data, offset + bytesRead, dataLength
						- bytesRead);
				addPart(part);
				bytesRead += part.getTotalLength();
			}
		} else {
			throw new RuntimeException(
					"Can only create message with data using this constructor.");
		}
		// connectionFunction = FUNCTION_NONE;
	}

	/** Returns the length of the data in the message (zero if there is no data). */
	public int getDataLength() {
		if (dataType == TYPE_SINGLE) {
			return data.length;
		} else if (dataType == TYPE_MULTIPART) {
			int length = 0;
			PartNotice notice = (PartNotice) parts.getFirst();
			while (notice != null) {
				length += notice.message.getTotalLength();
				notice = (PartNotice) notice.getNext();
			}
			return length;
		} else {
			return 0;
		}
	}

	/** Returns the length of the data including the header part in the message. */
	public int getNodeAndDataLength() {
		return getNodeLength(getDataLength());
	}

	/**
	 * Returns the total length of the message, including the length of the
	 * header with length and type.
	 */
	public int getTotalLength() {
		int length = 0;

		length += getHeaderNodesLength();

		if (dataType != TYPE_NONE) {
			length += getNodeLength(getDataLength());
		} else {
			if(rawData != null && rawData.length > length) {
				length = rawData.length;
			}
		}
		return length;
	}

	/**
	 * Get the length of all the current header nodes in the message.
	 * This is the same as the offset on the array with raw data that has been handled.
	 */
	public int getHeaderNodesLength() {
		int length = 0;
		MessageHeaderNode hNode = (MessageHeaderNode) headerNodes.getFirst();
		while (hNode != null) {
			length += hNode.toString().length();
			hNode = (MessageHeaderNode) hNode.getNext();
		}
		return length;
	}

	/**
	 * @return The length of a node when printed out.
	 * @param dataLength
	 *            The length of the data in the node
	 */
	private int getNodeLength(int dataLength) {
		int length = 3; // Format identifier and two semicolons
		length += getIntegerString(dataLength).length();
		length += dataLength;
		return length;
	}

	/**
	 * Returns a String with the sender deviceID, formatted as when written to a
	 * stream or buffer.
	 */
	private String getSenderDeviceIDString() {
		if (senderDeviceID == null) {
			return "";
		} else {
			return senderDeviceID;
		}
	}

	/**
	 * Returns a String with the sender selector, formatted as when written to a
	 * stream or buffer.
	 */
	private String getSenderSelectorString() {
		if (senderSelector == null) {
			return "";
		} else {
			return senderSelector.getString();
		}
	}

	/**
	 * Returns a String with the receiver deviceID, formatted as when written to a
	 * stream or buffer.
	 */
	private String getReceiverURNString() {
		if (receiverDeviceID == null) {
			return "";
		} else {
			return receiverDeviceID;
		}
	}

	/**
	 * Returns a String with the receiver selector, formatted as when written to
	 * a stream or buffer.
	 */
	private String getReceiverSelectorString() {
		if (receiverSelector == null) {
			return "";
		} else {
			return receiverSelector.getString();
		}
	}

	/**
	 * Returns a String with the integer value, formatted as when written to a
	 * stream or buffer.
	 */
	private String getIntegerString(int value) {
		return Integer.toString(value);
	}

	/** Returns the data type of the message. */
	public byte getType() {
		return dataType;
	}

	/**
	 * @return The data of the message (null if the data type is not of
	 *         TYPE_SINGLE).
	 */
	public byte[] getData() {
		return data;
	}

	/** Sets the data of the message (used when the data type is of TYPE_SINGLE). */
	public void setData(byte[] data) {
		this.data = data;
	}

	/**
	 * @return The entire raw data array with all the information.
	 */
	public byte[] getRawData() {
		return rawData;
	}

	public int getNumberOfParts() {
		if (dataType == TYPE_MULTIPART) {
			return parts.getSize();
		} else {
			return 0;
		}
	}

	public Message getPart(int index) {
		if (dataType != TYPE_MULTIPART) {
			return null;
		}

		PartNotice notice = (PartNotice) parts.getFirst();
		int i = 0;
		while (i < index && notice != null) {
			i++;
			notice = (PartNotice) notice.getNext();
		}

		if (notice == null) {
			return null;
		} else {
			return notice.message;
		}
	}

	public void addPart(Message part) {
		if (dataType == TYPE_MULTIPART) {
			parts.insertLast(new PartNotice(part));
		}
	}

	public void removePart(Message part) {
		if (dataType == TYPE_MULTIPART) {
			PartNotice notice = (PartNotice) parts.getFirst();
			while (notice != null && notice.message != part) {
				notice = (PartNotice) notice.getNext();
			}
			if (notice != null) {
				parts.remove(notice);
			}
		}
	}

	/**
	 * Set the linked list with message header nodes to the specified list.
	 * @param headerNodes LinkedList to set
	 */
	public void setHeaderNodes(LinkedList headerNodes) {
//		this.headerNodes.clear();
		LinkedList newList = new LinkedList();
		MessageHeaderNode hNode = (MessageHeaderNode)headerNodes.getFirst();
		while(hNode != null) {
			newList.insertLast(hNode.getCopy());
//			byte[] buffer = hNode.toByteArray();
//			insertHeaderNodeLast(new MessageHeaderNode(buffer, 0, buffer.length));
			hNode = (MessageHeaderNode)hNode.getNext();
		}
		this.headerNodes = newList;
	}

	/**
	 * Insert a message header node last in the linked list with nodes
	 * related to this message.
	 * @param hNode MessageHeaderNode to insert
	 */
	public void insertHeaderNodeLast(MessageHeaderNode hNode) {
		headerNodes.insertLast(hNode);
	}
	
	/**
	 * Insert a message header node after the specified node into the 
	 * linked list with nodes related to this message.
	 * @param newNode MessageHeaderNode to insert
	 * @param linkedNode MessageHeaderNode to insert after
	 */
	public void insertHeaderNodeAfter(MessageHeaderNode newNode, MessageHeaderNode linkedNode) {
		headerNodes.insertAfter(newNode, linkedNode);
	}

	/**
	 * Insert a message header node first in the linked list with nodes
	 * related to this message.
	 * @param hNode MessageHeaderNode to insert
	 */
	public void insertHeaderNodeFirst(MessageHeaderNode hNode) {
		headerNodes.insertFirst(hNode);
	}

	/**
	 * @return a LinkedList with all the message header nodes.
	 */
	public LinkedList getHeaderNodes() {
		return headerNodes;
	}


	/** Returns the sender deviceID. */
	public String getSenderDeviceID() {
		return senderDeviceID;
	}

	/** Sets the sender deviceID of this message. */
	public void setSenderDeviceID(String deviceID) {
		senderDeviceID = deviceID;
	}

	/** Returns the sender selector. */
	public Selector getSenderSelector() {
		return senderSelector;
	}

	/** Sets the sender selector of this message. */
	public void setSenderSelector(Selector selector) {
		senderSelector = selector;
	}

	/** Returns the receiver deviceID. */
	public String getReceiverDeviceID() {
		return receiverDeviceID;
	}

	/** Sets the receiver deviceID of this message. */
	public void setReceiverDeviceID(String deviceID) {
		receiverDeviceID = deviceID;
	}

	/** Returns the receiver selector. */
	public Selector getReceiverSelector() {
		return receiverSelector;
	}

	/** Sets the receiver selector. */
	public void setReceiverSelector(Selector selector) {
		receiverSelector = selector;
	}
	
	/**
	 * Returns the sender URL (protocol-specific URL), or null if it has not
	 * been set.
	 */
	public URL getSenderURL() {
		return senderURL;
	}

	/**
	 * Sets the sender URL of this message (protocol-specific URL, set when a
	 * message has been received over the network).
	 */
	public void setSenderURL(URL url) {
		senderURL = url;
	}

	/**
	 * Returns the receiver URL (protocol-specific URL), or null if it has not
	 * been set.
	 */
	public URL getReceiverURL() {
		return receiverURL;
	}

	/**
	 * Sets the receiver URL of this message (protocol-specific URL, set when a
	 * message has been received over the network).
	 */
	public void setReceiverURL(URL url) {
		receiverURL = url;
	}

	/**
	 * Write the message to a buffer, starting at index offset.
	 * 
	 * @return The number of bytes written.
	 * 
	 * @throws IndexOutOfBoundsException
	 *             if the buffer is too small.
	 */
	public int write(byte[] buffer, int offset) {
		int totalLength = getTotalLength();
		if (buffer.length < offset + totalLength) {
			throw new IndexOutOfBoundsException("Buffer too small.");
		}

		int pos = offset;

		if(rawData == null || rawData.length <= getHeaderNodesLength() + getDataLength()) {
			MessageHeaderNode hNode = (MessageHeaderNode) headerNodes.getFirst();
			while (hNode != null) {
				pos += writeStringToBuffer(hNode.toString(), buffer, pos);
				hNode = (MessageHeaderNode) hNode.getNext();
			}
			if (dataType != TYPE_NONE) {
				pos += writeDataNodeToBuffer(buffer, pos);
			}
		} else {
			pos += writeBytesToBuffer(rawData, buffer, pos);
		}

		return pos - offset;
	}

	/** @return The number of bytes written. */
	private int writeNodeToBuffer(byte format, byte[] nodeData, byte[] buffer,
			int offset) {
		int pos = offset;
		pos += writeFormatAndLengthToBuffer(format, nodeData.length, buffer,
				pos);
		pos += writeBytesToBuffer(nodeData, buffer, pos);
		return pos - offset;
	}

	/** @return The number of bytes written. */
	public int writeDataNodeToBuffer(byte[] buffer, int offset) {
		int pos = offset;
		pos += writeFormatAndLengthToBuffer(getType(), getDataLength(), buffer,
				pos);
		pos += writeData(buffer, pos);
		return pos - offset;
	}

	/** @return The number of bytes written. */
	private int writeFormatAndLengthToBuffer(byte format, int length,
			byte[] buffer, int offset) {
		int pos = offset;
		buffer[pos] = format;
		pos++;
		buffer[pos] = ';';
		pos++;
		String lengthString = Integer.toString(length);
		pos += writeStringToBuffer(lengthString, buffer, pos);
		buffer[pos] = ';';
		pos++;
		return pos - offset;
	}

	/** @return The number of bytes written. */
	private int writeStringToBuffer(String string, byte[] buffer, int offset) {
		byte[] bytes = string.getBytes();
		return writeBytesToBuffer(bytes, buffer, offset);
	}

	/** @return The number of bytes written. */
	private int writeBytesToBuffer(byte[] bytes, byte[] buffer, int offset) {
		System.arraycopy(bytes, 0, buffer, offset, bytes.length);
		return bytes.length;
	}

	/**
	 * Write the data of the message to buffer, starting at index offset in
	 * buffer.
	 * 
	 * @return The number of bytes written.
	 */
	protected int writeData(byte[] buffer, int offset) {
		if (dataType == TYPE_SINGLE) {
			System.arraycopy(data, 0, buffer, offset, data.length);
			return data.length;
		} else if (dataType == TYPE_MULTIPART) {
			int bytesWritten = 0;
			PartNotice notice = (PartNotice) parts.getFirst();
			while (notice != null) {
				bytesWritten += notice.message.write(buffer, offset
						+ bytesWritten);
				notice = (PartNotice) notice.getNext();
			}
			return bytesWritten;
		} else {
			return 0;
		}
	}

	/** @return A copy of the message, with the same data in new buffer(s). */
	public Message getCopy() {
		Message result = new Message();
//		result.version = version;
//		result.cachedInfo = cachedInfo;
		result.receiverDeviceID = receiverDeviceID;
		result.senderDeviceID = senderDeviceID;
//		result.connectionFunction = connectionFunction;
		result.receiverSelector = receiverSelector;
		result.senderSelector = senderSelector;
		result.dataType = dataType;
		result.senderURL = senderURL;
		result.receiverURL = receiverURL;
		if (data != null) {
			result.data = new byte[data.length];
			System.arraycopy(data, 0, result.data, 0, data.length);
		}
		if (rawData != null) {
			result.rawData = new byte[rawData.length];
			System.arraycopy(rawData, 0, result.rawData, 0, rawData.length);
		}
		for (int i = 0; i < getNumberOfParts(); i++) {
			result.addPart(getPart(i).getCopy());
		}
		result.setHeaderNodes(headerNodes);
		return result;
	}

	/**
	 * Factory method for parsing a Message from a byte array. The parsing
	 * starts at offset and parses to the end of a message, or at most length
	 * bytes.
	 * 
	 * @return The message read, or null if no message could be read.
	 */
	public static Message parse(byte[] buffer, int offset, int length) {
		int endOffset = offset + length;

		// Connection function and selectors
		/* byte connectionFunction;
        Selector receiverSelector = null;
        Selector senderSelector = null;
        if (result != null && result.format == NODE_FORMAT_CONNECTION) {
            connectionFunction = buffer[result.offset];
            int selectorPos = result.offset + 2;
            int selectorEndOffset = result.offset + result.length;
            int semicolonOffset = getSemicolonOffset(buffer, selectorPos,
                    selectorEndOffset);
            receiverSelector = new Selector(new String(buffer, selectorPos,
                    semicolonOffset - selectorPos));
            if (semicolonOffset < selectorEndOffset) {
                selectorPos = semicolonOffset + 1;
                senderSelector = new Selector(new String(buffer, selectorPos,
                        selectorEndOffset - selectorPos));
            }
            offset = selectorEndOffset;
            result = parseNode(buffer, offset, endOffset); // Parse next
        } else {
            connectionFunction = FUNCTION_NONE;
        }*/

		// Create message
		Message message = new Message();
		MessageHeaderNode currentNode = new MessageHeaderNode(buffer, offset, buffer.length);
		while(currentNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_NONE) {
			message.insertHeaderNodeLast(currentNode);
			currentNode = new MessageHeaderNode(buffer, message.getHeaderNodesLength(), buffer.length);
		}
		LinkedList tempHeaderNodes = message.getHeaderNodes();
		if(message.getHeaderNodesLength() > 0) {
			offset = message.getHeaderNodesLength();
		}
		NodeParsingResult result = parseNode(buffer, offset, endOffset);
		if (result == null) {
			message = new Message();
		} else {
			message = new Message(result.format, buffer, result.offset,
					result.length);
		}

		message.setHeaderNodes(tempHeaderNodes);
		message.rawData = buffer;

//		message.setReceiverURN(receiverURN);
//		message.setSenderURN(senderURN);
//		message.setSenderSelector(senderSelector);
		/*        message.setConnectionFunction(connectionFunction);
        message.setReceiverSelector(receiverSelector);
        message.setReceiverURN(receiverURN);
        message.setReceiverSelector(receiverSelector);
		 */
		//System.out.println(new String(data));
		//System.out.println("\n**********************************************************\n");

		return message;
	}
	
	/**
	 * Factory method for parsing a Message that have no 
	 * message header nodes from a byte array. The parsing starts at offset 
	 * and parses to the end of a message, or at most length
	 * bytes.
	 * 
	 * @return The message read, or null if no message could be read.
	 */
	private static Message parseWithOutHeaderNodes(byte[] buffer, int offset, int length) {
		int endOffset = offset + length;

		// Create message
		Message message;
		NodeParsingResult result = parseNode(buffer, offset, endOffset);
		if (result == null) {
			message = new Message();
		} else {
			message = new Message(result.format, buffer, result.offset,
					result.length);
		}

		return message;
	}

	private static NodeParsingResult parseNode(byte[] buffer, int offset,
			int endOffset) {
		if (offset >= endOffset) {
			return null;
		}
		NodeParsingResult result = new NodeParsingResult();
		result.format = buffer[offset];
		offset += 2;
		int pos = getSemicolonOffset(buffer, offset, endOffset);
		if (pos == endOffset) {
			return null;
		}
		try {
			String lengthString = new String(buffer, offset, pos - offset);
			result.length = Integer.parseInt(lengthString);
		} catch (NumberFormatException e) {
			// Wrong format of length
			return null;
		}
		result.offset = pos + 1;
		return result;
	}

	private static int getSemicolonOffset(byte[] data, int offset, int endOffset) {
		while (offset < endOffset && data[offset] != ';') {
			offset++;
		}
		return offset;
	}

	private static class NodeParsingResult {
		byte format;

		int offset;

		int length;
	}

	private static class PartNotice extends Link {
		final Message message;

		PartNotice(Message message) {
			this.message = message;
		}
	}
}
