/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import ist.palcom.collections.Link;

public class MessageHeaderNode extends Link{
	
	/** Special value used before a format has been assigned */
    public static final byte NODE_FORMAT_NONE = -1;
    
    /** The heartbeat message node 'h'. */
    public static final byte NODE_FORMAT_HEARTBEAT = 'h';
    
    /** The heartbeat ack message node 'H'. */
    public static final byte NODE_FORMAT_HEARTBEATACK = 'H';
    
    /** The heartattack message node 'X'. */
    public static final byte NODE_FORMAT_HEARTATTACK = 'X';
    
    /** The HBInfoRequest message node 'i'. */
    public static final byte NODE_FORMAT_HBINFOREQUEST = 'i';
    
    /** The HBInfoReply message node 'I'. */
    public static final byte NODE_FORMAT_HBINFOREPLY = 'I';
    
    /** The FormatVersion message node 'v'. */
    public static final byte NODE_FORMAT_VERSION = 'v';
    
    /** The Connection message node 'c'. */
    public static final byte NODE_FORMAT_CONNECTION = 'c';
    
    /** The RoutingR message node 'r'. */
    public static final byte NODE_FORMAT_ROUTINGR = 'r';
    
    /** The RoutingS message node 's'. */
    public static final byte NODE_FORMAT_ROUTINGS = 's';
    
    /** The mark message node 'm'. */
    public static final byte NODE_FORMAT_MARK = 'M';
    
    /** The Reliable message node 'R'. */
    public static final byte NODE_FORMAT_RELIABLE = 'R';
    
    /** The Reliable Ack message node 'A'. */
    public static final byte NODE_FORMAT_ACKMESSAGE = 'A';
    
    /** The Reliable resend message node 'B'. */
    public static final byte NODE_FORMAT_RESENDMESSAGE = 'B';
    
    /** The Chopped message node '-'. */
    public static final byte NODE_FORMAT_CHOPPED = '-';
    
    /** The single-shot message node 'S'. */
    public static final byte NODE_FORMAT_SINGLE_SHOT = 'S';
    
    /** The publish message node 'P'. */
    public static final byte NODE_FORMAT_PUBLISH = 'P';
    
    
    /** Special value used when function has not been set */
    public static final byte FUNCTION_NONE = -1;

    public static final byte FUNCTION_BROADCASTED = 'b';
    
    public static final byte FUNCTION_OPEN = 'o';

    public static final byte FUNCTION_OPEN_REPLY = 'p';
    
    public static final byte FUNCTION_CLOSE = 'c';
    
    public static final byte FUNCTION_RE_OPEN = 'r';

    public static final byte FUNCTION_MESSAGE_OVER_CONNECTION = 'm';

	private byte formatIdentifier;
	private byte functionIdentifier;
	private int length;
	private String[] data; //Contains the data from the header node as an array with strings 

	/**
	 * Constructs a message header node from an array with data.
	 * If the syntax is incorrect the header node will have format identifier
	 * set to NODE_FORMAT_NONE.
	 * @param data The array with data
	 * @param offset The offset to start parsing from
	 * @param length Parse the message to at most length bytes.
	 */
	public MessageHeaderNode(byte[] data, int offset, int length) {
		NodeParsingResult result = parseNode(data, offset, length);
		if(result != null && result.format != Message.TYPE_MULTIPART && 
				result.format != Message.TYPE_SINGLE) {
			formatIdentifier = result.format;
			if(formatIdentifier == NODE_FORMAT_CONNECTION) {
				functionIdentifier = result.function;
			} else {
				functionIdentifier = FUNCTION_NONE;
			}
			this.length = result.length;
			String dataString = "";
			if(functionIdentifier!=FUNCTION_NONE) {
				dataString = new String(data, result.offset, this.length-2);
			} else {
				dataString = new String(data, result.offset, this.length);
			}
			this.data = dataString.split(";");
		} else {
			this.length = 0;
			this.data = new String[0];
			formatIdentifier = NODE_FORMAT_NONE;
		}
	}
	
	/**
	 * Constructs a message header node with no data. (HBInfoRequests uses this format) 
	 * @param formatIdentifier The format of the node
	 */
	public MessageHeaderNode(byte formatIdentifier) {
		this.formatIdentifier = formatIdentifier;
		this.functionIdentifier = FUNCTION_NONE;
		this.data = new String[0];
		length = 0;
	}
	
	/**
	 * Constructs a message header node from the parameters with no function identifier. 
	 * @param formatIdentifier The format of the node
	 * @param data The data part of the node
	 */
	public MessageHeaderNode(byte formatIdentifier, String[] data) {
		this.formatIdentifier = formatIdentifier;
		this.functionIdentifier = FUNCTION_NONE;
		this.data = data;
		String dataString = "";
		for(int i = 0; i < data.length; ++i) {
			dataString += ";" + data[i];
		}
		//Special case when no data is used
		if(data.length == 0) {
			dataString = ";";
		}
		if(functionIdentifier==FUNCTION_NONE) {
			length = dataString.length()-1;
		}
	}

	/**
	 * Constructs a message header node from the parameters. 
	 * If the format is connection use the functionIdentifier parameter otherwise use FUNCTION_NONE for it.
	 * @param formatIdentifier The format of the node
	 * @param functionIdentifier The function of the node
	 * @param data The data part of the node
	 */
	public MessageHeaderNode(byte formatIdentifier, byte functionIdentifier, String[] data) {
		this.formatIdentifier = formatIdentifier;
		this.functionIdentifier = functionIdentifier;
		this.data = data;
		String dataString = "";
		for(int i = 0; i < data.length; ++i) {
			dataString += ";" + data[i];
		}
		//Special case when no data is used
		if(data.length == 0) {
			dataString = ";";
		}
		if(functionIdentifier==FUNCTION_NONE) {
			length = dataString.length()-1;
		} else {
			length = dataString.length()+1;
		}
	}

	/**
	 * @return The message header node as a string
	 */
	public String toString() {
		String dataString = "";
		for(int i = 0; i < data.length; ++i) {
			dataString += ";" + data[i];
		}
		//Special case when no data is used
		if(data.length == 0) {
			dataString = ";";
		} else {
			if(functionIdentifier!=FUNCTION_NONE) {
				dataString = ";" + new String(new byte[]{functionIdentifier}) + dataString;
			}
		}
		return new String(new byte[]{formatIdentifier}) + ";" + length + dataString;
	}

	public String[] getData() {
		return data;
	}

	public byte getFormatIdentifier() {
		return formatIdentifier;
	}
	
	public byte getFunctionIdentifier() {
		return functionIdentifier;
	}

	/**
	 * Returns the length of the data part of the node without the header part.
     */
	public int getLength() {
		return length;
	}
	
	/**
	 * Returns the total length of the message header node, including the length of the
	 * header with length and format.
     */
	public int getTotalLength() {
		return getNodeLength(length);
	}
	
	public byte[] toByteArray() {
		return toString().getBytes();
	}
	
	/**
	 * Returns the offset where the message data node or nodes start in a byte array 
	 * containing an entire message.
	 * @param message 
	 * 			The message to find the data node part in.
	 * @return The message data node offset.
	 */
	public static int getMessageDataNodeOffset(byte[] message) {
		int offset = 0;
		NodeParsingResult result = parseNode(message, 0, message.length);
		while(result != null && result.format != Message.TYPE_MULTIPART && 
				result.format != Message.TYPE_SINGLE) {
			offset = result.offset + result.length;
			result = parseNode(message, offset, message.length);
		}
		return offset;
	}
	
	public MessageHeaderNode getCopy() {
		return new MessageHeaderNode(formatIdentifier, functionIdentifier, data);
	}
	
	/**
     * @return The length of a node when printed out.
     * @param dataLength
     *            The length of the data in the node
     */
    private int getNodeLength(int dataLength) {
        int length = 3; // Format identifier and two semicolons
        length += getIntegerString(dataLength).length();
        length += dataLength;
        return length;
    }
    
    /**
     * Returns a String with the integer value, formatted as when written to a
     * stream or buffer.
     */
    private String getIntegerString(int value) {
        return Integer.toString(value);
    }
	
	private static NodeParsingResult parseNode(byte[] buffer, int offset,
            int endOffset) {
        if (offset >= endOffset) {
            return null;
        }
        NodeParsingResult result = new NodeParsingResult();
        result.format = buffer[offset];
        offset += 2;
        int pos = getSemicolonOffset(buffer, offset, endOffset);
        if (pos == endOffset) {
            return null;
        }
        try {
            String lengthString = new String(buffer, offset, pos - offset);
            result.length = Integer.parseInt(lengthString);
        } catch (NumberFormatException e) {
            // Wrong format of length
            return null;
        }
        if((pos+1 < endOffset  ) && result.format == NODE_FORMAT_CONNECTION) {
        	result.function = buffer[++pos];
        	pos++; //Put it at the following ';'
        } else {
        	result.function = FUNCTION_NONE;
        }
        result.offset = pos + 1;
        return result;
    }

    private static int getSemicolonOffset(byte[] data, int offset, int endOffset) {
        while (offset < endOffset && data[offset] != ';') {
            offset++;
        }
        return offset;
    }
	
	private static class NodeParsingResult {
        byte format;
        
        byte function;

        int offset;

        int length;
    }
}