/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class MultiPartMessage extends Message {

    private LinkedList parts = new LinkedList();

    public MultiPartMessage() {
    }

    public MultiPartMessage(byte[] data, int offset, int informationLength) {
        int bytesRead = 0;
        while (bytesRead < informationLength) {
            Message part = parse(data, offset + bytesRead, informationLength
                    - bytesRead);
            addPart(part);
            bytesRead += part.getTotalLength();
        }
    }

    public int getInformationLength() {
        int length = 0;
        PartNotice notice = (PartNotice) parts.getFirst();
        while (notice != null) {
            length += notice.message.getTotalLength();
            notice = (PartNotice) notice.getNext();
        }
        return length;
    }

    public byte getType() {
        return TYPE_MULTIPART;
    }

    public int getNumberOfParts() {
        return parts.getSize();
    }

    public Message getPart(int index) {
        PartNotice notice = (PartNotice) parts.getFirst();
        int i = 0;
        while (i < index && notice != null) {
            i++;
            notice = (PartNotice) notice.getNext();
        }

        if (notice == null) {
            return null;
        } else {
            return notice.message;
        }
    }

    public void addPart(Message part) {
        parts.insertLast(new PartNotice(part));
    }

    public void removePart(Message part) {
        PartNotice notice = (PartNotice) parts.getFirst();
        while (notice != null && notice.message != part) {
            notice = (PartNotice) notice.getNext();
        }
        if (notice != null) {
            parts.remove(notice);
        }
    }

    protected int writeInformation(byte[] buffer, int offset) {
        int bytesWritten = 0;
        PartNotice notice = (PartNotice) parts.getFirst();
        while (notice != null) {
            bytesWritten += notice.message.write(buffer, offset + bytesWritten);
            notice = (PartNotice) notice.getNext();
        }
        return bytesWritten;
    }

    public Message getCopy() {
        MultiPartMessage result = new MultiPartMessage();
        for (int i = 0; i < getNumberOfParts(); i++) {
            result.addPart(getPart(i).getCopy());
        }
        result.setReceiverSelector(getReceiverSelector());
        result.setSenderSelector(getSenderSelector());
        result.setSenderURL(getSenderURL());
        result.setReceiverURL(getReceiverURL());
        result.setSenderDeviceID(getSenderDeviceID());
        result.setReceiverDeviceID(getReceiverDeviceID());
        return result;
    }

    private class PartNotice extends Link {
        final Message message;

        PartNotice(Message message) {
            this.message = message;
        }
    }
}
