/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.bt;

/**
 * This class implements ist.palcom.mal.BluetoothConnectionManager
 * 
 * @see ist.palcom.palcomthreads.palcomprioritythread
 * 
 * @author Jacob Frolund Pedersen, frolund@daimi.au.dk
 */

import ist.palcom.base.networking.BluetoothSocket;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.CreateConnectionEvent;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.MessageArrayWithURLEvent;
import ist.palcom.mal.StopEvent;
import ist.palcom.mal.AbstractMediaManager.DistributorThread;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.IOAvailableEvent;
import ist.palcom.palcomthreads.IOTimeoutEvent;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.resource.descriptor.URL;

import java.io.IOException;
import java.util.ArrayList;

public class BluetoothConnectionManagerThread extends PalcomPriorityThread {

	private ArrayList threadURLs;

	private ArrayList threads;
	
	private DistributorThread distributor;
	
	public BluetoothCollector bluetoothMulticastCollector;
	
	private BluetoothCollector bluetoothUnicastCollector;
	
	private BluetoothSocket bluetoothServerSocket;
	
	private boolean running;
	private int timeout = 1000;
	private PalcomScheduler scheduler;
	
	public BluetoothConnectionManagerThread(PalcomScheduler scheduler, DistributorThread distributor) throws IOException {
		this.distributor = distributor;
		this.scheduler = scheduler;
		threadURLs = new ArrayList();
		threads = new ArrayList();
		bluetoothServerSocket = new BluetoothSocket();
		
		bluetoothMulticastCollector = new BluetoothCollector(new URL("bluetooth://00:00:00:00:00:00"),this);
	//	setThread("bluetooth://00:00:00:00:00:00", bluetoothMulticastCollector); 
		scheduler.scheduleThread(bluetoothMulticastCollector);

		bluetoothUnicastCollector = new BluetoothCollector(new URL("bluetooth://"+BluetoothSocket.getLocalAddress()),this);
		//setThread("bluetooth://"+BluetoothSocket.getLocalAddress(), bluetoothUnicastCollector); 
		scheduler.scheduleThread(bluetoothUnicastCollector);
	}

	public void stop(){
		for (int i = 0; i < threads.size(); i++) {
			CommunicationThread thread = (CommunicationThread) threads.get(i);
			thread.putEvent(new StopEvent());
		}
		threadURLs.clear();
		threads.clear();
	}
	
	protected CommunicationThread getReceivingThread(URL url) {
		CommunicationThread thread = null;
		
		int index = getIndexOfThread(url);
		if (index != -1) {
			thread = (CommunicationThread) threads.get(index);
		} else {
			if (url.equals(new URL("bluetooth://00:00:00:00:00:00"))) {
				thread = bluetoothMulticastCollector;
			} else if (url.equals( new URL("bluetooth://"
					+ BluetoothSocket.getLocalAddress()))) {
				thread = bluetoothUnicastCollector;
			} else {
				System.err.println("The url: " + url
						+ " does not have a receiving thread.");
			}
		}
		return thread;
	}

	protected void removeReceivingThread(URL url) {
		int index = getIndexOfThread(url);
		if (index != -1) {
			CommunicationThread thread = (CommunicationThread) threads
					.get(index);
			thread.putEvent(new StopEvent());
			threadURLs.remove(index);
			threads.remove(index);
		}
	}

	/** @return the index of the thread for url, or -1 if not found. */
	private int getIndexOfThread(URL url) {
		int index = 0;
		while (index < threadURLs.size() && !threadURLs.get(index).equals(url)) {
			index++;
		}
		if (index == threadURLs.size()) {
			return -1;
		} else {
			return index;
		}
	}

	/** Sets the receiving thread for URL url. */
	private void setThread(URL url, CommunicationThread thread) {
		int index = getIndexOfThread(url);
		if (index == -1) {
			threadURLs.add(url);
			threads.add(thread);
		} else {
			threads.set(index, thread);
		}
	}

	protected boolean hasConnection(){
		return !(threads.size() == 0);
	}
	
	//TODO: Implement a more inteligent bluetooth discovery strategy
	public void run() {
		running = true;
		enableIOEvent(bluetoothServerSocket,timeout);
		while (running) {
			Event event = waitEvent();
			if (event instanceof IOAvailableEvent) {
				if (bluetoothServerSocket.isConnected()) {
					try {
						BluetoothCommunicationThread bthread = new BluetoothCommunicationThread(scheduler,distributor, bluetoothServerSocket);
						setThread(bthread.getURL(), bthread);
						scheduler.scheduleThread(bthread);
						bluetoothServerSocket = new BluetoothSocket();
						enableIOEvent(bluetoothServerSocket,timeout);
					
					} catch (IOException e) {
						distributor.putEvent(new MALErrorEvent(e));
					}
				}
			} else if (event instanceof IOTimeoutEvent) {
				// Try again
				enableIOEvent(bluetoothServerSocket,timeout);
			} else if (event instanceof MessageArrayWithURLEvent) {
				//This is used when the ConnectionThread send messages
				MessageArrayWithURLEvent message = (MessageArrayWithURLEvent) event;
				if (message.getDeviceURL().equals(new URL("bluetooth://00:00:00:00:00:00"))) {
					//The message is a multicast message
					for (int i = 0; i < threads.size(); i++) {
						CommunicationThread cThread = (CommunicationThread) threads
								.get(i);
						cThread.putEvent(event);
					}
				} else {
					//The message is a unicast message
					CommunicationThread cThread = getReceivingThread(message.getDeviceURL());
					if(cThread != null){
						cThread.putEvent(event);
					}
				}
			} else if (event instanceof CreateConnectionEvent) {
				try {
					BluetoothCommunicationThread bthread = new BluetoothCommunicationThread(scheduler,distributor, ((CreateConnectionEvent)event).getConnectionUrl());
					scheduler.scheduleThread(bthread);
					setThread(bthread.getURL(), bthread);
					
				} catch (IOException e) {
					distributor.putEvent(new MALErrorEvent(e));
				}
			} else if (event instanceof StopEvent) {
				// Closing down
				stop();
				terminate();
			}
		}
	}
}
