/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.bt;

import ist.palcom.base.networking.BluetoothDiscovery;
import ist.palcom.mal.CreateConnectionEvent;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.MessageArrayWithURLEvent;
import ist.palcom.mal.StopEvent;
import ist.palcom.mal.AbstractMediaManager.DistributorThread;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.IOAvailableEvent;
import ist.palcom.palcomthreads.IOTimeoutEvent;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PeriodicEvent;

import javax.bluetooth.BluetoothStateException;

/**
 * This class implements ist.palcom.mal.BluetoothCommunicationThread
 * 
 * BluetoothCommunicationThread is a CommunicationThread for Bluetooth communication. It
 * sends an receives on one Bluetooth socket and is thus the rfcomm connection to a specific
 * host
 * 
 * @see ist.palcom.mal.CommunicationThread CommunicationThread
 * 
 * @author Jacob Frolund Pedersen, frolund@daimi.au.dk
 */

public class BluetoothDiscoveryThread extends PalcomPriorityThread {

	private PalcomScheduler scheduler;

    private BluetoothConnectionManagerThread connectionManager;

	private BluetoothDiscovery bluetoothDiscoverer;
	
	private DistributorThread distributor;

    protected BluetoothDiscoveryThread(PalcomScheduler scheduler, 
    		BluetoothConnectionManagerThread connectionManager, DistributorThread distributor) throws BluetoothStateException {
    	this.scheduler = scheduler;
    	this.distributor = distributor;
    	this.connectionManager = connectionManager;
    	setUp();
    }
    
    /** Perform socket set-up if needed. */
    protected void setUp() throws BluetoothStateException {
    	bluetoothDiscoverer = new BluetoothDiscovery();
    }

    /** Close socket if needed. */
    protected void tearDown() {
    }

    public void run() {
    	final int timeout = 1000;
    	try {
    		setUp();
	        bluetoothDiscoverer.searchDevices();
			enableIOEvent(bluetoothDiscoverer, timeout);
			
    	} catch (BluetoothStateException e) {
    		distributor.putEvent(new MALErrorEvent(e));
    	}
			
		
        while (true) {
            Event event = waitEvent();
            // System.out.println(hashCode() + " " + event);
            if (event instanceof IOAvailableEvent) {
            	if (bluetoothDiscoverer.getNumberOfDevices() > 0 || bluetoothDiscoverer.getNumberOfServices() > 0){
	            	try {	
						// We have received a event from the bluetooth discoverer
						if (bluetoothDiscoverer.getNumberOfServices() == 0) {
							bluetoothDiscoverer.searchServices();
						} else {
							for (int i = 0; i < bluetoothDiscoverer
									.getNumberOfServices(); i++) {
								String url = bluetoothDiscoverer
										.getServiceAddress();
								if (url != null) {
									connectionManager.putEvent(new CreateConnectionEvent(url));
								}
							}
							if(!connectionManager.hasConnection()){
								bluetoothDiscoverer.searchServices();
							}
						}
	            	} catch (BluetoothStateException e) {
						distributor.putEvent(new MALErrorEvent(e));
					}
            	} else {
            		//We found no devices to trying again
                    enablePeriodicEvents(bluetoothDiscoverer.timeToNextDiscovery()*1000);
            	}
				enableIOEvent(bluetoothDiscoverer, timeout);
            } else if (event instanceof IOTimeoutEvent) {
                // Try again
            	enableIOEvent(bluetoothDiscoverer, timeout);
            } else if (event instanceof PeriodicEvent){
            	disablePeriodicEvents();
            	if(!connectionManager.hasConnection()){
            		try {
            			bluetoothDiscoverer.searchDevices();
            		} catch (BluetoothStateException e) {
            			distributor.putEvent(new MALErrorEvent(e));
            		}
            	} else {
            		enablePeriodicEvents(5000);
            	}
            	enableIOEvent(bluetoothDiscoverer, timeout);
            } else if (event instanceof MessageArrayWithURLEvent) {
            	enableIOEvent(bluetoothDiscoverer, timeout);
            } else if (event instanceof StopEvent) {
                tearDown();
                terminate();
            }
        }
    }
}