/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.bt;

import ist.palcom.base.networking.BluetoothSocket;
import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.URL;

import java.io.IOException;

/**
 * This class implements ist.palcom.mal.BluetoothMediaManager
 * 
 * @see ist.palcom.mal.MediaManager MediaManager
 * 
 * @author Jacob Frolund Pedersen, frolund@daimi.au.dk
 */

public class BluetoothMediaManager extends AbstractMediaManager {
	
	private BluetoothConnectionManagerThread bluetoothConnectionManager;
	
	private BluetoothDiscoveryThread bluetoothDiscoveryThread;
	
    private URL[] defaultGroupAddresses;
    
    private boolean isStarted;
    
    /**
     * @deprecated See superclass
     */
	public BluetoothMediaManager(PalcomScheduler scheduler) {
		this(scheduler, null);
		System.err.println("Warning: Using deprecated BluetoothMediaManager constructor");
	}
	
	public BluetoothMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler) {
		super(scheduler, errorHandler);
		startIfNotAlready();
	}

	public void stop() {
		bluetoothConnectionManager.stop();
		super.stop();
	}
	
	//TODO: this might need to be changed, as we do not have a default receiving url
	/**
	 * Start receiving all incomming packages on the different bluetooth links.
	 * 
	 *
	 public String startReceiving(PalcomThread receiver) {
	     return "";
	 } */
	 
	//TODO: this needs to be implemented
	/**
	 * If the exists a connection to a device with this url, then subscribe to these 
	 * messages. If no connection exists open a new one and start receiving chunks of 
	 * data from the new connection. The later will offend be used in connections to 
	 * 3th party devices eg. gps.
	 *
	 public String startReceiving(String url, PalcomThread receiver) {
		// getReceivingThread(url).
		 return "";
	 }*/
	
	//TODO: this needs to be implemented
	public void suggestSendBufferSize(int size) {
		
	}
	
//	TODO: this needs to be implemented
	public int getSendBufferSize() {
		return -1;
	}
	 
	 public URL getDefaultUnicastURL() {
		startIfNotAlready();
		return new URL("bluetooth://"+BluetoothSocket.getLocalAddress());
	}

	protected CommunicationThread getSendingThread(URL url) {
		startIfNotAlready();
		String protocol = getProtocol(url);
		if (protocol.equals("bluetooth")) {
			if(isBroadcast(url)){
				return bluetoothConnectionManager.bluetoothMulticastCollector;
			} else {
				// Use receiving
				return getReceivingThread(url);
			}
		} else {
			throw new IllegalArgumentException("Protocol of " + url + " not supported.");
		}
	}

	/**
	 * This does not make sense in a bluetooth world, as there is only point to point
	 * connections. We have to create routing on top of that and this is not in this 
	 * layer
	 */
	protected boolean isBroadcast(URL url) {
		return (url.equals(getDefaultGroupAddresses()[0]));
	}

	/**
	 * We use the address 00:00:00:00:00:00 as the default group address, as
	 * no such thing exists in Bluetooth. A call to receive(00:00:00:00:00:00,palcomThread)
	 * will result in the thread receiving all bluetooth messages send to this node
	 */
	public URL[] getDefaultGroupAddresses() {
		startIfNotAlready();
        if (defaultGroupAddresses == null){
            defaultGroupAddresses = new URL[] {new URL("bluetooth://00:00:00:00:00:00")};
        }
		return defaultGroupAddresses;
	}

	protected CommunicationThread getReceivingThread(URL url) {
		startIfNotAlready();
		return bluetoothConnectionManager.getReceivingThread(getProtocolSpecificUrl(url));
	}

	protected void removeReceivingThread(URL url) {
		startIfNotAlready();
		bluetoothConnectionManager.removeReceivingThread(getProtocolSpecificUrl(url));
	}	
	
	/**
	 * Placed here, instead of in the constructor, so that..
	 * (1) The client has a chance to register an error handler
	 * (2) The user has a chance to fix any problems (attach dongle etc.) after first failure
	 */
	private void startIfNotAlready() {
		if (isStarted) return;
		try {
			bluetoothConnectionManager = new BluetoothConnectionManagerThread(scheduler, distributorThread);
			scheduler.scheduleThread(bluetoothConnectionManager);
			bluetoothDiscoveryThread = new BluetoothDiscoveryThread(scheduler, bluetoothConnectionManager, distributorThread);
			scheduler.scheduleThread(bluetoothDiscoveryThread);
			isStarted = true;
			
		} catch (IOException e) {
			raiseErrorEvent(new MALErrorEvent(e));
		}
	}
}
