/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.ir;

import ist.palcom.base.networking.IRManager;
import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.URL;

/**
 * Media manager, capable of exchanging PalCom messages through
 * InfraRed (IR) communication.
 * <p>
 * URL format (Tentative): palcom:ir://$device$:$ir-port$/$selector$, where
 * $ir-port$ is one of 1 through 4, or 0 for broadcast to all 4 ports.
 * 
 * @see ist.palcom.mal.MediaManager MediaManager
 * @author Henrik Gammelmark, geemark@daimi.au.dk
 */
public class IRMediaManager extends AbstractMediaManager {
	private static final String IR_URL_PREFIX = PALCOM_URL_PREFIX + "ir://";
	public static final int BROADCAST_HOST = 0;
	public static final String SERIAL_DEVICE = "/dev/ttyS0";
	
	private IRSenderThread commThreadSend;
	private IRReceiverThread commThreadReceive;
	private IRManager ir;		// SmallTalk instance
    private URL[] defaultGroupAddresses;
	
    /**
     * @deprecated See superclass
     */
    public IRMediaManager(PalcomScheduler scheduler) {
    	this(scheduler, null);
    	System.err.println("Warning: Using deprecated IRMediaManager constructor");
    }
    
	public IRMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler) {
        super(scheduler, errorHandler);
        
        // Start IR manager
        ir = new IRManager();
		ir.open(SERIAL_DEVICE);
        
		// Start threads for sending and receiving messages
        commThreadSend = new IRSenderThread(scheduler, this, IR_URL_PREFIX);
        commThreadReceive = new IRReceiverThread(scheduler, this, IR_URL_PREFIX);
        
        scheduler.scheduleThread(commThreadSend);
        scheduler.scheduleThread(commThreadReceive);
	}
	
	public void stop() {
		commThreadSend.putEvent(new StopEvent());
		commThreadReceive.putEvent(new StopEvent());
		ir.close();
		super.stop();
	}
	
//	TODO: this needs to be implemented
	public void suggestSendBufferSize(int size) {
		
	}
	
//	TODO: this needs to be implemented
	public int getSendBufferSize() {
		return -1;
	}
	
	public URL getDefaultUnicastURL() {
		return new URL(IR_URL_PREFIX + BROADCAST_HOST + ":0");
	}
	
	protected CommunicationThread getReceivingThread(URL url) {
		return commThreadReceive;
	}
	
	public IRManager getIRManager() {
		return ir;
	}
		
	protected void removeReceivingThread(URL url) {
		// You wish
	}
	
	protected CommunicationThread getSendingThread(URL url) {
		return commThreadSend;
	}
	
	protected boolean isBroadcast(URL url) {
		return getHost(url).equals("" + BROADCAST_HOST);
	}
	
	public URL[] getDefaultGroupAddresses() {
	    if (defaultGroupAddresses == null) {
	        defaultGroupAddresses = new URL[] { new URL(IR_URL_PREFIX + BROADCAST_HOST + ":0") };
	    }
	    return defaultGroupAddresses;
	}
}
