/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.palcomthreads;

import ist.palcom.base.Channel;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class PalcomScheduler {

    private SortedPalcomThreadQueue readyQueue;

    private SortedPalcomThreadQueue timeQueue;

    /** Array of channels used when calling select. */
    private Channel[] channels;

    /** The number of channels in channels that are used. */
    private int numberOfChannels;

    /** array for results from select. */
    private int[] results;

    public PalcomScheduler() {
        readyQueue = new SortedPalcomThreadQueue();
        timeQueue = new SortedPalcomThreadQueue();
        channels = new Channel[10];
        numberOfChannels = 0;
        results = new int[channels.length];
    }

    /** Connect the PalcomThread t to this scheduler. */
    public void scheduleThread(PalcomThread t) {
        t.setScheduler(this);
        t.schedule();
    }

    /** Force t to be terminated and disconnect it from this PalcomScheduler. */
    public void kill(PalcomThread t) {
        throw new RuntimeException("Not implemented.");
    }

    /**
     * Hand over control to this PalcomScheduler (typically called once in the
     * main program).
     */
    public void run() {
        boolean finished = false;
        while (!finished) {
            // The following yield is currently necessary on pre-vm-c
            // It is a no-op in the java-environment
            ist.palcom.base.System.instance().yield();
            PalcomThreadNotice notice = readyQueue.getFirst();
            if (notice != null) {
                PalcomThread current = notice.getThread();
                if (current.state == PalcomThread.STATE_READY) {
                    current.coroutine.attach();
                }
                readyQueue.removeFirst();
                if (current.state == PalcomThread.STATE_READY) {
                    readyQueue.sortInto(notice);
                }
                waitForInterrupt(false);
            } else {
                if (!waitForInterrupt(true)) {
                    finished = true;
                }
            }
        }
    }

    /**
     * @return True if events have been scheduled, false if there will be no
     *         more events.
     */
    private boolean waitForInterrupt(boolean blocking) {
        // Measure once for consistent value
        long time = ist.palcom.base.System.instance().time();

        boolean eventGenerated = false;
        TimeNotice timeNotice = (TimeNotice) timeQueue.getFirst();
        int n = 0; // Result from select
        if (!blocking) {
            // Don't wait - just check if data is available
            if (numberOfChannels > 0) {
                n = ist.palcom.base.System.instance().selectNow(channels,
                        numberOfChannels, results);
            }
        } else if (timeNotice != null) {
            long diff = timeNotice.nextTime - time;
            if (diff > 0) {
                n = ist.palcom.base.System.instance().select(channels,
                        numberOfChannels, results, (int) diff);
            } else {
                n = 0; // Timeout already late
            }

            if (n == 0) {
                // Timeout - generate event
                PalcomThread thread = timeNotice.getThread();
                if (timeNotice instanceof PeriodicNotice) {
                    timeQueue.removeFirst();
                    PeriodicNotice periodicNotice = (PeriodicNotice) timeNotice;
                    if (thread.state != PalcomThread.STATE_TERMINATED) {
                        periodicNotice.nextTime += periodicNotice.period;
                        timeQueue.sortInto(periodicNotice);
                        thread.putEvent(new PeriodicEvent());
                    }
                } else if (timeNotice instanceof TimeoutNotice) {
                    TimeoutNotice timeoutNotice = (TimeoutNotice) timeNotice;
                    timeQueue.removeFirst();
                    if (thread.state != PalcomThread.STATE_TERMINATED) {
                        removeChannel(timeoutNotice.getChannel());
                        thread.putEvent(new IOTimeoutEvent());
                    }
                }
                eventGenerated = true;
            }
        }
        if (n > 0) {
            // I/O available/possible
            PalcomThreadNotice notice = timeQueue.getFirst();
            while (notice != null) {
                PalcomThreadNotice nextNotice = (PalcomThreadNotice) notice
                        .getNext();
                if (notice instanceof TimeoutNotice) {
                    TimeoutNotice timeoutNotice = (TimeoutNotice) notice;
                    int index = indexOfChannel(timeoutNotice.getChannel());
                    if (index != -1 && results[index] > 0) {
                        // Only generate event once to a PalcomThread
                        PalcomThread t = timeoutNotice.getThread();
                        if (t.state == PalcomThread.STATE_WAITING
                                || t.state == PalcomThread.STATE_READY) {
                            t.putEvent(new IOAvailableEvent());
                            eventGenerated = true;
                        }
                        timeQueue.remove(notice);
                        removeChannel(timeoutNotice.getChannel());
                    }
                }
                notice = nextNotice;
            }
        }
        return eventGenerated;
    }

    /**
     * @return A reference to the PalcomThread currently scheduled by this
     *         PalcomScheduler.
     */
    PalcomThread getCurrentThread() {
        if (!readyQueue.isEmpty()) {
            return readyQueue.getFirst().getThread();
        } else {
            return null;
        }
    }

    void makeReadyForScheduling(PalcomThread t) {
        readyQueue.sortInto(new PalcomThreadNotice(t));
    }

    void enablePeriodicEvents(PalcomThread t, long period) {
        long nextTime = ist.palcom.base.System.instance().time() + period;

        PeriodicNotice notice = timeQueue.findPeriodicNotice(t);
        if (notice == null) {
            // Create a new one
            notice = new PeriodicNotice(t, nextTime, period);
        } else {
            // Change existing PeriodicNotice
            notice.nextTime = nextTime;
            notice.period = period;
            timeQueue.remove(notice);
        }
        timeQueue.sortInto(notice);
    }

    void disablePeriodicEvents(PalcomThread t) {
        PeriodicNotice notice = timeQueue.findPeriodicNotice(t);
        while (notice != null) {
            timeQueue.remove(notice);
            notice = timeQueue.findPeriodicNotice(t);
        }
    }

    void enableIOEvent(PalcomThread t, Channel channel, int timeout) {
        long nextTime = ist.palcom.base.System.instance().time() + timeout;
        disableIOEvents(t);
        TimeoutNotice timeoutNotice = new TimeoutNotice(t, nextTime, channel);
        timeQueue.sortInto(timeoutNotice);
        addChannel(channel);
    }

    void disableIOEvents(PalcomThread t) {
        TimeoutNotice timeoutNotice = timeQueue.findTimeoutNotice(t);
        while (timeoutNotice != null) {
            timeQueue.remove(timeoutNotice);
            removeChannel(timeoutNotice.getChannel());
            timeoutNotice = timeQueue.findTimeoutNotice(t);
        }
    }

    private void addChannel(Channel channel) {
        if (indexOfChannel(channel) == -1) {
            // Not already added
            if (numberOfChannels == channels.length) {
                Channel[] newArray = new Channel[2 * channels.length];
                for (int i = 0; i < channels.length; i++) {
                    newArray[i] = channels[i];
                }
                channels = newArray;
                results = new int[channels.length];
            }
            channels[numberOfChannels] = channel;
            numberOfChannels++;
        }
    }

    private void removeChannel(Channel channel) {
        int index = indexOfChannel(channel);
        if (index != -1) {
            for (int i = index; i < numberOfChannels - 1; i++) {
                channels[i] = channels[i + 1];
                results[i] = results[i + 1]; // Must update this array also
            }
            channels[numberOfChannels - 1] = null;
            numberOfChannels--;
        }
    }

    /** @return The index of channel in the array, or -1 if not found. */
    private int indexOfChannel(Channel channel) {
        int index = 0;
        while (index < numberOfChannels && channels[index] != channel) {
            index++;
        }
        if (index == numberOfChannels) {
            return -1;
        } else {
            return index;
        }
    }
}