/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.rascal;

import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.mal.Message;

/**
 * Envelop of a PalCom message which includes the URN/URL/Selectors 
 * of the receiver/sender.
 * 
 * @author rgh
 *
 */
public class RascalMessage {

	private URN urn;
	
	private String senderDeviceID;
	
	private URL senderURL;
	
	private Selector senderSelector;
	
	private String receiverDeviceID;
	
	private URL receiverURL;
	
	private Selector receiverSelector;

	private Message msg;
	
	public RascalMessage(URN urn, Message msg) {
		this.urn = urn;
		this.msg = msg;
		
		this.senderDeviceID = msg.getSenderDeviceID();
		this.senderURL = msg.getSenderURL();
		this.senderSelector = msg.getSenderSelector();
		
		this.receiverDeviceID = msg.getReceiverDeviceID();
		this.receiverURL = msg.getReceiverURL();
		this.receiverSelector = msg.getReceiverSelector();
		
	}
	
	public int getTotalLength() {
        int length = getURNString(urn).length();
        length += 1;

        // SENDER
        length += getDeviceIDString(senderDeviceID).length();
        length += 1;
        
        length += getSelectorString(senderSelector).length();
        length += 1;

        length += getURLString(senderURL).length();
        length += 1;

        // RECEIVER
        length += getDeviceIDString(receiverDeviceID).length();
        length += 1;
        
        length += getSelectorString(receiverSelector).length();
        length += 1;

        length += getURLString(receiverURL).length();
        length += 1;
        
        length += msg.getTotalLength();
		
		return length;
	}

    private String getURLString(URL url) {
        if (url == null) {
            return "";
        } else {
            return url.toString();
        }
    }
    
    private String getURNString(URN urn) {
        if (urn == null) {
            return "";
        } else {
            return urn.toString();
        }
    }
    
    private String getDeviceIDString(String deviceID) {
        if (deviceID == null) {
            return "";
        } else {
            return deviceID;
        }
    }
    
    private String getSelectorString(Selector selector) {
        if (selector == null) {
            return "";
        } else {
            return selector.toString();
        }
    }
    

    /**
     * Write the RASCAL message to a buffer, starting at index offset.
     * 
     * @return The number of bytes written.
     * 
     * @throws IndexOutOfBoundsException
     *             if the buffer is too small.
     */
    public int write(byte[] buffer, int offset) {
        int totalLength = getTotalLength();
        if (buffer.length < offset + totalLength) {
            throw new IndexOutOfBoundsException("Buffer too small.");
        }

        int pos = offset;

        pos += writeStringToBuffer(getURNString(urn), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        // SENDER
        pos += writeStringToBuffer(getDeviceIDString(senderDeviceID), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        pos += writeStringToBuffer(getSelectorString(senderSelector), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        pos += writeStringToBuffer(getURLString(senderURL), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        // RECEIVER
        pos += writeStringToBuffer(getDeviceIDString(receiverDeviceID), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        pos += writeStringToBuffer(getSelectorString(receiverSelector), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);

        pos += writeStringToBuffer(getURLString(receiverURL), buffer, pos);
        pos += writeStringToBuffer(";", buffer, pos);
        
        pos += msg.write(buffer, pos);
        
        return pos - offset;
    }

    /**
     * Factory method for parsing a Rascal Message from a byte array. The parsing
     * starts at offset and parses to the end of a message, or at most length
     * bytes.
     * 
     * @return The message read, or null if no message could be read.
     */
    public static RascalMessage parse(byte[] data, int offset, int length) {
        int pos = offset;
        int endPos = offset + length;

        int len = getHeaderPartLength(data, pos, endPos - pos);
        if (len == -1)
            return null;
        String urn = new String(data, pos, len);

        pos += (len + 1);

        // SENDER
        len = getHeaderPartLength(data, pos, endPos - pos);
        String senderURN = new String(data, pos, len);
        pos += (len + 1);

        len = getHeaderPartLength(data, pos, endPos - pos);
        String senderSelector = new String(data, pos, len);
        pos += (len + 1);

        len = getHeaderPartLength(data, pos, endPos - pos);
        String senderURL = new String(data, pos, len);
        pos += (len + 1);

        // RECEIVER        
        len = getHeaderPartLength(data, pos, endPos - pos);
        String receiverURN = new String(data, pos, len);
        pos += (len + 1);

        len = getHeaderPartLength(data, pos, endPos - pos);
        String receiverSelector = new String(data, pos, len);
        pos += (len + 1);

        len = getHeaderPartLength(data, pos, endPos - pos);
        String receiverURL = new String(data, pos, len);
        pos += (len + 1);

        byte[] subData = new byte[endPos - pos];
        System.arraycopy(data, pos, subData, 0, endPos - pos);
        
        Message msg = Message.parse(subData, 0, subData.length);

        if (!senderURN.equals(""))
        	msg.setSenderDeviceID(senderURN);
        
        if (!senderSelector.equals(""))
        	msg.setSenderSelector(new Selector(senderSelector));
        
        if (!senderURL.equals(""))
        	msg.setSenderURL(new URL(senderURL));
        
        if (!receiverURN.equals(""))
        	msg.setReceiverDeviceID(receiverURN);
        
        if (!receiverSelector.equals(""))
        	msg.setReceiverSelector(new Selector(receiverSelector));
        
        if (!receiverURL.equals(""))
        	msg.setReceiverURL(new URL(receiverURL));

        RascalMessage message;
        if ( (urn != null) && !(urn.equals("")) )
        	message = new RascalMessage(new URN(urn), msg);
        else
        	message = new RascalMessage(null, msg);
        
        return message;
    }

    private int writeStringToBuffer(String string, byte[] buffer, int offset) {
        byte[] bytes = string.getBytes();
        return writeBytesToBuffer(bytes, buffer, offset);
    }

    private int writeBytesToBuffer(byte[] bytes, byte[] buffer, int offset) {
        System.arraycopy(bytes, 0, buffer, offset, bytes.length);
        return bytes.length;
    }

    /**
     * @return The length of the next header part (up to but not including a
     *         semicolon), starting at offset. If no semicolon is found for
     *         length bytes, -1 is returned.
     */
    private static int getHeaderPartLength(byte[] buffer, int offset, int length) {
        int pos = offset;
        int endPos = offset + length;
        while (pos < endPos && buffer[pos] != ';') {
            pos++;
        }
        if (pos == endPos) {
            return -1;
        } else {
            return pos - offset;
        }
    }

	public Message getMessage() {
		return msg;
	}

	public URN getURN() {
		return urn;
	}
}
