/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing;

import ist.palcom.collections.LinkedList;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URN;

import java.util.ArrayList;
import java.util.Iterator;

/** Abstract class for a routing manager, implements RoutingManager. 
 * Handles chopped and reliable messages. */
public abstract class AbstractRoutingManager2 implements RoutingManager{

	private ChoppedMessageHandler choppedMessageHandler;
	/*****hgraph impl*****/
	protected HgraphNode _node;


	public AbstractRoutingManager2() {
		choppedMessageHandler = new ChoppedMessageHandler();
		_node = new HgraphNode("routing manager");
	}

	//Old wire protocol methods
	public abstract void stopReceiving(PalcomThread receiver);
	public abstract void stopReceiving(Selector selector);
	public abstract boolean send(Message msg, URN urn1, Selector selector1, URN urn2, Selector selector2);

	//The following methods supports the new wire protocol
	public abstract boolean send(Message message, String receiverDeviceID);

	/** Clears the choppedMessageHandler.
	 * Subclasses override this method to stop the media manager(s), 
	 * and call super.stop from there.
	 */
	public void stop() {
		choppedMessageHandler.clear();
	}

	/** Remove the device with deviceID from the list of registered devices. 
	 * Subclasses override this method and call super.unregisterDevice from there.
	 * */
	public void unregisterDevice(String deviceID) {
		choppedMessageHandler.removeAllMessagesFromSender(deviceID);
	}

	public abstract void addErrorHandler(PalcomThread handler);
	public abstract void setLocalDeviceCache(String localDeviceCache);
	public abstract String getLocalDeviceCache();
	public abstract void setLocalDeviceID(String localDeviceID);
	public abstract void setLocalStatusFlag(byte statusFlag);
	public abstract void setLocalDiscoveryVersion(String discoveryVersion);
	public abstract void setDiscoverySelector(Selector selector);
	public abstract void sendBroadcast(Message message);
	public abstract void sendHeartAttack();
	public abstract void registerAsListener(PalcomThread communicationManager);
	public abstract void registerForHeartBeats(PalcomThread resourceManager);
	public abstract void sendHeartBeat(boolean isHeartBeatAck);


	public ChoppedMessageHandler getChoppedMessageHandler() {
		return choppedMessageHandler;
	}

	/**
	 * This method is required, in the current implementation, when a
	 * message header node has been removed. It creates a new message without
	 * the raw data array.
	 * 
	 * @param Message
	 * @return New message
	 */
	protected Message adjustMessageRawData(Message message) {
		Message correctedMessage;
		if (message.getType() != Message.TYPE_NONE) {
			if (message.getType() == Message.TYPE_SINGLE) {
				correctedMessage = new Message(message.getType(), message
						.getData());
			} else {
				correctedMessage = new Message(message.getType());
				for (int k = 0; k < message.getNumberOfParts(); ++k) {
					correctedMessage.addPart(message.getPart(k));
				}
			}
		} else {
			correctedMessage = new Message();
		}
		correctedMessage.setSenderDeviceID(message.getSenderDeviceID());
		return correctedMessage;
	}
	
	/** Handles chopping and reassembling of messages */
	protected class ChoppedMessageHandler {

		private ArrayList messageData;

		private ArrayList messageSenderKey;

		private ArrayList previousPart;

		private int messageID;

		ChoppedMessageHandler() {
			messageSenderKey = new ArrayList();
			messageData = new ArrayList();
			previousPart = new ArrayList();
		}

		void clear() {
			messageSenderKey.clear();
			messageData.clear();
			previousPart.clear();

			messageID = 1;
		}

		/**
		 * @return the index for key key, or -1 if not found.
		 */
		private int getIndexOf(String key) {
			int index = 0;
			while (index < messageSenderKey.size()
					&& !messageSenderKey.get(index).equals(key)) {
				index++;
			}
			if (index == messageSenderKey.size()) {
				return -1;
			} else {
				return index;
			}
		}

		/**
		 * Sets the first part of the chopped message with messageID. Sent from
		 * senderURN. If set before, the old listener is removed.
		 */
		private void setFirstMessagePart(String messageID, String senderDeviceID,
				byte[] message) {
			String searchString = messageID + "/" + senderDeviceID;
			int index = getIndexOf(searchString);
			if (index == -1) {
				messageSenderKey.add(searchString);
				messageData.add(message);
				previousPart.add(new Integer(1));
			} else {
				messageData.set(index, message);
				previousPart.set(index, new Integer(1));
			}
		}

		/**
		 * Sets a middle part of the chopped message with messageID. Sent from
		 * senderURN. Returns true if the message was inserted or false if it
		 * was dropped becasuse it was out of order or not found.
		 */
		private void setMessagePart(String messageID, String senderDeviceID, byte[] message,
				int part) {
			String searchString = messageID + "/" + senderDeviceID;
			int index = getIndexOf(searchString);
			if (index != -1) {
				if (((Integer) (previousPart.get(index))).intValue() + 1 == part) {
					byte[] currentMessage = (byte[]) messageData.get(index);
					byte[] newMessage = new byte[currentMessage.length
					                             + message.length];
					System.arraycopy(currentMessage, 0, newMessage, 0,
							currentMessage.length);
					System.arraycopy(message, 0, newMessage,
							currentMessage.length, message.length);
					messageData.set(index, newMessage);
					previousPart.set(index, new Integer(part));

				} else {
					messageSenderKey.remove(index);
					messageData.remove(index);
					previousPart.remove(index);
				}
			}
		}

		/**
		 * Sets the last part of the chopped message with messageID. Sent from
		 * senderURN. The stored data is removed from the list after.
		 */
		private Message setLastMessagePart(String messageID, String senderDeviceID,
				Message message) {
			String searchString = messageID + "/" + senderDeviceID;

			int index = getIndexOf(searchString);
			if (index != -1) {
				byte[] currentMessage = (byte[]) messageData.get(index);
				byte[] newMessage = new byte[currentMessage.length
				                             + message.getDataLength()];
				System.arraycopy(currentMessage, 0, newMessage, 0,
						currentMessage.length);
				System.arraycopy(message.getData(), 0, newMessage,
						currentMessage.length, message.getDataLength());

				Message tempMessage = Message.parse(newMessage, 0,
						newMessage.length);
				tempMessage.setHeaderNodes(message.getHeaderNodes());
				tempMessage.setReceiverSelector(message.getReceiverSelector());
				tempMessage.setSenderSelector(message.getSenderSelector());
				tempMessage.setSenderDeviceID(message.getSenderDeviceID());

				messageSenderKey.remove(index);
				messageData.remove(index);
				previousPart.remove(index);
				return tempMessage;
			} else {
				return null;
			}
		}

		void removeReceiver(String messageID, String senderDeviceID) {
			String searchString = messageID + "/" + senderDeviceID;
			int index = getIndexOf(searchString);
			if (index != -1) {
				messageSenderKey.remove(index);
				messageData.remove(index);
				previousPart.remove(index);
			}
		}

		void removeAllMessagesFromSender(String senderDeviceID) {
			String searchString = senderDeviceID;
			for(int index = 0; index < messageSenderKey.size(); ++index) {
				if(((String)messageSenderKey.get(index)).indexOf(searchString) >= 0) {
					messageSenderKey.remove(index);
					messageData.remove(index);
					previousPart.remove(index);
					index--;
				}
			}
		}

		/**
		 * @param sendBufferSize The largest size of a message to send.
		 * @param message
		 *            The message to chop.
		 * @return an array with either the original message, the original message
		 *         chopped in small messages or an empty array if the send buffer is
		 *         to small for sending the message.
		 */
		public Message[] chopMessage(int sendBufferSize, Message message) {
			
			//Reduce sendBufferSize a bit to be on the safe side. The MediaManager might add a MessageHeaderNode.
			sendBufferSize -= 40;
			
			//A node of this type will be inserted later on so its length must be accounted for.
			//1000 is a "big" number unlikely to be exceeded. 
			MessageHeaderNode extraNode = new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_CHOPPED, new String[] {
							"m" + messageID, "1000",
							"1000"});
			//3 is the extra length in the new message by Type;length; as a header
			int partLength = sendBufferSize
			- (message.getHeaderNodesLength() + 3 + extraNode.toString().length());
			//The length of the data part will also be included in the message
			partLength -= Integer.toString(partLength).length();
			
			if (message.getTotalLength() >= sendBufferSize
					&& partLength > 0) {
				
				int dataLength = message.getNodeAndDataLength();
				
				int nbrOfParts = dataLength / partLength;
				int controlLength = partLength * nbrOfParts;
				Message[] messages = new Message[nbrOfParts];
				int fixedNbrOfParts = nbrOfParts;
				// Is there a round-off error?
				if (controlLength < dataLength) {
					messages = new Message[nbrOfParts + 1];
					fixedNbrOfParts++;
				}
				byte[] data = new byte[dataLength];
				message.writeDataNodeToBuffer(data, 0);
				LinkedList headerNodes = message.getHeaderNodes();
				byte[] partData = new byte[partLength];

				for (int i = 0; i < nbrOfParts; ++i) {
					System.arraycopy(data, i * partLength, partData, 0,
							partData.length);
					Message partMessage = new Message(partData);
					partMessage.setHeaderNodes(headerNodes);

					MessageHeaderNode hNode = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_CHOPPED, new String[] {
									"m" + messageID, "" + (i + 1),
									"" + fixedNbrOfParts });
					partMessage.insertHeaderNodeLast(hNode);

					messages[i] = partMessage;
				}
				// Handle any end part that is different in size. Caused by
				// round-off error.
				if (controlLength < dataLength) {
					partData = new byte[dataLength - controlLength];
					System.arraycopy(data, nbrOfParts * partLength, partData, 0,
							dataLength - controlLength);
					Message partMessage = new Message(partData);
					partMessage.setHeaderNodes(headerNodes);
					MessageHeaderNode hNode = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_CHOPPED, new String[] {
									"m" + messageID, "" + fixedNbrOfParts,
									"" + fixedNbrOfParts });
					partMessage.insertHeaderNodeLast(hNode);
					
					messages[nbrOfParts] = partMessage;
				}
				messageID++;
				
				return messages;
			} else if (message.getTotalLength() < sendBufferSize) {
				return new Message[] { message };
			} else {
				System.err
				.println("ERROR: Could not send message. SendBufferSize is smaller than the minimum required size.");
				return new Message[0];
			}
		}

		/**
		 * Reassembles chopped messages.
		 * 
		 * @param choppedMessage
		 *            The message part to assemble with any earlier received
		 *            parts of the message.
		 * @return the the entire message if it was the last part or null
		 *         otherwise.
		 */
		public Message rebuildChopped(Message choppedMessage) {
			LinkedList headerNodes;
			headerNodes = choppedMessage.getHeaderNodes();
			MessageHeaderNode hNode = (MessageHeaderNode) headerNodes
			.getFirst();
			// Find header node with function Chopped
			while (hNode != null
					&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_CHOPPED) {
				hNode = (MessageHeaderNode) hNode.getNext();
			}
			if (hNode != null) {
				String messageID = hNode.getData()[0];
				int part = Integer.parseInt(hNode.getData()[1]);
				int parts = Integer.parseInt(hNode.getData()[2]);

				if (part == 1) {
					choppedMessageHandler.setFirstMessagePart(messageID,
							choppedMessage.getSenderDeviceID().toString(),
							choppedMessage.getData());
					choppedMessage = null;
				} else if (part == parts) {
					headerNodes.remove(hNode);
					choppedMessage.setHeaderNodes(headerNodes);
					choppedMessage = choppedMessageHandler.setLastMessagePart(
							messageID,
							choppedMessage.getSenderDeviceID().toString(),
							choppedMessage);
				} else {
					choppedMessageHandler.setMessagePart(messageID,
							choppedMessage.getSenderDeviceID().toString(),
							choppedMessage.getData(), part);
					choppedMessage = null;
				}
			}
			return choppedMessage;
		}
	}
	
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}
}
