/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing;

import ist.palcom.collections.LinkedList;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;

/**
 * This class provides the simplest routing manager possible. This routing
 * manager will only provide a translation between URN and to URL.
 * 
 * This implementation will give the same behavior as if no routing manager
 * used. Messages will be send to one hop neighbours only.
 * 
 * @author Jacob Frolund - frolund@daimi.au.dk
 *
 */

public class BasicRoutingManager extends AbstractRoutingManager2 {
	
	private MediaManager manager;
	
	private MessageThread internalThread;
	
	private PalcomThread communicationHandlerThread;
	
	private ChoppedMessageHandler choppedMessageHandler;
	
	public BasicRoutingManager(MediaManager manager, PalcomScheduler scheduler){
		this.manager = manager;
		
		choppedMessageHandler = new ChoppedMessageHandler();
		
		internalThread = new MessageThread();
		scheduler.scheduleThread(internalThread);
		
	}
	
	public void stopReceiving(URL url, PalcomThread receiver){
//		manager.stopReceiving(url, receiver);
	}
	public URL startReceiving(URL url, PalcomThread receiver){
		return null;
//		return manager.startReceiving(url, receiver);
	}
	public void stopReceiving(PalcomThread receiver){
//		manager.stopReceiving(receiver);
	}
	public URL startReceiving(PalcomThread receiver){
		return null;
//		return manager.startReceiving(receiver);
	}
	public Selector startReceiving2(PalcomThread receiver){
		return null;
//		return manager.startReceiving2(receiver);
	}
	public void stopReceiving(Selector selector){
//		manager.stopReceiving(selector);
	}
	public void stopReceiving(URL url){
//		manager.stopReceiving(url);
	}
//	public void send(Message msg, URL url){
//		manager.send(msg, url);
//	}
	public boolean send(Message msg, URN urn1, Selector selector1, URN urn2, Selector selector2){
		return manager.send(msg, urn2.getDeviceURN().toString());
	}
	
	
	//The following methods supports the new wire protocol
	public boolean send(Message message, String receiverDeviceID) {
		Message[] messages = new Message[] {message};
		if(manager.getSendBufferSize() >= 0) {
			messages = choppedMessageHandler.chopMessage(
					manager.getSendBufferSize(), message);
		}
		boolean success = true;
		for(int i = 0; i < messages.length; ++i) {
			if(!manager.send(messages[i], receiverDeviceID)) {
				success = false;
			}
		}
		return success;
	}
	
	public void stop() {
		super.stop();
		manager.stop();
	}
	
	public void addErrorHandler(PalcomThread handler) {
		manager.addErrorHandler(handler);
	}
	
	public void setLocalDeviceCache(String localDeviceCache) {
		manager.setLocalDeviceCache(localDeviceCache);
	}
	
	public String getLocalDeviceCache() {
		return manager.getLocalDeviceCache();
	}

	public void setLocalDeviceID(String localDeviceID) {
		manager.setLocalDeviceID(localDeviceID);
	}

	public void setLocalStatusFlag(byte statusFlag) {
		manager.setLocalStatusFlag(statusFlag);
	}

	public void setLocalDiscoveryVersion(String discoveryVersion) {
		manager.setLocalDiscoveryVersion(discoveryVersion);
	}

	public void setDiscoverySelector(Selector selector) {
		manager.setDiscoverySelector(selector);
	}

	public void sendBroadcast(Message message) {
		Message[] messages = new Message[] {message};
		if(manager.getSendBufferSize() >= 0) {
			messages = choppedMessageHandler.chopMessage(
				manager.getSendBufferSize(), message);
		}
		for(int j = 0; j < messages.length; ++j) {
			manager.sendBroadcast(messages[j]);
		}
	}
	
	/**Send a message saying that the device will close down*/
	public void sendHeartAttack() {
		manager.sendHeartAttack();
	}
	
	/**
	 * Registers as a listener to messages received 
	 * by this manager from the network.
	 * @param communicationManager  The PalcomThread used in CommunicationManager
	 */
	public void registerAsListener(PalcomThread communicationManager) {
		communicationHandlerThread = communicationManager;
		manager.registerAsListener(internalThread);
	}
	
	public void registerForHeartBeats(PalcomThread resourceManager) {
		manager.registerForHeartBeats(resourceManager);
	}
	
	public void sendHeartBeat(boolean isHeartBeatAck) {
		manager.sendHeartBeat(isHeartBeatAck, new LinkedList());
	}
	
	/** Remove the device with ID deviceID from the list of registered devices. */
	public void unregisterDevice(String deviceID) {
		super.unregisterDevice(deviceID);
		manager.unRegisterDevice(deviceID);
	}
	
	/**
	 * This private class is the thread for receiving messages from the 
	 * mediamanager, this thread will distribute messages based on their type
	 * routing/ordinary.
	 */
	private class MessageThread extends PalcomPriorityThread {

		private Event event;

		public MessageThread() {
			event = null;
		}

		public void run() {
			while (true) {
				event = waitEvent();
				Message message = ((MessageEvent) event).getMessage();
				message = choppedMessageHandler.rebuildChopped(message);

				if(message != null) {
					event = new MessageEvent(message, ((MessageEvent)event).getSource());
					communicationHandlerThread.putEvent(event);
				}
			}
		}
	}
}
