/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.routing.test;

import ist.palcom.communication.CommunicationManager;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.discovery.InfoEvent;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.udp.UDPMediaManager;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.resource.manager.ResourceManager;
import ist.palcom.routing.util.AbstractRoutingManager;
import ist.palcom.routing.util.DSDVRoutingManager;
import ist.palcom.routing.util.RoutingPubSubManager;

/**
 * This is a simple sender-service that use the DSDVRoutingManager.
 * 
 * @author jmahle
 *
 */
public class DSDVRoutingSender {

	private MediaManager mediaManager;

	private AbstractRoutingManager routingManager;

	private PointToPointManager pointToPointManager;

	private DiscoveryManager discoveryManager;

	private ResourceManager resourceManager;

	private PRDDevice deviceInfo;

	private final PalcomScheduler scheduler = new PalcomScheduler();

	/**
	 *
	 */
	public DSDVRoutingSender() {
		PalcomThread t = new PalcomPriorityThread() {
			public void run() {
				// Setup
				mediaManager = new UDPMediaManager(scheduler, null);
				//mediaManager = new BluetoothMediaManager(scheduler, null);

				final DeviceID receiverDeviceID = new DeviceID("X:device_10.11.23.191:3000");

				routingManager = new DSDVRoutingManager(scheduler, mediaManager, new RoutingPubSubManager(scheduler, mediaManager), new URN(receiverDeviceID.toString()));
//				CommunicationManager comManager = new CommunicationManager(scheduler, new BasicRoutingManager(mediaManager)); //Without routing
				CommunicationManager comManager = new CommunicationManager(scheduler, routingManager);
				pointToPointManager = new PointToPointManager(comManager);

				System.out.println("DSDV SenderURN: " + receiverDeviceID);
				deviceInfo = new PRDDevice(receiverDeviceID);
				deviceInfo.setName("DSDV Test Sender Service");

				discoveryManager = new DiscoveryManager(scheduler, pointToPointManager, comManager, deviceInfo);
				resourceManager = new ResourceManager(scheduler, pointToPointManager, comManager, deviceInfo);

				// Start discovery and broadcasting messages to DSDV receivers
				discoveryManager.start();
				resourceManager.start();

				System.out.println("Looking for DSDV receiver service");
				PalcomThread infoThread = new DiscoveryThread();
				resourceManager.updatePRDDeviceReceiver(infoThread);
				scheduler.scheduleThread(infoThread);
			}
		};
		scheduler.scheduleThread(t);
		scheduler.run();
	}

	private void receiverFound(PRDDevice device, PRDService service) {
		final DeviceID receiverDeviceID = service.getLocalSID().getDeviceID();
		final Selector receiverSelector = device.getDiscoverySelector();
		System.out.println("DSDV receiver found: " + device.getDeviceID() + " - Service: " + service.getLocalSID());

		// Send heartbeats to the discovered device, using it's DeviceID
		PalcomThread t = new PalcomPriorityThread() {
			public void run() {
				while(true) {
					byte[] data = "testdsdv".getBytes();
					Message message = new Message(data, 0, data.length);
					System.out.println("Sending testdsdv to " + receiverDeviceID.toString());
					boolean sent = pointToPointManager.send(message, deviceInfo.getDiscoverySelector(), receiverDeviceID, receiverSelector, false);

					enablePeriodicEvents(5000);
					waitEvent();
					disablePeriodicEvents();
				}
			}
		};
		scheduler.scheduleThread(t);
	}

	private class DiscoveryThread extends PalcomPriorityThread {
		public void run() {
			PRDDevice receiver = null;
			while (true) {
				InfoEvent event = (InfoEvent) waitEvent();
				if(event.getInfo().getLocalSID() != null) {
					DeviceID deviceID = event.getInfo().getLocalSID().getDeviceID();

					if (deviceID.toString().startsWith("X:device_")) {

						if (event.getInfo() instanceof PRDDevice && receiver == null) {		
							PRDDevice info = (PRDDevice) event.getInfo();
							if (info.getName().startsWith("DSDV Test Receiver")) {
								System.out.println("111");
								resourceManager.updatePRDServiceListReceiver(new LocalSID(info.getDeviceID(), null), this);
								receiver = info;
							}

						} else if (event.getInfo() instanceof PRDServiceList && receiver != null) {
							System.out.println("222");
							PRDServiceList list = (PRDServiceList) event.getInfo();
							if(list.getPRDServiceListItem(0) instanceof PRDService) {
								PRDService serviceInfo = (PRDService)list.getPRDServiceListItem(0);
								receiverFound(receiver, serviceInfo);
							}
							receiver = null;
							//break;
						}
					}
				}
			}   	
		}
	};

	public static void main(String[] args) {
		new DSDVRoutingSender();
	}

}
