/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing.util;


import ist.palcom.collections.LinkedList;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.routing.util.messages.AbstractRoutingMessage;
import ist.palcom.routing.util.messages.DSDVRoutingMessage;
import ist.palcom.routing.util.messages.DSDVUpdateMessage;
import ist.palcom.routing.util.messages.PingMessage;
import ist.palcom.routing.util.messages.PongMessage;

/**
 * 
 * The DSDVRoutingManager is a table-driven routing manager (DSDV tables)
 * where the nodes in the network periodically exchange routing information.
 * The direction of the traffic is chosen based on number of nodes between
 * sender and receiver and not on actually bandwidth for instance.
 * 
 * 
 * @author Jacob Mahler-Andersen, jmahle@daimi.au.dk
 */
public class DSDVRoutingManager extends AbstractRoutingManager {
	
	private DSDVBootstrap listener;
	
	private DSDVRoutine routine;
	
	private DSDVTable table;
	
	private DSDVTable changedPeers;
	
	/**
	 * 
	 * @param manager
	 */
	public DSDVRoutingManager(PalcomScheduler scheduler, MediaManager manager, RoutingPubSubManager pubSubManager, URN deviceURN) {
		super(scheduler, manager, pubSubManager, deviceURN);
	}
	
	/**
	 * Main method
	 */
	protected void startRouting() {
		/* Initialization */
		table = new DSDVTable();
		changedPeers = new DSDVTable();
		
		routine = new DSDVRoutine(this);
		listener = new DSDVBootstrap(this, pubSubManager);
		
		System.out.println("Routing URN is: " + getRoutingURN() + " selector: " + getRoutingSelector());
		
		/* Start DSDV-routine */
		scheduler.scheduleThread(routine.startRoutine());
		
		/* Start DSDVListener - listen for new devices in our subnet */
		scheduler.scheduleThread(listener.startListening());
		scheduler.scheduleThread(listener.startPublish());
		
	}
	
	/**
	 * 
	 */
	public void send(Message message, URL receiverURL) {
		throw new IllegalArgumentException("This message is unsupported in the routing manager");
	}
	
	/**
	 * Send message lookups the receiver URN in the DSDV-table and route the message to the next routingManager.
	 * If the routingManager can't lookup the receiver URN, the message isn't send and the method return false. 
	 */
	public boolean send(Message message, URN senderURN, Selector senderSelector,
            URN receiverURN, Selector receiverSelector) {
		
		URN lookup = receiverURN.getDeviceURN();

		DSDVPeer peer;
		peer = table.getPeer(lookup);
		
		// It's possible to lookup the device URN and route the message
		if (lookup != null && (peer) != null) {
			DSDVRoutingMessage routingMsg = new DSDVRoutingMessage(message, senderURN, senderSelector, receiverURN, receiverSelector);
			return sendDirect(routingMsg, getRoutingURN(), getRoutingSelector(), peer.getNextHopURN(), peer.getNextHopSelector());
			
		} else if (receiverURN.getDeviceURN().equals(getRoutingURN())) {
			DSDVRoutingMessage routingMsg = new DSDVRoutingMessage(message, senderURN, senderSelector, receiverURN, receiverSelector);
			return sendDirect(routingMsg, getRoutingURN(), getRoutingSelector(), receiverURN, receiverSelector);
			
		} else {
			System.out.println("DSDVRoutingManager: Could not find " + receiverURN.getDeviceURN() + " / " + receiverURN + " - tried to send the message anyway: " + message);
		}
		
		return false;
	}
	
	protected boolean sendDirect(Message message, URN senderURN, Selector senderSelector,
            URN receiverURN, Selector receiverSelector) {
		//TODO MessageHeaderNode for routing should be inserted before sending the message.
		return manager.send(message, receiverURN.getDeviceURN().toString());
	}


	/**
	 * This method handle different types of routing messages
	 */
	protected void handleRoutingMessage(AbstractRoutingMessage message) {
    
        if (message instanceof PingMessage) {
        	PingMessage pingMsg = (PingMessage) message;
        	URN urn = pingMsg.getPingURN();
        	Selector selector = pingMsg.getPingSelector();
        	updateStatus(urn);
        	
			// Add the peer
        	addLocalPeer(urn, selector);
        	
        	// Reply with a pong message
        	PongMessage pong = new PongMessage(getRoutingURN(), getRoutingSelector());
        	sendDirect(pong, getRoutingURN(), getRoutingSelector(), urn, selector);
        }
        
        else if (message instanceof PongMessage) {
        	PongMessage pongMsg = (PongMessage) message;
        	URN urn = pongMsg.getPongURN();
        	Selector selector = pongMsg.getPongSelector();

        	updateStatus(urn);
        }
        
        else if (message instanceof DSDVRoutingMessage) {
        	DSDVRoutingMessage routingMsg = (DSDVRoutingMessage) message;
        	URN deviceURN = routingMsg.getReceiverServiceURN().getDeviceURN();
        	updateStatus(deviceURN);
        	
        	/*
        	 * Assign to the right receiver thread on this device
        	 */
        	if (deviceURN.equals(getRoutingURN())) {
        		PalcomThread thread = (PalcomThread) registeredThreads.get(routingMsg.getReceiverServiceSelector());
        		Message m = routingMsg.getMessage();
        		m.setReceiverDeviceID(routingMsg.getReceiverServiceURN().getDeviceURN().toString());
        		m.setReceiverSelector(routingMsg.getReceiverServiceSelector());
        		m.setSenderDeviceID(routingMsg.getDSDVSenderURN().getDeviceURN().toString());
        		m.setSenderSelector(routingMsg.getDSDVSenderSelector());
        		m.setReceiverURL(routingMsg.getReceiverURL());
        		m.setSenderURL(routingMsg.getSenderURL());
       		
        		thread.putEvent(new MessageEvent(m, this));	
        		
        	/*
        	 * Forward the message to next hop
        	 */
        	} else {
        		send(routingMsg.getMessage(), routingMsg.getDSDVSenderURN(), routingMsg.getDSDVSenderSelector(), 
        				routingMsg.getReceiverServiceURN(), routingMsg.getReceiverServiceSelector());
        	}
    	
        }
        
        else if (message instanceof DSDVUpdateMessage) {
        	DSDVUpdateMessage update = (DSDVUpdateMessage) message;
        	updateStatus(update.getDSDVSenderURN());
        	DSDVTable newTable = update.getUpdateDSDVTable();

        	updateTable(table, newTable, update.getDSDVSenderURN(), update.getDSDVSenderSelector());
        }
	}
	
	protected void updateStatus(URN urn) {
		table.updateStatus(true, urn);
	}
	
	/**
	 * Add a peer to the DSDV-table and ensure no duplicates.
	 * Only local peers are added (metric 0).
	 */
	protected void addLocalPeer(URN routingURN, Selector routingSelector) {
		// Synchronize with DSDV-Table
		LinkedList list = table.getPeers();
		DSDVPeer peer = (DSDVPeer) list.getFirst();
		boolean more = peer != null;
		boolean found = false;
		
		/*
		 * Check for duplicates in DSDV-table
		 */
		while (more) {
			if (peer.getRoutingURN().equals(routingURN) && peer.getRoutingSelector().equals(routingSelector)) {
				
				// Update because of a better metric
				if (peer.getMetric() > 0) {
					peer.setMetric(0);
					peer.setNextHop(routingURN, routingSelector);
					peer.setSequenceNumber(peer.getSequenceNumber()+2);
					
					changedPeers.addPeer(peer.copy());
				}
				
				found = true;
				break;
			}
			
			peer = (DSDVPeer) peer.getNext();
			more = peer != null;
		}
		
		/*
		 * No duplicates, add it
		 */
		if (!found) {
			DSDVPeer newpeer = new DSDVPeer(routingURN, routingSelector, routingURN, routingSelector, 0, 0);
			// Send our DSDV-table to the new peer
			
			table.addPeer(newpeer);
			changedPeers.addPeer(newpeer.copy());

			DSDVUpdateMessage update = new DSDVUpdateMessage(getRoutingURN(), getRoutingSelector(), table);
			sendDirect(update, getRoutingURN(), getRoutingSelector(), routingURN, routingSelector);
			
			
		}
	}
	
	/**
	 * This method takes one original DSDVTable and update it with another DSDVTable.
	 * Table entries with highest sequence number and best connection to a peer remains.   
	 */
	protected void updateTable(DSDVTable original, DSDVTable updater, URN updateSenderURN, Selector updateSenderSelector) {
		DSDVPeer peer = (DSDVPeer) updater.getPeers().getFirst();
		boolean more = peer != null;
		while (more) {
			DSDVPeer old = original.getPeer(peer.getRoutingURN());
			
			/*
			 * Unknown not active peer
			 */
			if (old == null && !peer.isActive()) {
				break;
			
				
			/*
			 * Add new peer
			 */
			} else if (old == null && peer.isActive()) {

				if (!peer.getRoutingURN().equals(getRoutingURN())) {
					DSDVPeer newpeer = new DSDVPeer(peer.getRoutingURN(), peer.getRoutingSelector(), updateSenderURN, 
							updateSenderSelector, peer.getMetric() + 1, peer.getSequenceNumber());
					
					original.addPeer(newpeer);
					
					// The DSDVRoutingManager will send our changedPeers to our neighbours 
					changedPeers.addPeer(newpeer.copy());
				}
				
			/*
			 * Greater sequence number - update the peer
			 */
			} else if (old.getMetric() > 0 && peer.getSequenceNumber() > old.getSequenceNumber()) {
				old.setSequenceNumber(peer.getSequenceNumber());
				old.setMetric(peer.getMetric() + 1);
				old.setNextHop(updateSenderURN, updateSenderSelector);
				old.setStatus(peer.getStatus());
				old.setIsActive(peer.isActive());
				changedPeers.addPeer(old.copy());
				
				// Peer is currently not on the network at the moment
				if (!peer.isActive())
					original.deletePeer(old.getRoutingURN(), old.getRoutingSelector());
		
			/*
			 * Equal sequence number
			 */
			} else if (old.getMetric() > 0 && peer.getSequenceNumber() == old.getSequenceNumber()) {
				
				// Better metric?
				if ((peer.getMetric()+1) < old.getMetric()) {
					old.setMetric(peer.getMetric() + 1);
					old.setNextHop(updateSenderURN, updateSenderSelector);
					old.setStatus(peer.getStatus());	
				}
				
				// Peer is currently not on the network at the moment
				if (!peer.isActive()) {
					old.setIsActive(peer.isActive());
					original.deletePeer(old.getRoutingURN(), old.getRoutingSelector());
					changedPeers.addPeer(old.copy());
				}
				
				
			}
			
			peer = (DSDVPeer) peer.getNext();
			more = peer != null;
		}
	}
	
	
	protected DSDVTable getDSDVTable() {
		return table;
	}
	
	
	protected DSDVTable getDSDVChangedPeers() {
		return changedPeers;
	}

}
