/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing.util;


import ist.palcom.collections.LinkedList;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.routing.util.messages.AbstractRoutingMessage;
import ist.palcom.routing.util.messages.SMURFFullAdvertisement;
import ist.palcom.routing.util.messages.SMURFUpdateAdvertisement;

public class SMURFRoutingManager extends AbstractRoutingManager {

	private LinkedList oneHopNeighbours;

	private LinkedList twoHopNeighbours;

	private LinkedList MPRNodes;
	
	private LinkedList nodesLostSinceLastAnnouncement;
	
	private int reluctance = 10;
	
	private int sequence = 0;

	private MPRconstructorAndAnnouncer MPRconstructor;
	
	public SMURFRoutingManager(PalcomScheduler scheduler, MediaManager manager, RoutingPubSubManager pubSubManager, URN deviceURN) {
		super(scheduler, manager, pubSubManager, deviceURN);
	}

	/**
	 * Handle the routing messages used by the SMURF protocol
	 */
	protected void handleRoutingMessage(AbstractRoutingMessage message) {
		//If the node has not been seen before, add it as a one hop neighbour list		
		if(findeNode(message.getSenderURL(), oneHopNeighbours) == null){
			oneHopNeighbours.insertLast(new SMURFNode(message.getSenderURL()));
		}
		
		//Handling Full advertisements
		if (message instanceof SMURFFullAdvertisement) {
			SMURFFullAdvertisement adMessage = (SMURFFullAdvertisement) message;
			//Updating the two hop neighbour list
			updateTwoHopNeighbours(adMessage.getOneHopNeighbours());
			//Updating the MPR list
			updateMPRs(adMessage.getMPRs());
		}

		//Handling update advertisements
		if (message instanceof SMURFUpdateAdvertisement) {
			SMURFUpdateAdvertisement adMessage = (SMURFUpdateAdvertisement) message;
			//Updating the two hop neighbour list
			updateTwoHopNeighbours(adMessage.getOneHopNeighbours());
			//Updating the MPR list
			updateMPRs(adMessage.getMPRs());
			//Removing potential lost nodes from the one hop neighbour list
			removeOneHopNeighbours(adMessage.getLost());
			//Removing potential lost nodes from the two hop neighbour list
			removeTwoHopNeighbours(adMessage.getLost());
			//Removing potential lost MPRs
			removeMPRs(adMessage.getLost());
		}
	}

	public void send(Message message, URL UID) {
		// TODO Auto-generated method stub
	}
	
	public boolean send(Message message, URN senderURN, Selector senderSelector, URN receiverURN, Selector receiverSelector) {
		// TODO Auto-generated method stub
		return false;
	}

	protected void startRouting() {
		MPRconstructor = new MPRconstructorAndAnnouncer();
		scheduler.scheduleThread(MPRconstructor);
	}

	/**
	 * Updating the MPR list with nodesToAdd a node is only represented once.
	 * If the node is in the onehopneighbour or twohopneighbour list it is 
	 * remove from this list.
	 * 
	 * @param nodesToAdd - nodes to add to the MPR list
	 */
	private void updateMPRs(LinkedList nodesToAdd){
		LinkedList oneHopNeighboursToRemove = new LinkedList();
		LinkedList twoHopNeighboursToRemove = new LinkedList();
		SMURFNode node2insert = (SMURFNode) nodesToAdd.getFirst();
		while(node2insert != null){
			if(findeNode(node2insert.getAddress(), oneHopNeighbours) != null){
				oneHopNeighboursToRemove.insertFirst(node2insert);
			}
			if(findeNode(node2insert.getAddress(), twoHopNeighbours) != null){
				twoHopNeighboursToRemove.insertFirst(node2insert);
			}
			if(findeNode(node2insert.getAddress(), MPRNodes) == null){
				MPRNodes.insertLast(node2insert);
			}
			node2insert = (SMURFNode) node2insert.getNext();
		}
	}
	
	/**
	 * Adding the nodes provided in nodesToAdd to the twoHopNeighbours ensuring
	 * that nodes only is repesened once in the list and not adding nodes in the
	 * oneHopNeighbour and MPR list.
	 * 
	 * @param nodesToAdd -
	 *            nodes to add to the twoHopNeighbours linkedlist
	 */
	private void updateTwoHopNeighbours(LinkedList nodesToAdd) {		
		SMURFNode node2insert = (SMURFNode) nodesToAdd.getFirst();
		while (node2insert != null) {
			if (findeNode(node2insert.getAddress(), oneHopNeighbours) == null) {
				if (findeNode(node2insert.getAddress(), MPRNodes) == null) {
					if (findeNode(node2insert.getAddress(), twoHopNeighbours) == null) {
						twoHopNeighbours.insertLast(node2insert);
					}
				}
			}
			node2insert = (SMURFNode) node2insert.getNext();
		}
	}

	/**
	 * Removes the nodes in nodesToRemove from the MPR list.
	 * 
	 * @param node2remove - nodes to be removed from the MPR list
	 */
	private void removeMPRs(LinkedList nodesToRemove) {
		SMURFNode node2remove = (SMURFNode) nodesToRemove.getFirst();
		while(node2remove != null){
			removeNode(node2remove, MPRNodes);
		}
	}
	
	/**
	 * Removes a linked list of nodes from the two hop neighbour list
	 * 
	 * @param nodesToRemove - nodes to be removed
	 */
	private void removeTwoHopNeighbours(LinkedList nodesToRemove) {
		SMURFNode node2remove = (SMURFNode) nodesToRemove.getFirst();
		while(node2remove != null){
			removeNode(node2remove, twoHopNeighbours);
		}
	}

	/**
	 * Inserting the nodes in nodesToAdd into the oneHopNeighbours, ensuring
	 * that nodes are only represented once and that nodes also in the
	 * twoHopNeighbours are removed.
	 * 
	 * @param nodesToAdd -
	 *            nodes to add to the one hop neighbour list.
	 */
	private void updateOneHopNeighbours(LinkedList nodesToAdd) {
		SMURFNode node2insert = (SMURFNode) nodesToAdd.getFirst();
		while (node2insert != null) {
			SMURFNode node2remove;
			if((node2remove = findeNode(node2insert.getAddress(), twoHopNeighbours)) != null){
				removeNode(node2remove, twoHopNeighbours);
			}
			if (findeNode(node2insert.getAddress(), MPRNodes) == null) {
				if (findeNode(node2insert.getAddress(), oneHopNeighbours) == null) {
					oneHopNeighbours.insertLast(node2insert);
				}
			}
		}
	}
	
	/**
	 * Removes a linked list of nodes from the one hop neighbour list
	 * 
	 * @param nodesToRemove - nodes to be removed
	 */
	private void removeOneHopNeighbours(LinkedList nodesToRemove) {
		SMURFNode node2remove = (SMURFNode) nodesToRemove.getFirst();
		while(node2remove != null){
			removeNode(node2remove, oneHopNeighbours);
		}
	}

	/**
	 * Removes the node from a given LinkedList based on the address. If no node
	 * exists nothing is done.
	 * 
	 * @param node -
	 *            the node to remove
	 * @param list -
	 *            the list from where the node should be removed
	 */
	private void removeNode(SMURFNode node, LinkedList list) {
		SMURFNode node2remove = findeNode(node.getAddress(), list);
		if (node2remove != null) {
			list.remove(node2remove);
		}
	}

	/**
	 * Find the node in the linkedlist based on the input address
	 * 
	 * @param address - address of the node to finde.
	 * @param list - list where to finde the node.
	 * @return - null or the node
	 */
	private SMURFNode findeNode(URL address, LinkedList list) {
		SMURFNode node = (SMURFNode) list.getFirst();
		while (node != null) {
			if (node.getAddress().equals(address)) {
				return node;
			}
			node = (SMURFNode) node.getNext();
		}
		return null;
	}
	
	/**
	 * PalcomPriorityThread that calculates the MPR backbone
	 * @author frolund
	 *
	 */
	private class MPRconstructorAndAnnouncer extends PalcomPriorityThread{
		private SMURFNode MPR = null;
		private boolean shutDown = false;
		
		public void stop(){
			shutDown = true;
		}
		
		public void run(){
			while (!shutDown) {
				//Calculating the lates MPR backbone
				if (!oneHopNeighbours.isEmpty() && !twoHopNeighbours.isEmpty()) {
					SMURFNode node = (SMURFNode) oneHopNeighbours.getFirst();
					while (node != null) {
						if (((MPR == null) || (node.getNeighbours().getSize() > MPR
								.getNeighbours().getSize()))
								&& (findeNode(node.getAddress(), MPRNodes) == null)) {
							MPR = node;
						}
					}
					if (MPR != null) {
						MPRNodes.insertLast(MPR);
					}
				}
				//sending announcements to all other nodes
				sequence++;
				SMURFUpdateAdvertisement advertisement = new SMURFUpdateAdvertisement(Integer.toString(sequence), Integer.toString(reluctance), "c", oneHopNeighbours, MPRNodes, nodesLostSinceLastAnnouncement);
				manager.sendBroadcast(advertisement);
				enablePeriodicEvents(100);
				waitEvent();
			}
		}
	}
}
