/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.service.manager;

import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.resource.descriptor.URNLeaf;
import ist.palcom.device.DeviceContext;
import ist.palcom.discovery.CommandEvent;
import ist.palcom.services.AbstractService;
import ist.palcom.services.CommandEventListener;
import ist.palcom.services.CommandListenerThread;

/**
 * UnboundService
 * 
 * A class describing an Unbound Service, i.e., a service fully implemented 
 * in software, that can be loaded into devices.
 * 
 * This class is generic to all UnboundServices, the specific functionality is given
 * by a subclass to AbstractService, which is instantiated when loadService(.) is
 * called. The service class is identified either by giving its qualified name, 
 * or a class reference.
 * 
 * @author sven
 *
 */
public class UnboundService implements CommandEventListener {
    static final String COMMAND_UNLOAD = "unload";
	
	private String serviceClassName;
	private AbstractService service;	
	private DeviceContext myContext;

	
	// A control provider for generic unbound service control
	private AbstractService control;
	private ServiceProxy sccsd;

	private CommandListenerThread inCommandThread;

	private ServiceInstanceID myServiceInstanceID;
	
	private ServiceManager man;



	public UnboundService(String serviceClassName, ServiceManager m) {
		super();
		this.serviceClassName = serviceClassName;
		man = m;
	}
	
	public String getClassName() {
		return serviceClassName;
	}
	
	public ServiceInstanceID getSIID() {
		return myServiceInstanceID;
	}
	
/**
 * load the service of an UnboundService into a device
 * @param context   the device context the service will be added to
 * @param siid   the URN of the service
 * @return true on success
 * 
 */
	public boolean loadService(DeviceContext context, ServiceInstanceID siid, PRDSubList swcList) {
		try {
			myContext = context;
			myServiceInstanceID = siid;
			service = (AbstractService) ist.palcom.base.System.instance()
                    .newInstance(serviceClassName,
                            new Object[] { myContext, siid });
            
//			myContext.addServiceAtDeviceLevel(service);
			addToServiceList(context, swcList);


//			sccsd = new ServiceProxy("Unbound Service Control");
//	        Command quitCmd = new Command(COMMAND_UNLOAD,
//	                Command.DIRECTION_IN);
//	        sccsd.addControlItemAndInfo(quitCmd);

//			control = new AbstractService(myContext, service
//                    .getServiceInfo().getURN().createSubUrn(
//                            new URNLeaf("UScontrol")),
//                    "Unbound Service Control", sccsd);
//            service.addSubservice(control);
//	        inCommandThread = new CommandListenerThread(this);
//	        context.getScheduler().scheduleThread(inCommandThread);
//	        sccsd.addInCommandReceiver(inCommandThread);

//			control.start();
			
			service.start();
			return true;
		} catch (Exception e) {
			e.printStackTrace();
		}
		myContext = null;
		return false;
	}
	
	/** Remove the service from its device 
	 */
	public void unloadService() {
		man.serviceUnloaded(this);
		if(myContext != null) {
			System.out.println("unloadService:"+service);
			try {
				if(control != null) {
					control.stop();
	    	    	//TODO: this can't be right. When a service is removed,
					//there should be some automagic removal of stale ServiceDescriptionInfos
					myContext.getDiscoveryManager()
                            .removeLocalServiceDescription(
                                    (PRDServiceFMDescription) sccsd.getInfo());
				}
				if(service != null)	{
					service.stop();
//					myDM.removeLocalServiceDescriptionInfo((ServiceDescriptionInfo));
					myContext.removeService(service);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		if(inCommandThread != null) {
			inCommandThread.terminate();
		}
	}

//	/** helper method used by services to invoke commands 
//	 * 
//	 * @param commandID
//	 * @param args
//	 */
//	public void doCommand(String commandID, String[] args) {
//		synchronized (Coroutine.class) {
//			Command command = (Command) sd.findItem(commandID);
//			if (command != null) {
//				for (int i = 0; i < args.length; i++) {
//					command.getParam(i).setData(args[i].getBytes());
//				}
//				command.invoke();
//			}
//		}
//	}

	public void commandReceived(CommandEvent e) {
        String command = e.getCommand().getID();
        if (command.equals(COMMAND_UNLOAD)) {
        	unloadService();
        }
	}

	public void addToServiceList(DeviceContext context, PRDSubList swcList) {
		context.addServiceToList(service, swcList.getListLocalSID());
	}
}
