/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.services;

import ist.palcom.collections.LinearMap;
import ist.palcom.collections.LinearSet;
import ist.palcom.connections.Connection;
import ist.palcom.connections.ConnectionEvent;
import ist.palcom.device.DeviceContext;
import ist.palcom.device.Service;
import ist.palcom.discovery.CommandEvent;
import ist.palcom.discovery.InfoEvent;
import ist.palcom.discovery.SelectorEvent;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.Monitor;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.ControlItem;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.xml.XMLException;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class is the basis for a standard service, with things like subservice
 * handling implemented.
 */
public class AbstractService implements Service {

	protected final DeviceContext context;

	/** The info about the service. */
	protected final PRDService info;
	protected ServiceInstanceID siid; //XXX this ought to be final

	private ServiceProxy serviceProxy;

	private ServiceProxyMonitor remoteServiceProxies;

	private ServiceThread serviceThread;

	protected ConnectionsMonitor connectionsMonitor;

	private ControlItemFactory itemFactory;

	/** True when the service is active. */
	private boolean active;

	private LinearMap serviceLocalSIDs;

	private LinearSet existingConnections;

	private Selector selector;

	/*****hgraph impl*****/
	protected HgraphNode _node;
	/*********************/

	/**
	 * Create a new service instance.
	 * 
	 * @param context
	 *            The device context of the service
	 * @param name
	 *            The human-readable name of the service
	 * @param hasDesc
	 *            Does the service have a description
	 * @param remoteConnect
	 *            Will the service act on RemoteConnect messages
	 * @param distribution 
	 * 			  The way the service connects to other services
	 * @param protocol 
	 * 			  The service interaction protocol spoken by the service
	 * @param reliable
	 * 			  Do connecting services have to use reliable communication
	 * @param readableVersionName
	 * 			  A version presentation for the service
	 * @param serviceHelpText
	 * 			  A describing text for the service
	 * @param localSID
	 * 			  A Local ServiceID for the service
	 */
	public AbstractService(DeviceContext context, String name,
			boolean hasDesc, boolean remoteConnect, byte distribution, String protocol, 
			boolean reliable, String readableVersionName, String serviceHelpText, LocalSID localSID) throws IOException {
		this.context = context;

		active = false;

		itemFactory = new ControlItemFactory();

		info = new PRDService(name, localSID, distribution, hasDesc, remoteConnect, 
				protocol, reliable, readableVersionName, serviceHelpText);
		selector = null;

		connectionsMonitor = new ConnectionsMonitor();
		serviceThread = new ServiceThread();

		remoteServiceProxies = new ServiceProxyMonitor(context.getScheduler());

		serviceLocalSIDs = new LinearMap();
		existingConnections = new LinearSet();
		
		_node = new HgraphNode(name);
	}

	public AbstractService(DeviceContext context,
			String serviceName, boolean hasDesc, boolean remoteConnect, byte distribution, String protocol, boolean reliable, 
			String versionName, String helpText, ServiceProxy sp, LocalSID localSID) throws IOException {
		this(context, serviceName, hasDesc, remoteConnect, distribution, 
				protocol, reliable, versionName, helpText, localSID);
		setServiceProxy(sp);
	}
	
	public AbstractService(DeviceContext context,
			String serviceName, boolean hasDesc, boolean remoteConnect, byte distribution, String protocol, boolean reliable, 
			String versionName, String helpText, ServiceProxy sp) throws IOException {
		this(context, serviceName, hasDesc, remoteConnect, distribution, 
				protocol, reliable, versionName, helpText, new LocalSID(context.getDeviceInfo().getDeviceID(),
				context.getNewLocalSIDString()));
		setServiceProxy(sp);
	}
	
	public AbstractService(DeviceContext context,
			String serviceName, boolean hasDesc, boolean remoteConnect, byte distribution, String protocol, boolean reliable, 
			String versionName, String helpText) throws IOException {
		this(context, serviceName, hasDesc, remoteConnect, distribution, 
				protocol, reliable, versionName, helpText, new LocalSID(context.getDeviceInfo().getDeviceID(),
				context.getNewLocalSIDString()));
	}

	public PRDService getServiceInfo() {
		return info;
	}

	public Selector getSelector() {
		return selector;
	}

	/**
	 * Sets the LocalSID service. The LocalSID will be used for uniquely identifying the
	 * service. The LocalSID must contain deviceID.
	 * 
	 * @throws IllegalStateException
	 *             If the LocalSID has already been set.
	 */
	public void setLocalSID(LocalSID localSID) {
		if (info.getLocalSID() != null) {
			throw new IllegalStateException(
			"The LocalSID of a service cannot be set twice.");
		}
		info.setLocalSID(localSID);
	}
	
	/** Sets the ServiceInstanceID for this service. 
	 * The DiscoveryManager will be updated with this info.
	 */
	public void setServiceInstanceID(ServiceInstanceID instanceID) {
		siid = instanceID;
		context.getDiscoveryManager().updateLocalServiceInstanceID(info.getLocalSID(), instanceID);
	}

	//XXX HERE BE DRAGONS! This should *really* be an abstract method
	public ServiceID getServiceID() {
		throw new Error("AbstractService.getServiceID must be overridden");
	}
	
	/**  
	 * @return the ServiceInstanceID
	 */
	public ServiceInstanceID getServiceInstanceID() {
		return siid;
	}
	
	/** Set the status for the service. 
	 * The byte representations for the status can be found in PRDData.
	 */
	public void setStatus(byte status, String helpText) {
		context.getDiscoveryManager().updateServiceStatus(info.getLocalSID(), status, helpText);
	}
	
	/**
	 * Sets the name of the service to name.
	 */
	public void setName(String name) {
		info.setName(name);
	}

	/**
	 * Start the service. The LocalSID must be set before calling this method
	 * (through a constructor, or through a call to the setLocalSID method).
	 * 
	 * If the service is started already, nothing happens.
	 * 
	 * The DiscoveryManager will be updated with info about the service.
	 * 
	 * @throws IllegalStateException
	 *             if the LocalSID has not been set.
	 */
	public void start() throws IOException {
		context.getScheduler().scheduleThread(serviceThread);

		if (selector == null) {
			selector = context.getConnectionsManager().startListening2(
					serviceThread);
		}
		
		if (!active) {
			if (info.getLocalSID() == null) {
				throw new IllegalStateException();
			}
			context.getDiscoveryManager().updateLocalSelector(info.getLocalSID(), selector);
			active = true;
		}
	}

	public void stop() throws IOException {
		if (selector != null) {
			context.getConnectionsManager().stopListening(selector);
		}

		// Update the info
		active = false;
		context.getDiscoveryManager().removeLocalSelector(info.getLocalSID());

		// Close all connections
		Connection[] connections = connectionsMonitor.getConnections();
		for(int i = 0; i < connections.length; ++i) {
			connections[i].close();
		}

		serviceThread.terminate();
	}

	public void setServiceProxy(ServiceProxy proxy) {
		this.serviceProxy = proxy;
		if (proxy != null) {
			proxy.addOutCommandReceiver(serviceThread);
			
			info.setHasDesc(true);
				
			((PRDServiceFMDescription) proxy.getInfo())
			.setLocalSID(info.getLocalSID());
			context.getDiscoveryManager().updateLocalServiceDescription(
					(PRDServiceFMDescription) getServiceProxy().getInfo());
		}
	}

	public ServiceProxy getServiceProxy() {
		return serviceProxy;
	}

	/**
	 * Connect to the service identified by localSID. If there is already a
	 * connection to that service, nothing happens.
	 * 
	 * Subclasses are notified by a call to connectionCreated, and InfoEvents
	 * are sent to registered listeners, once the connection is established (may
	 * happen after this method has returned).
	 * 
	 * ConnectionInfoEvents will be sent to registered listeners if a new
	 * connection is established.
	 * 
	 * @param localSID
	 *            ID of the service to connect to.
	 */
	public void connectTo(final LocalSID localSID) {
		PRDConnection connectionInfo = new PRDConnection(localSID, info.getLocalSID());
		serviceThread.putEvent(new InfoEvent(connectionInfo, true));

		context.getResourceManager2().updateSelectorReceiver(localSID, serviceThread);

		context.getResourceManager2().updatePRDDeviceReceiver(serviceThread);
	}

	/**
	 * Add a listener for received ServiceProxies, for established connections
	 * to providers.
	 */
	public void addServiceProxyListener(PalcomThread listener) {
		remoteServiceProxies.addListener(listener);
	}

	/** Remove a listener for ServiceProxies. */
	public void removeServiceProxyListener(PalcomThread listener) {
		remoteServiceProxies.removeListener(listener);
	}

	protected ServiceProxy getRemoteServiceProxy(LocalSID localSID) {
		return remoteServiceProxies.getServiceProxy(localSID);
	}

	protected Iterator remoteServiceProxyIterator() {
		return remoteServiceProxies.serviceProxyIterator();
	}

	/**
	 * Closes the connection to the given provider and removes it from the list
	 * of connections. InfoEvents are sent to registered listeners.
	 * serviceLocalSID must include deviceID.
	 */
	public void disconnectFrom(LocalSID serviceLocalSID) throws IOException {
		int index = serviceLocalSIDs.values().indexOf(serviceLocalSID);

		Connection connection = null;
		if(index > -1) {
			LinearSet set = serviceLocalSIDs.keySet();
			String localSelector = (String)(set.toArray())[index];
			connection = connectionsMonitor.removeConnection(localSelector);
			existingConnections.remove(serviceLocalSID);
			serviceLocalSIDs.remove(localSelector.toString());
		}


		if (connection != null) {
			connection.close();

			PRDConnection connectionInfo = new PRDConnection(serviceLocalSID,
					this.info.getLocalSID());
			context.getDiscoveryManager().removeLocalConnectionInfo(connectionInfo);

			ServiceProxy sp = remoteServiceProxies
			.getServiceProxy(serviceLocalSID);

			if (sp != null) {
				remoteServiceProxies.removeServiceProxy(serviceLocalSID);
				fireServiceProxy(sp, false);
			} 

		}
	}

	/**
	 * Closes the connection to the given provider and removes it from the list
	 * of connections.
	 */
	private void disconnectFrom(String localSelector) throws IOException {

		Connection connection = connectionsMonitor.removeConnection(localSelector);
		if (connection != null) {
			connection.close();

		}
		serviceLocalSIDs.remove(localSelector.toString());
	}

	/**
	 * Closes all connections to the given provider and removes them from the list
	 * of connections. InfoEvents are sent to registered listeners.
	 */
	private void disconnectAllFromDevice(DeviceID deviceID) throws IOException {
		LocalSID serviceLocalSID = null;
		ArrayList urns = serviceLocalSIDs.values();
		for(int index = 0; index < urns.size(); ++index) {
			serviceLocalSID = (LocalSID)urns.get(index);
			if(deviceID.equals(serviceLocalSID.getDeviceID())) {

				LinearSet set = serviceLocalSIDs.keySet();
				Object[] selectors = set.toArray();
				if(index < selectors.length) {
					String localSelector = (String)selectors[index];
					Connection connection = connectionsMonitor.removeConnection(localSelector);
					existingConnections.remove(serviceLocalSID);
					serviceLocalSIDs.remove(localSelector.toString());
					urns = serviceLocalSIDs.values();
					index--;

					if (connection != null) {
						connection.close();

						PRDConnection connectionInfo = new PRDConnection(serviceLocalSID,
								this.info.getLocalSID());
						context.getDiscoveryManager().removeLocalConnectionInfo(connectionInfo);

						ServiceProxy sp = remoteServiceProxies
						.getServiceProxy(serviceLocalSID);

						if (sp != null) {
							remoteServiceProxies.removeServiceProxy(serviceLocalSID);
							fireServiceProxy(sp, false);
						} 
					}
				} else {
					System.err.println("Could not find selector for deviceID: " + deviceID);
				}
			}
		}
	}

	protected void sendToAll(Message message) {
		Connection[] connections = connectionsMonitor.getConnections();
		for (int i = 0; i < connections.length; i++) {
			Connection connection = connections[i];
			if (connection != null) {
				connection.sendMessage(message);
			}
		}
	}

	/**
	 * Notification that a new connection has been established. Service-specific
	 * things may be performed by subclasses. The default implementation here is
	 * empty.
	 */
	protected void connectionCreated(Connection connection) {
	}

	protected Connection getConnectionForLocalSID(LocalSID serviceLocalSID) {
		int index = serviceLocalSIDs.values().indexOf(serviceLocalSID);

		Connection connection = null;
		if(index > -1) {
			LinearSet set = serviceLocalSIDs.keySet();
			String localSelector = (String)(set.toArray())[index];
			connection = connectionsMonitor.getConnection(localSelector);
		}
		return connection;
	}

	/**
	 * Method for testing.
	 * 
	 * @return All connections to customers.
	 */
	protected Connection[] getConnections() {
		return connectionsMonitor.getConnections();
	}

	/**
	 * Notifies all listeners about a ServiceProxy.
	 */
	private void fireServiceProxy(ServiceProxy sp, boolean active) {
		PalcomThread[] listeners = remoteServiceProxies.getListenerThreads();
		for (int i = 0; i < listeners.length; i++) {
			listeners[i].putEvent(new ServiceProxyEvent(sp, active));
		}
	}

	/**
	 * Method for testing. Blocks until the number of connections becomes equal
	 * to nbr.
	 */
	public void waitUntilNumberOfConnectionsIs(int nbr) {
		connectionsMonitor.waitUntilNumberOfConnectionsIs(nbr);
	}

	private class ServiceThread extends PalcomPriorityThread {
		public void run() {
			ArrayList pendingSendingOpen = new ArrayList();
			ArrayList pendingOpenReply = new ArrayList();
			while (true) {
				Event event = waitEvent();
				if (event instanceof MessageEvent) {
					MessageEvent messageEvent = (MessageEvent) event;
					Message message = messageEvent.getMessage();
					try {
						ControlItem item = itemFactory
						.createFromMessage(message);
						if (item instanceof Command) {
							Command command = (Command) item;
							if (command.getDirection().equals(
									Command.DIRECTION_IN)) {
								if (getServiceProxy() != null) {
									getServiceProxy().update(command);
								}
							} else if (command.getDirection().equals(
									Command.DIRECTION_OUT)) {
								// Update the right ServiceProxy with the
								// received OutCommand
								Connection source = (Connection) messageEvent
								.getSource();

								LocalSID providerLocalSID = (LocalSID)serviceLocalSIDs.get(
										source.getLocalSelector().toString());

								ServiceProxy sp = remoteServiceProxies
								.getServiceProxy(providerLocalSID);
								if (sp != null) {
									sp.update(command);
								}
							}
						}
					} catch (XMLException e1) {
						e1.printStackTrace();
					}
				} else if (event instanceof CommandEvent) {
					CommandEvent commandEvent = (CommandEvent) event;
					Command command = commandEvent.getCommand();
					if (command.getDirection().equals(Command.DIRECTION_OUT)) {
						Message message = command.getMessage();
						sendToAll(message);
					} else if (command.getDirection().equals(
							Command.DIRECTION_IN)) {
						// Write the InCommand, received from a ServiceProxy,
						// to the right connection
						ServiceProxy source = (ServiceProxy) commandEvent
						.getSource();
						Connection connection = getConnectionForLocalSID(source.getLocalSID());
						if (connection != null) {
							Message message = commandEvent.getCommand()
							.getMessage();
							connection.sendMessage(message);
						}
					}
				} else if (event instanceof InfoEvent) {
					InfoEvent infoEvent = (InfoEvent) event;
					PRDData info = infoEvent.getInfo();
					if (info instanceof PRDConnection) {
						PRDConnection connectionInfo = (PRDConnection) info;
						if(!pendingSendingOpen.contains(connectionInfo)) {
							pendingSendingOpen.add(connectionInfo);
						}
					} else if (info instanceof PRDServiceFMDescription) {
						if (infoEvent.infoShouldBeKept()) {
//							System.err.println("AbstractService:["+getServiceInfo().getName()+"] got PRDServiceFMD: ");
							PRDServiceFMDescription sd = (PRDServiceFMDescription) infoEvent
							.getInfo();
							if (!remoteServiceProxies
									.hasServiceProxy(sd.getLocalSID())) {
								ServiceProxy sp = (ServiceProxy) sd
								.createInstance();
								remoteServiceProxies.putServiceProxy(
										sd.getLocalSID(), sp);
								sp.addInCommandReceiver(this);
								fireServiceProxy(sp, true);
								context.getResourceManager2()
								.removePRDServiceFMDescriptionReceiver(
										sd.getLocalSID(), this);
							}
						}
					} else if (info instanceof PRDDevice) {
						//Only interested in devices that has shut down.
						if(!infoEvent.infoShouldBeKept()) {
							PRDDevice deviceInfo = (PRDDevice) info;
							try {
								disconnectAllFromDevice(deviceInfo.getDeviceID());
							} catch (IOException e) {
								// Ignore exceptions on close
							}
						}
					} else if(info instanceof PRDService) {
						PRDService service = (PRDService) info;
						if (service.getHasDesc()) {
							context.getResourceManager2()
							.updatePRDServiceFMDescriptionReceiver(
									service.getLocalSID(), this);
						}
					}
				} else if (event instanceof SelectorEvent) {
					SelectorEvent selEvent = (SelectorEvent) event;
					for (int i = 0; i < pendingSendingOpen.size(); i++) {
						PRDConnection connectionInfo = (PRDConnection) pendingSendingOpen
						.get(i);
						LocalSID providerLocalSID = connectionInfo.getLocalSID1();
						LocalSID selectorLocalSID = selEvent.getLocalSID();
						if (selectorLocalSID != null && selectorLocalSID.equals(providerLocalSID)) {
							Selector connectionSelector = selEvent.getSelector();
							if (!existingConnections.contains(providerLocalSID) && 
									!pendingOpenReply.contains(connectionInfo)) {
								// Connect to the service

								Selector localSelector = context.getConnectionsManager().open(
										providerLocalSID.getDeviceID(), connectionSelector,
										this);
								pendingSendingOpen.remove(i);
								pendingOpenReply.add(connectionInfo);
								serviceLocalSIDs.put(localSelector.toString(), providerLocalSID);
							} else {
								pendingSendingOpen.remove(i);
							}
						}
					}
				} else if (event instanceof ConnectionEvent) {
//					System.err.println("AbstractService:["+getServiceInfo().getName()+"] got ConnectionEvent");
					ConnectionEvent connectionEvent = (ConnectionEvent) event;
					Connection connection = connectionEvent.getConnection();

					LocalSID remoteLocalSID = (LocalSID)serviceLocalSIDs.get(
							connection.getLocalSelector().toString());
//					System.err.println("AbstractService:["+getServiceInfo().getName()+"] LocalSID =="+remoteLocalSID);					
					if(remoteLocalSID != null) {
						existingConnections.add(remoteLocalSID);
					} else {
						remoteLocalSID = new LocalSID(connection.getRemoteDeviceID(), null);
					}
					if (connectionEvent.isConnected()) {
						connectionsMonitor.updateConnection(
								connection.getLocalSelector().toString(), connection);
						connection.addDisconnectReceiver(this);
						
						//TODO Now there are only messages but when StreamInfos starts to appear a check will have to be made here.
						connection.addMessageReceiver(serviceThread);
						
						connectionCreated(connection);
						PRDConnection connectionInfo = new PRDConnection(
								remoteLocalSID, info.getLocalSID());
						if (pendingOpenReply.contains(connectionInfo)) {
//							System.err.println("AbstractService:["+getServiceInfo().getName()+"] handling CI: "+connectionInfo);
//							if (remoteServiceProxies
//									.hasServiceProxy(remoteLocalSID)) {
//								ServiceProxy sp = remoteServiceProxies.getServiceProxy(remoteLocalSID);
//								fireServiceProxy(sp, true);
//							}
							context.getResourceManager2()
							.updatePRDServiceFMDescriptionReceiver(remoteLocalSID, this);
							context.getDiscoveryManager()
							.updateLocalConnectionInfo(connectionInfo);
							pendingOpenReply.remove(connectionInfo);
						} else {
//							System.err.println("AbstractService:["+getServiceInfo().getName()+"] ignoring CI: "+connectionInfo);
//							Iterator ii = pendingOpenReply.iterator();
//							while(ii.hasNext()) {
//								System.err.println("..."+ii.next());
//							}
						}
					} else {
						// A disconnect

						ServiceProxy sp = remoteServiceProxies
						.getServiceProxy(remoteLocalSID);

						if (sp != null) {
							remoteServiceProxies.removeServiceProxy(remoteLocalSID);
							fireServiceProxy(sp, false);
						} 

						try {
							if(!remoteLocalSID.getDeviceID().equals(connection.getRemoteDeviceID())) {
								disconnectFrom(remoteLocalSID);
							} else {
								disconnectFrom(connection.getLocalSelector().toString());
							}
						} catch (IOException e) {
							// Ignore exceptions on close
						}


					}
				}
			}
		}
	}

	private class ConnectionsMonitor extends Monitor {

		/** Map with connections to services, indexed by service URN. */
		private LinearMap connections;

		private Condition modified = new Condition();

		ConnectionsMonitor() {
			super(context.getScheduler());
			connections = new LinearMap();
		}

		void updateConnection(String connectionKey, Connection connection) {
			enterMonitor();
			try {
				connections.put(connectionKey, connection);
				modified.causeAll();
			} finally {
				exitMonitor();
			}
		}

		Connection removeConnection(String connectionKey) {
			enterMonitor();
			try {
				Connection connection = (Connection) connections
				.remove(connectionKey);
				modified.causeAll();
				return connection;
			} finally {
				exitMonitor();
			}
		}

		Connection getConnection(String connectionKey) {
			enterMonitor();
			try {
				return (Connection) connections.get(connectionKey);
			} finally {
				exitMonitor();
			}
		}

		boolean connectionExists(String connectionKey) {
			enterMonitor();
			try {
				return connections.containsKey(connectionKey);
			} finally {
				exitMonitor();
			}
		}

		public String[] getServiceConnectionKeys() {
			enterMonitor();
			try {
				Object[] array = connections.keySet().toArray();
				String[] connectionKeys = new String[array.length];
				for (int i = 0; i < array.length; i++) {
					connectionKeys[i] = (String) array[i];
				}
				return connectionKeys;
			} finally {
				exitMonitor();
			}
		}

		Connection[] getConnections() {
			enterMonitor();
			try {
				Connection[] array = new Connection[connections.size()];
				for (int i = 0; i < array.length; i++) {
					array[i] = (Connection) connections.values().get(i);
				}
				return array;
			} finally {
				exitMonitor();
			}
		}

		void waitUntilNumberOfConnectionsIs(int nbr) {
			enterMonitor();
			try {
				while (connections.size() != nbr) {
					modified.await();
				}
			} finally {
				exitMonitor();
			}
		}
	}

	private class ServiceProxyMonitor extends Monitor {

		/** The list of registered ServiceProxy listener threads. */
		private ArrayList spListenerThreads;

		/** A HashMap with mappings from provider LocalSIDs to ServiceProxies. */
		// private HashMap<String, ServiceProxy> urnToSPMap;
		private LinearMap localSIDToSPMap;

		public ServiceProxyMonitor(PalcomScheduler scheduler) {
			super(scheduler);
			spListenerThreads = new ArrayList();
			// urnToSPMap = new HashMap<String, ServiceProxy>();
			localSIDToSPMap = new LinearMap();
		}

		void putServiceProxy(LocalSID providerLocalSID, ServiceProxy sp) {
			enterMonitor();
			try {
				System.err.println("AbstractService["+getServiceInfo().getName()+"].putServiceProxy: "+providerLocalSID);
				localSIDToSPMap.put(providerLocalSID, sp);
			} finally {
				exitMonitor();
			}
		}

		void removeServiceProxy(LocalSID providerLocalSID) {
			enterMonitor();
			try {
				localSIDToSPMap.remove(providerLocalSID);
			} finally {
				exitMonitor();
			}
		}

		ServiceProxy getServiceProxy(LocalSID providerLocalSID) {
			enterMonitor();
			try {
				return (ServiceProxy) localSIDToSPMap.get(providerLocalSID);
			} finally {
				exitMonitor();
			}
		}

		boolean hasServiceProxy(LocalSID providerLocalSID) {
			enterMonitor();
			try {
				return localSIDToSPMap.containsKey(providerLocalSID);
			} finally {
				exitMonitor();
			}
		}

		void addListener(PalcomThread listener) {
			enterMonitor();
			try {
				spListenerThreads.add(listener);
			} finally {
				exitMonitor();
			}
		}

		void removeListener(PalcomThread listener) {
			enterMonitor();
			try {
				spListenerThreads.remove(listener);
			} finally {
				exitMonitor();
			}
		}

		PalcomThread[] getListenerThreads() {
			enterMonitor();
			try {
				Object[] array = spListenerThreads.toArray();
				PalcomThread[] threads = new PalcomThread[array.length];
				for (int i = 0; i < array.length; i++) {
					threads[i] = (PalcomThread) array[i];
				}
				return threads;
			} finally {
				exitMonitor();
			}
		}

		Iterator serviceProxyIterator() {
			return localSIDToSPMap.values().iterator();
		}

	}


	/****************hgraph impl below********************/
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}
}