/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.simulated.device;

import ist.palcom.device.AbstractDevice;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.URL;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JTextArea;

/**
 * This abstract class is intended to be used as the basis for a simulated
 * device.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public abstract class AbstractSimulatedDeviceGUI extends JFrame implements SimulatedDeviceGUI, 
		ActionListener {

    private AbstractDevice device;

    protected final DeviceDescriptor descriptor;

    protected AbstractSimulatedDeviceGUI(DeviceID deviceID) {
        descriptor = DeviceDescriptor.createFromPropertiesFile(deviceID);
        descriptor.type = getClass().getName();

        device = createDevice(deviceID);
        
        //Insert a help menu with an about item
        JMenuBar menuBar = new JMenuBar();
        JMenu menu = new JMenu("Help");
        JMenuItem aboutItem = new JMenuItem("About");
        aboutItem.addActionListener(this);
        menu.add(aboutItem);
        menuBar.setOpaque(true);
        menuBar.add(menu);
        this.setJMenuBar(menuBar);
        
        addWindowListener(new WindowClosingHandler());
    }

    protected abstract AbstractDevice createDevice(DeviceID deviceID);

    public AbstractDevice getDevice() {
        return device;
    }

    public void setName(String name) {
        if (device != null) {
            device.setName(name);
        }
        setTitle(name);
        descriptor.name = name;
        descriptor.writeToPropertiesFile();
    }

    public String getName() {
        return descriptor.name;
    }

    /**
     * A WindowListener that will dispose the JFrame and close the
     * AbstractDevice as the window closes.
     */
    public class WindowClosingHandler extends WindowAdapter {
        public void windowClosing(WindowEvent event) {
            dispose();
            if (device != null) {
                device.stop();
                device = null;
            }
        }

        public void windowClosed(WindowEvent event) {
            if (device != null) {
                device.stop();
                device = null;
            }
        }
    }
    
    public JFrame getGUI() {
    	return this;
    }
    
    public void close() {dispose();}
    
    public void actionPerformed(ActionEvent e) {
    	JDialog aboutDialog = new JDialog(this);
		
		aboutDialog.setTitle("About");
		
		JTextArea textArea = new JTextArea();
		textArea.setEditable(false);
		
		URL[] urls = ((AbstractMediaManager)device.getDeviceContext().getMediaManager())
		.getDefaultGroupAddresses();
		
		String addresses = "";
		for(int i = 0; i < urls.length; ++i) {
			if(i == 0) {
				addresses = "" + urls[i];
			} else {
				addresses = addresses + System.getProperty("line.separator") + "   " + urls[i];
			}
		}
		
		
		textArea.setText("Device information:" + System.getProperty("line.separator") + 
				System.getProperty("line.separator") + "Device name: " + device.getDeviceContext().getDeviceInfo().getName() + 
				System.getProperty("line.separator") + "Device ID: " + device.getDeviceContext().getDeviceInfo().getDeviceID() + 
				System.getProperty("line.separator") + "Wire protocol version: " + AbstractMediaManager.WIRE_PROTOCOL_VERSION + 
				System.getProperty("line.separator") + "Discovery protocol version: " + DiscoveryManager.DISCOVERY_VERSION + 
				System.getProperty("line.separator") + "Discovery on: " + addresses);
		
		aboutDialog.add(textArea);
		
		Point loc = getLocation();
		aboutDialog.setLocation(loc.x, loc.y);
		
		aboutDialog.pack();
		aboutDialog.setVisible(true);
    }
    
}