/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.simulated.device.factory;

import ist.palcom.device.AbstractDevice;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.simulated.device.AbstractSimulatedDeviceGUI;
import ist.palcom.simulated.device.DeviceDescriptor;
import ist.palcom.simulated.device.SimulatedDeviceGUI;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

/**
 * Abstract base class for a factory that can create simulated devices with
 * persistent identities (identities that survive computer reboots).
 * <p>
 * There is a static method {@link #createAndStart(Class) createAndStart(Class)}
 * for creating and starting a new device.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public abstract class AbstractDeviceFactory {

    private static final File SCENARIOS_FILE = new File(
            AbstractSimulatedDeviceGUI.DEVICES_DIRECTORY, ".scenarios");

    private ArrayList<DeviceDescriptor> deviceDescriptors;

    private ArrayList<ScenarioDescriptor> scenarioDescriptors;

    /** URL to resource where device GUI class names can be read. */
    private URL deviceClassNameURL;

    /** Array of device GUI class names (computed lazily). */
    private List<String> deviceGUIClassNames;

    /**
     * Map from device device IDs to GUIs for devices, for devices that have been
     * started and are running, or have been stopped and not yet removed from
     * the map.
     */
    private HashMap<DeviceID, SimulatedDeviceGUI> deviceGUIs;

    /**
     * @param classNameURL
     *            URL to resource where device GUI class names can be read.
     */
    public AbstractDeviceFactory(URL classNameURL) {
        deviceClassNameURL = classNameURL;
        deviceDescriptors = new ArrayList<DeviceDescriptor>();
        scenarioDescriptors = new ArrayList<ScenarioDescriptor>();
        deviceGUIs = new HashMap<DeviceID, SimulatedDeviceGUI>();

        try {
            readDeviceDescriptorsFromFiles();
            readScenarioDescriptorsFromFile();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static void createAndStart(Class deviceType)
            throws DeviceFactoryException {
        new AbstractDeviceFactory(null) {
        }.createAndStartNewDevice(deviceType.getName());
    }

    private void readDeviceDescriptorsFromFiles() throws IOException {
        File devicesDirectory = new File(
                AbstractSimulatedDeviceGUI.DEVICES_DIRECTORY);
        File[] deviceDirectories = devicesDirectory.listFiles(new FileFilter() {
            public boolean accept(File pathname) {
                return pathname.isDirectory();
            }
        });
        if (deviceDirectories != null) {
            Arrays.sort(deviceDirectories);
            for (int i = 0; i < deviceDirectories.length; i++) {
                try {
                    DeviceID deviceID = new DeviceID(URLDecoder.decode(
                            deviceDirectories[i].getName(), "UTF-8"));
                    DeviceDescriptor descriptor = DeviceDescriptor
                            .createFromPropertiesFile(deviceID);
                    if (!getDeviceGUIClassNames().contains(descriptor.type)) {
                        // Show error/informational message
                        // System.err.println("Class " + descriptor.type + " for
                        // "
                        // + descriptor.urn + " is not available.");
                    }
                    // Add the descriptor in any case
                    deviceDescriptors.add(descriptor);
                } catch (UnsupportedEncodingException e) {
                    // Should not happen with UTF-8...
                    e.printStackTrace();
                }
            }
        }
    }

    private void readScenarioDescriptorsFromFile() {
        try {
            FileInputStream fileInputStream = new FileInputStream(
                    SCENARIOS_FILE);
            BufferedReader reader = new BufferedReader(new InputStreamReader(
                    fileInputStream));
            ScenarioDescriptor descriptor = ScenarioDescriptor
                    .createFromReader(reader);
            while (descriptor != null) {
                scenarioDescriptors.add(descriptor);
                descriptor = ScenarioDescriptor.createFromReader(reader);
            }
            fileInputStream.close();
        } catch (FileNotFoundException e) {
            // There was no file, don't read
        } catch (IOException e) {
            // Just continue...
        }
    }

    private void writeScenarioDescriptorsToFile() {
        try {
            FileOutputStream fileOutputStream = new FileOutputStream(
                    SCENARIOS_FILE);
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(
                    fileOutputStream));
            for (int i = 0; i < scenarioDescriptors.size(); i++) {
                ScenarioDescriptor descriptor = scenarioDescriptors.get(i);
                descriptor.writeToWriter(writer);
            }
            fileOutputStream.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private int getScenarioDescriptorIndex(String name) {
        for (int i = 0; i < scenarioDescriptors.size(); i++) {
            ScenarioDescriptor descriptor = (ScenarioDescriptor) scenarioDescriptors
                    .get(i);
            if (descriptor.getName().equals(name)) {
                return i;
            }
        }
        return -1;
    }

    private ScenarioDescriptor getScenarioDescriptor(String name) {
        int index = getScenarioDescriptorIndex(name);
        if (index == -1) {
            return null;
        } else {
            return scenarioDescriptors.get(index);
        }
    }

    public boolean scenarioExists(String name) {
        return getScenarioDescriptorIndex(name) != -1;
    }

    public void createScenario(String name) {
        ScenarioDescriptor descriptor = new ScenarioDescriptor(name);
        scenarioDescriptors.add(descriptor);
        writeScenarioDescriptorsToFile();
    }

    public void deleteScenario(String name) {
        int index = getScenarioDescriptorIndex(name);
        if (index != -1) {
            scenarioDescriptors.remove(index);
            writeScenarioDescriptorsToFile();
        }
    }

    /**
     * Adds the device to the scenario. If the device is already in the
     * scenario, it won't be added again.
     */
    public void addDeviceToScenario(DeviceID deviceID, String scenarioName) {
        ScenarioDescriptor scenarioDescriptor = getScenarioDescriptor(scenarioName);
        if (scenarioDescriptor != null) {
            scenarioDescriptor.addDeviceID(deviceID);
            writeScenarioDescriptorsToFile();
        }
    }

    public void removeDeviceFromScenario(DeviceID deviceID, String scenarioName) {
        ScenarioDescriptor scenarioDescriptor = getScenarioDescriptor(scenarioName);
        if (scenarioDescriptor != null) {
            scenarioDescriptor.removeDeviceID(deviceID);
            writeScenarioDescriptorsToFile();
        }
    }

    public int getNumberOfScenarios() {
        return scenarioDescriptors.size();
    }

    public String getScenarioName(int index) {
        ScenarioDescriptor descriptor = scenarioDescriptors.get(index);
        return descriptor.getName();
    }

    public Iterator<DeviceID> getDeviceIDIterator(String scenarioName) {
        ScenarioDescriptor descriptor = getScenarioDescriptor(scenarioName);
        return descriptor.getDeviceIDIterator();
    }

    public int getNumberOfDevices() {
        return deviceDescriptors.size();
    }

    public int getDeviceDescriptorIndex(DeviceID deviceID) {
        for (int i = 0; i < deviceDescriptors.size(); i++) {
            DeviceDescriptor descriptor = deviceDescriptors.get(i);
            if (descriptor.deviceID.equals(deviceID)) {
                return i;
            }
        }
        return -1;
    }

    public DeviceDescriptor getDeviceDescriptor(int index) {
        return deviceDescriptors.get(index);
    }

    public DeviceDescriptor getDeviceDescriptor(DeviceID deviceID) {
        int index = getDeviceDescriptorIndex(deviceID);
        if (index == -1) {
            return null;
        } else {
            return deviceDescriptors.get(index);
        }
    }

    private SimulatedDeviceGUI getDeviceGUI(DeviceID deviceID) {
        return deviceGUIs.get(deviceID);
    }

    private File getDeviceDir(DeviceID deviceID) {
        try {
            return new File(AbstractSimulatedDeviceGUI.DEVICES_DIRECTORY,
                    URLEncoder.encode(deviceID.toString(), "UTF-8"));
        } catch (UnsupportedEncodingException e) {
            // Should not happen with UTF-8...
            e.printStackTrace();
            return null;
        }
    }

    public DeviceID getNewDeviceID() {
        DeviceID deviceID;
        do {
            deviceID = new DeviceID("C:" + UUID.randomUUID().toString());
        } while (getDeviceDir(deviceID).exists());
        return deviceID;
    }

    /** Creates a device, without starting it. */
    public void createDevice(DeviceDescriptor descriptor) {
        descriptor.writeToPropertiesFile();
        deviceDescriptors.add(descriptor);
    }

    public List<String> getDeviceGUIClassNames() throws IOException {
        if (deviceGUIClassNames == null) {
            // Compute GUI class names lazily

            deviceGUIClassNames = new ArrayList<String>();
            if (deviceClassNameURL != null) {
                BufferedReader reader = new BufferedReader(
                        new InputStreamReader(deviceClassNameURL.openStream()));
                String line = reader.readLine();
                while (line != null) {
                    line = line.trim(); // Remove leading and trailing
                    // whitespace
                    if (line.length() > 0 && !line.startsWith("#")) {
                        // Not an empty line or a comment
                        String className = line;

                        // Add the classname to the list if the class is
                        // available
                        try {
                            Class clazz = Class.forName(className);
                            deviceGUIClassNames.add(className);
                        } catch (ClassNotFoundException e) {
                            // Show error/information message
                            System.err.println("Device GUI class " + className
                                    + " is not on the classpath.");
                        }
                    }
                    line = reader.readLine();
                }
            }
        }
        return deviceGUIClassNames;
    }

    public void deleteDevice(DeviceID deviceID) throws IOException {
        stopDevice(deviceID); // Just in case

        int index = getDeviceDescriptorIndex(deviceID);
        if (index != -1) {
            deviceDescriptors.remove(index);
        }

        deleteRecursively(getDeviceDir(deviceID));
    }

    /**
     * Deletes file, and, if file is a directory, all files contained inside,
     * recursively.
     */
    private void deleteRecursively(File file) {
        if (file.isDirectory()) {
            File[] files = file.listFiles();
            for (int i = 0; i < files.length; i++) {
                deleteRecursively(files[i]);
            }
        }
        file.delete();
    }

    /**
     * Start the device with ID deviceID, if it exists and if it is not
     * already running.
     * 
     * @return The AbstractSimulatedDeviceGUI of the started device, or of the
     *         already started device, or null if no device was started.
     * 
     * @throws DeviceFactoryException
     *             if the device cannot be started, e.g. if the stored GUI type
     *             is wrong.
     */
    public SimulatedDeviceGUI startDevice(DeviceID deviceID)
            throws DeviceFactoryException {
        SimulatedDeviceGUI gui = getDeviceGUI(deviceID);
        if (gui == null || !gui.getGUI().isVisible()) {
            deviceGUIs.remove(deviceID); // Clean up if stopped
            DeviceDescriptor descriptor = getDeviceDescriptor(deviceID);
            if (descriptor != null && descriptor.type != null) {
                // Try to construct an AbstractDevice object by calling a
                // constructor with 1 String parameter (the deviceID)
                try {
                    Class clazz = Class.forName(descriptor.type);
                    if (!SimulatedDeviceGUI.class.isAssignableFrom(clazz)) {
                        throw new DeviceFactoryException(clazz.getName()
                                + " does not implement "
                                + SimulatedDeviceGUI.class.getName()
                                + ". Please recreate the device.");
                    }
                    Class[] parameterTypes = new Class[] { DeviceID.class };
                    Constructor constructor = clazz
                            .getConstructor(parameterTypes);
                    gui = (SimulatedDeviceGUI) constructor
                            .newInstance(new Object[] { descriptor.deviceID });

                    deviceGUIs.put(deviceID, gui);

                    // Set the name of the device and start it in a separate
                    // thread
                    gui.setName(descriptor.name);
                    gui.getGUI().setVisible(true);
                    final AbstractDevice device = gui.getDevice();
                    Thread thread = new Thread() {
                        public void run() {
                            device.run();
                        }
                    };
                    thread.start();
                } catch (NoSuchMethodException e) {
                    throw new DeviceFactoryException(
                            "Found no public constructor for "
                                    + descriptor.type
                                    + " with 1 String parameter.");
                } catch (Exception e) {
                    e.printStackTrace();
                    throw new DeviceFactoryException(e.getMessage());
                }
            } else {
                throw new DeviceFactoryException(
                        "Found no stored descriptor for " + deviceID);
            }
        }
        return gui;
    }

    /**
     * Starts a new device with a given type and a random name and device ID.
     * 
     * @throws DeviceFactoryException
     *             if the device cannot be started.
     */
    public void createAndStartNewDevice(String deviceType)
            throws DeviceFactoryException {
        DeviceID deviceID = getNewDeviceID();

        int lastDotPos = deviceType.lastIndexOf('.');
        String name = deviceType.substring(lastDotPos + 1)
                + "-"
                + deviceID.toString().substring(DeviceID.DID_PREFIX.length(),
                        DeviceID.DID_PREFIX.length() + 5);

        createAndStartDevice(deviceID, name, deviceType);
    }

    /**
     * Starts a new device with a given device ID, name and type.
     * 
     * @throws DeviceFactoryException
     *             if the device cannot be started.
     */
    public void createAndStartDevice(DeviceID deviceID, String name,
            String deviceType) throws DeviceFactoryException {
        DeviceDescriptor descriptor = new DeviceDescriptor(deviceID);
        descriptor.name = name;
        descriptor.type = deviceType;
        createDevice(descriptor);

        startDevice(deviceID);
    }

    /** Stop the device with ID deviceID, if it is running. */
    public void stopDevice(DeviceID deviceID) throws IOException {
        SimulatedDeviceGUI gui = getDeviceGUI(deviceID);
        if (gui != null && gui.getGUI().isVisible()) {
            gui.close();
        }
        deviceGUIs.remove(deviceID);
    }

    public boolean deviceIsRunning(DeviceID deviceID) {
        SimulatedDeviceGUI device = getDeviceGUI(deviceID);
        if (device == null) {
            return false;
        } else {
            return device.getGUI().isVisible();
        }
    }
}
