/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.simulated.device.factory;

import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.simulated.device.DeviceDescriptor;
import ist.palcom.simulated.device.SimulatedDeviceGUI;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import javax.swing.DefaultListModel;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableModel;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class DeviceFactoryGUI {

    private static final String ACTION_NEW_SCENARIO = "New scenario...";

    private static final String ACTION_DELETE_SCENARIO = "Delete this scenario...";

    private static final String ACTION_STOP_ALL_DEVICES_IN_SCENARIO = "Stop all devices in this scenario";

    private static final String ACTION_START_ALL_DEVICES_IN_SCENARIO = "Start all devices in this scenario";

    private static final String ACTION_NEW_DEVICE = "New device...";

    private static final String ACTION_DELETE_SELECTED_DEVICES = "Delete...";

    private static final String ACTION_ADD_SELECTED_DEVICES_TO_A_SCENARIO = "Add to a scenario...";

    private static final String ACTION_REMOVE_SELECTED_DEVICES_FROM_SCENARIO = "Remove from this scenario";

    private static final String ACTION_START_SELECTED_DEVICES = "Start";

    private static final String ACTION_STOP_SELECTED_DEVICES = "Stop";

    private AbstractDeviceFactory factory;

    private JFrame frame;

    private JList scenariosList;

    private JTable devicesTable;

    private JScrollPane devicesScrollPane;

    private JMenu scenarioMenu;

    private JMenu deviceMenu;

    private JPopupMenu scenarioPopup;

    private JPopupMenu devicePopup;

    public DeviceFactoryGUI(AbstractDeviceFactory factory) {
        this.factory = factory;

        frame = new JFrame("Device factory");
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                frame.dispose();
            }
        });

        scenariosList = new JList(new DefaultListModel());
        scenariosList.addListSelectionListener(new ScenarioSelectionListener());
        scenariosList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        JScrollPane scenariosScrollpane = new JScrollPane(scenariosList);

        devicesTable = new JTable(new DefaultTableModel() {
            public boolean isCellEditable(int row, int column) {
                return false;
            }
        });
        getDevicesTableModel().addColumn("");
        getDevicesTableModel().addColumn("Device name");
        getDevicesTableModel().addColumn("Device ID");
        getDevicesTableModel().addColumn("Device type");
        devicesTable.getColumn("").setMaxWidth(20);
        devicesTable
                .setPreferredScrollableViewportSize(new Dimension(800, 600));
        devicesTable.getSelectionModel().addListSelectionListener(
                new DeviceSelectionListener());
        devicesTable
                .setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        devicesScrollPane = new JScrollPane(devicesTable);

        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                scenariosScrollpane, devicesScrollPane);

        frame.getContentPane().add(splitPane);

        createMenus();

        putAllScenariosInList();

        frame.pack();
        frame.setVisible(true);

        scenariosList.setSelectedIndex(0);

        enableActions();
    }

    private DefaultListModel getScenariosListModel() {
        return (DefaultListModel) scenariosList.getModel();
    }

    private void putAllScenariosInList() {
        getScenariosListModel().clear();
        getScenariosListModel().addElement("All devices");
        for (int i = 0; i < factory.getNumberOfScenarios(); i++) {
            String name = factory.getScenarioName(i);
            getScenariosListModel().addElement(name);
        }
    }

    private DefaultTableModel getDevicesTableModel() {
        return (DefaultTableModel) devicesTable.getModel();
    }

    private void putAllDevicesInTable() {
        getDevicesTableModel().setRowCount(0);
        for (int i = 0; i < factory.getNumberOfDevices(); i++) {
            addDeviceToTable(factory.getDeviceDescriptor(i));
        }
    }

    private void putDevicesForScenarioInTable(String scenarioName) {
        Iterator<DeviceID> iterator = factory.getDeviceIDIterator(scenarioName);
        getDevicesTableModel().setRowCount(0);
        while (iterator.hasNext()) {
            DeviceID deviceID = iterator.next();
            DeviceDescriptor deviceDescriptor = factory
                    .getDeviceDescriptor(deviceID);
            if (deviceDescriptor != null) {
                addDeviceToTable(deviceDescriptor);
            }
        }
    }

    private void addDeviceToTable(DeviceDescriptor descriptor) {
        getDevicesTableModel().addRow(
                new String[] { getDeviceRunningStatusIndicator(descriptor.deviceID),
                        descriptor.name, descriptor.deviceID.toString(),
                        descriptor.type });
    }

    private int getDeviceTableIndex(DeviceID deviceID) {
        DefaultTableModel model = getDevicesTableModel();
        for (int i = 0; i < model.getRowCount(); i++) {
            String id = (String) model.getValueAt(i, 2);
            if (id.equals(deviceID.toString())) {
                return i;
            }
        }
        return -1;
    }

    private String getDeviceRunningStatusIndicator(DeviceID deviceID) {
        if (factory.deviceIsRunning(deviceID)) {
            return "R";
        } else {
            return "";
        }
    }

    private void showDeviceRunningStatus(DeviceID deviceID) {
        int index = getDeviceTableIndex(deviceID);
        if (index != -1) {
            getDevicesTableModel().setValueAt(
                    getDeviceRunningStatusIndicator(deviceID), index, 0);
        }
    }

    /** @return the device IDs of the selected devices. */
    private DeviceID[] getSelectedDeviceIDs() {
        int[] selectedIndices = devicesTable.getSelectedRows();
        DeviceID[] deviceIDs = new DeviceID[selectedIndices.length];
        for (int i = 0; i < selectedIndices.length; i++) {
            deviceIDs[i] = new DeviceID((String) getDevicesTableModel().getValueAt(
                    selectedIndices[i], 2));
        }
        return deviceIDs;
    }

    private void createMenus() {
        JMenuBar menuBar = new JMenuBar();
        frame.setJMenuBar(menuBar);

        scenarioMenu = new JMenu("Scenario");
        menuBar.add(scenarioMenu);

        deviceMenu = new JMenu("Device");
        menuBar.add(deviceMenu);

        scenarioPopup = new JPopupMenu();
        scenariosList.addMouseListener(new PopupListener(scenarioPopup));

        devicePopup = new JPopupMenu();
        devicesTable.addMouseListener(new PopupListener(devicePopup));
        devicesScrollPane.addMouseListener(new PopupListener(devicePopup));

        addScenarioAction(ACTION_NEW_SCENARIO, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                createNewScenario();
            }
        });

        addScenarioAction(ACTION_DELETE_SCENARIO, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                deleteSelectedScenario();
            }
        });

        addScenarioSeparator();

        addScenarioAction(ACTION_START_ALL_DEVICES_IN_SCENARIO,
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        startAllDevicesInSelectedScenario();
                    }
                });

        addScenarioAction(ACTION_STOP_ALL_DEVICES_IN_SCENARIO,
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        stopAllDevicesInSelectedScenario();
                    }
                });

        addDeviceAction(ACTION_NEW_DEVICE, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                createNewDevice();
            }
        });

        addDeviceAction(ACTION_DELETE_SELECTED_DEVICES, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                deleteSelectedDevices();
            }
        });

        addDeviceSeparator();

        addDeviceAction(ACTION_START_SELECTED_DEVICES, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                startSelectedDevices();
            }
        });

        addDeviceAction(ACTION_STOP_SELECTED_DEVICES, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                stopSelectedDevices();
            }
        });

        addDeviceSeparator();

        addDeviceAction(ACTION_ADD_SELECTED_DEVICES_TO_A_SCENARIO,
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        addSelectedDevicesToAScenario();
                    }
                });

        addDeviceAction(ACTION_REMOVE_SELECTED_DEVICES_FROM_SCENARIO,
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        removeSelectedDevicesFromSelectedScenario();
                    }
                });
    }

    private void addScenarioAction(String action, ActionListener listener) {
        JMenuItem menuItem = new JMenuItem(action);
        menuItem.addActionListener(listener);
        scenarioMenu.add(menuItem);

        menuItem = new JMenuItem(action);
        menuItem.addActionListener(listener);
        scenarioPopup.add(menuItem);
    }

    private void addScenarioSeparator() {
        scenarioMenu.addSeparator();
        scenarioPopup.addSeparator();
    }

    private void addDeviceSeparator() {
        deviceMenu.addSeparator();
        devicePopup.addSeparator();
    }

    private void addDeviceAction(String action, ActionListener listener) {
        JMenuItem menuItem = new JMenuItem(action);
        menuItem.addActionListener(listener);
        deviceMenu.add(menuItem);

        menuItem = new JMenuItem(action);
        menuItem.addActionListener(listener);
        devicePopup.add(menuItem);
    }

    private void setActionEnabled(String action, boolean enabled) {
        setActionEnabled(action, enabled, scenarioMenu);
        setActionEnabled(action, enabled, deviceMenu);
        setActionEnabled(action, enabled, scenarioPopup);
        setActionEnabled(action, enabled, devicePopup);
    }

    private void setActionEnabled(String action, boolean enabled, JMenu menu) {
        for (int i = 0; i < menu.getMenuComponentCount(); i++) {
            if (menu.getMenuComponent(i) instanceof JMenuItem) {
                JMenuItem menuItem = (JMenuItem) menu.getMenuComponent(i);
                if (menuItem.getText().equals(action)) {
                    menuItem.setEnabled(enabled);
                }
            }
        }
    }

    private void setActionEnabled(String action, boolean enabled,
            JPopupMenu popup) {
        for (int i = 0; i < popup.getComponentCount(); i++) {
            if (popup.getComponent(i) instanceof JMenuItem) {
                JMenuItem menuItem = (JMenuItem) popup.getComponent(i);
                if (menuItem.getText().equals(action)) {
                    menuItem.setEnabled(enabled);
                }
            }
        }
    }

    private void enableActions() {
        int selectedScenarioIndex = scenariosList.getSelectedIndex();
        boolean allDevicesScenarioSelected = (selectedScenarioIndex == 0);
        setActionEnabled(ACTION_NEW_SCENARIO, true);
        setActionEnabled(ACTION_DELETE_SCENARIO, !allDevicesScenarioSelected
                && selectedScenarioIndex != -1);
        setActionEnabled(ACTION_START_ALL_DEVICES_IN_SCENARIO,
                !allDevicesScenarioSelected && selectedScenarioIndex != -1);
        setActionEnabled(ACTION_STOP_ALL_DEVICES_IN_SCENARIO,
                !allDevicesScenarioSelected && selectedScenarioIndex != -1);

        int selectedDeviceRowCount = devicesTable.getSelectedRowCount();
        setActionEnabled(ACTION_DELETE_SELECTED_DEVICES,
                selectedDeviceRowCount > 0);
        setActionEnabled(ACTION_START_SELECTED_DEVICES,
                selectedDeviceRowCount > 0);
        setActionEnabled(ACTION_STOP_SELECTED_DEVICES,
                selectedDeviceRowCount > 0);
        setActionEnabled(ACTION_ADD_SELECTED_DEVICES_TO_A_SCENARIO,
                selectedDeviceRowCount > 0
                        && factory.getNumberOfScenarios() > 0);
        setActionEnabled(ACTION_REMOVE_SELECTED_DEVICES_FROM_SCENARIO,
                selectedDeviceRowCount > 0 && !allDevicesScenarioSelected);
    }

    private void createNewScenario() {
        String name = JOptionPane.showInputDialog(frame,
                "Please enter a scenario name.", "New scenario",
                JOptionPane.QUESTION_MESSAGE);
        if (name != null) {
            if (factory.scenarioExists(name)) {
                JOptionPane
                        .showMessageDialog(frame, "A scenario with the name "
                                + name + " already exixts.", "New scenario",
                                JOptionPane.ERROR_MESSAGE);
            } else {
                factory.createScenario(name);
                getScenariosListModel().addElement(name);
            }
        }
    }

    private void deleteSelectedScenario() {
        if (scenariosList.getSelectedIndex() > 0) {
            // A scenario, not "All devices", is selected

            String name = (String) scenariosList.getSelectedValue();
            if (name != null) {
                int answer = JOptionPane.showConfirmDialog(frame,
                        "Are you sure you want to delete the scenario named "
                                + name + "?", "Delete scenario",
                        JOptionPane.YES_NO_OPTION);
                if (answer == JOptionPane.YES_OPTION) {
                    factory.deleteScenario(name);
                    getScenariosListModel().remove(
                            scenariosList.getSelectedIndex());
                    scenariosList.setSelectedIndex(0);
                }
            }
        }
    }

    private void startAllDevicesInSelectedScenario() {
        if (scenariosList.getSelectedIndex() > 0) {
            // A scenario, not "All devices", is selected

            String name = (String) scenariosList.getSelectedValue();
            Iterator<DeviceID> iterator = factory.getDeviceIDIterator(name);
            try {
                while (iterator.hasNext()) {
                    DeviceID deviceID = iterator.next();
                    startDevice(deviceID);
                }
            } catch (DeviceFactoryException e) {
                JOptionPane.showMessageDialog(frame, e.getMessage());
            }
        }
    }

    private void stopAllDevicesInSelectedScenario() {
        if (scenariosList.getSelectedIndex() > 0) {
            // A scenario, not "All devices", is selected

            String name = (String) scenariosList.getSelectedValue();
            Iterator<DeviceID> iterator = factory.getDeviceIDIterator(name);
            while (iterator.hasNext()) {
                DeviceID deviceID = iterator.next();
                try {
                    factory.stopDevice(deviceID);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(frame, e.toString(),
                            "Stop device", JOptionPane.ERROR_MESSAGE);
                }
                showDeviceRunningStatus(deviceID);
            }
        }
    }

    private void createNewDevice() {
        DeviceID deviceID = factory.getNewDeviceID();

        final String nl = System.getProperty("line.separator");
        String type = askForDeviceType();
        if (type != null) {
            String name = JOptionPane.showInputDialog(frame,
                    "Your new device will get the following device ID: " + deviceID
                            + "." + nl + nl + "What is the device name?",
                    "New device", JOptionPane.QUESTION_MESSAGE);
            if (name != null) {
                DeviceDescriptor descriptor = new DeviceDescriptor(deviceID);
                descriptor.name = name;
                descriptor.type = type;
                factory.createDevice(descriptor);
                if (scenariosList.getSelectedIndex() == 0) {
                    // "All devices" is selected
                    addDeviceToTable(descriptor);
                } else {
                    // Select "All devices" so that the new device shows
                    scenariosList.setSelectedIndex(0);
                }
            }
        }
    }

    private void deleteSelectedDevices() {
        DeviceID[] deviceIDs = getSelectedDeviceIDs();
        if (deviceIDs.length > 0) {
            int answer = JOptionPane
                    .showConfirmDialog(
                            frame,
                            deviceIDs.length == 1 ? "Are you sure you want to delete the selected device?"
                                    : "Are you sure you want to delete the "
                                            + deviceIDs.length
                                            + " selected devices?", "Delete",
                            JOptionPane.YES_NO_OPTION);
            if (answer == JOptionPane.YES_OPTION) {
                try {
                    for (int i = 0; i < deviceIDs.length; i++) {
                        factory.deleteDevice(deviceIDs[i]);
                    }
                    while (devicesTable.getSelectedRowCount() > 0) {
                        getDevicesTableModel().removeRow(
                                devicesTable.getSelectedRows()[0]);
                    }
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(frame, e.toString(),
                            "Delete device", JOptionPane.ERROR_MESSAGE);
                }
            }
        }
    }

    private void addSelectedDevicesToAScenario() {
        if (factory.getNumberOfScenarios() == 0) {
            JOptionPane.showMessageDialog(frame,
                    "There are no scenarios. Create one first.",
                    "Add device to scenario", JOptionPane.ERROR_MESSAGE);
        } else {
            DeviceID[] deviceIDs = getSelectedDeviceIDs();
            if (deviceIDs.length > 0) {
                String[] scenarioNames = new String[factory
                        .getNumberOfScenarios()];
                for (int i = 0; i < scenarioNames.length; i++) {
                    scenarioNames[i] = factory.getScenarioName(i);
                }
                String scenarioName = (String) JOptionPane.showInputDialog(
                        frame, "Scenario:", "Add to scenario",
                        JOptionPane.QUESTION_MESSAGE, null, scenarioNames,
                        scenarioNames[0]);
                if (scenarioName != null) {
                    for (int i = 0; i < deviceIDs.length; i++) {
                        factory
                                .addDeviceToScenario(deviceIDs[i],
                                        scenarioName);
                    }

                    // We don't need to redraw the table here, because if the
                    // device was added again to the current scenario, it is
                    // already there
                }
            }
        }
    }

    private void removeSelectedDevicesFromSelectedScenario() {
        DeviceID[] deviceIDs = getSelectedDeviceIDs();
        if (deviceIDs.length > 0) {
            if (scenariosList.getSelectedIndex() > 0) {
                // A real scenario is selected, not "All devices"
                String scenarioName = (String) scenariosList.getSelectedValue();
                for (int i = 0; i < deviceIDs.length; i++) {
                    factory.removeDeviceFromScenario(deviceIDs[i],
                            scenarioName);
                }
                while (devicesTable.getSelectedRowCount() > 0) {
                    getDevicesTableModel().removeRow(
                            devicesTable.getSelectedRows()[0]);
                }
            }
        }
    }

    private String askForDeviceType() {
        try {
            List<String> classNames = factory.getDeviceGUIClassNames();

            if (classNames.isEmpty()) {
                JOptionPane.showMessageDialog(frame,
                        "There are no registered device GUI types.");
            } else {
                String selectedClassName = (String) JOptionPane
                        .showInputDialog(frame, "Device type:", "New device",
                                JOptionPane.QUESTION_MESSAGE, null, classNames
                                        .toArray(), classNames.get(0));
                for (int i = 0; i < classNames.size(); i++) {
                    if (classNames.get(i).equals(selectedClassName)) {
                        return classNames.get(i);
                    }
                }
            }
        } catch (IOException e) {
            JOptionPane.showMessageDialog(frame, e.toString());
        }
        return null;
    }

    private void startDevice(DeviceID deviceID) throws DeviceFactoryException {
        SimulatedDeviceGUI device = factory.startDevice(deviceID);
        if (device != null) {
            device.getGUI().addWindowListener(
                    new DeviceWindowListener(deviceID));
            device.getGUI().requestFocus();
        }
        showDeviceRunningStatus(deviceID);
    }

    private void startSelectedDevices() {
        try {
            DeviceID[] deviceIDs = getSelectedDeviceIDs();
            for (int i = 0; i < deviceIDs.length; i++) {
                startDevice(deviceIDs[i]);
            }
        } catch (DeviceFactoryException e) {
            JOptionPane.showMessageDialog(frame, e.getMessage());
        }
    }

    private void stopSelectedDevices() {
        DeviceID[] deviceIDs = getSelectedDeviceIDs();
        if (deviceIDs.length > 0) {
            for (int i = 0; i < deviceIDs.length; i++) {
                try {
                    factory.stopDevice(deviceIDs[i]);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(frame, e.toString(),
                            "Stop device", JOptionPane.ERROR_MESSAGE);
                }
                showDeviceRunningStatus(deviceIDs[i]);
            }
        }
    }

    private class PopupListener extends MouseAdapter {
        private JPopupMenu popup;

        PopupListener(JPopupMenu popup) {
            this.popup = popup;
        }

        public void mousePressed(MouseEvent e) {
            maybeShowPopup(e);
        }

        public void mouseReleased(MouseEvent e) {
            maybeShowPopup(e);
        }

        private void maybeShowPopup(MouseEvent e) {
            if (e.isPopupTrigger()) {
                popup.show(e.getComponent(), e.getX(), e.getY());
            }
        }
    }

    private class ScenarioSelectionListener implements ListSelectionListener {
        public void valueChanged(ListSelectionEvent e) {
            // Ignore extra messages.
            if (e.getValueIsAdjusting())
                return;

            int index = scenariosList.getSelectedIndex();
            if (index == 0) {
                // Show all devices
                putAllDevicesInTable();
            } else if (index > 0) {
                // A real scenario is selected
                String scenarioName = (String) scenariosList.getSelectedValue();
                putDevicesForScenarioInTable(scenarioName);
            }
            enableActions();
        }
    }

    private class DeviceSelectionListener implements ListSelectionListener {
        public void valueChanged(ListSelectionEvent e) {
            // Ignore extra messages.
            if (e.getValueIsAdjusting())
                return;

            enableActions();
        }
    }

    private class DeviceWindowListener extends WindowAdapter {
        private DeviceID deviceID;

        DeviceWindowListener(DeviceID deviceID) {
            this.deviceID = deviceID;
        }

        public void windowClosed(WindowEvent e) {
            showDeviceRunningStatus(deviceID);
        }
    }
}