/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.simulated.device.widgets;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.image.BufferedImage;
import java.net.URL;

import javax.swing.JPanel;

/**
 * An ImagePanel that scales its image to fit the panel, and for which the image
 * can be changed.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class ScalingImagePanel extends JPanel {

    /** The image. */
    private Image image;

    /**
     * An image that has been scaled to the current size of the panel, or null
     * if there is no image. This is protected by synchronized methods.
     */
    private BufferedImage scaledImage;

    // TODO Use a Monitor instead of synchronized methods

    public ScalingImagePanel() {
        this(null);
    }

    public ScalingImagePanel(URL imageResource) {
        Image newImage;
        if (imageResource != null) {
            newImage = Toolkit.getDefaultToolkit().getImage(imageResource);
            MediaTracker tracker = new MediaTracker(this);
            tracker.addImage(newImage, 0);
            try {
                tracker.waitForID(0);
            } catch (InterruptedException exception) {
            }
        } else {
            newImage = null;
        }
        setImage(newImage);
        addComponentListener(new ResizeListener());
    }

    /** Set the image to show. If image is null, no image will be shown. */
    public void setImage(Image image) {
        this.image = image;
        createScaledImage();
    }

    private synchronized void createScaledImage() {
        int boundsWidth = this.getWidth();
        int boundsHeight = this.getHeight();

        if (image != null && boundsWidth > 0 && boundsHeight > 0) {
            int imageWidth = image.getWidth(this);
            int imageHeight = image.getHeight(this);
            double imageRatio = (double) imageHeight / imageWidth;
            double boundsRatio = (double) boundsHeight / boundsWidth;
            int height, width;
            if (imageRatio > boundsRatio) {
                height = boundsHeight;
                width = (int) (height / imageRatio);
            } else {
                width = boundsWidth;
                height = (int) (width * imageRatio);
            }

            scaledImage = new BufferedImage(width, height,
                    BufferedImage.TYPE_INT_RGB);
            Graphics2D graphics2D = scaledImage.createGraphics();
            graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
                    RenderingHints.VALUE_INTERPOLATION_BILINEAR);
            graphics2D.drawImage(image, 0, 0, width, height, null);
        } else {
            scaledImage = null;
        }
    }

    private synchronized Image getScaledImage() {
        return scaledImage;
    }

    private class ResizeListener extends ComponentAdapter {
        public void componentResized(ComponentEvent e) {
            createScaledImage();
        }
    }

    public void paintComponent(Graphics g) {
        super.paintComponent(g);

        Image scaledImage = getScaledImage();
        if (scaledImage != null) {
            int width = scaledImage.getWidth(this);
            int height = scaledImage.getHeight(this);
            int x = (this.getWidth() - width) / 2;
            int y = (this.getHeight() - height) / 2;
            g.drawImage(scaledImage, x, y, width, height, null);
        }
    }
}
